/* copyright (c) 1996, 1997, 1998, 1999 William R. Pearson and the
   U. of Virginia */

/* $Name: fa33t09d0 $ - $Id: comp_thr.c,v 1.22.2.9.2.2 2001/03/29 22:29:43 wrp Exp $ */

/* threaded version of complib.c 

 *	December, 1998 - DNA searches are now down with forward and reverse
 *			 strands
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#ifdef UNIX
#if !defined(__NetBSD__) && !defined(MacOSX)
#include <values.h>
#else
#include <limits.h>
#define MAXLONG LONG_MAX
#endif
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#else
#define MAXFLOAT 1E+37
#endif

#include <math.h>

#include "mw.h"			/* defines beststr */
#include "defs.h"		/* various constants */
#include "structs.h"		/* mngmsg, libstruct */
#include "param.h"		/* pstruct, thr_str, buf_head, rstruct */

#define XTERNAL
#include "upam.h"
#include "uascii.h"
#undef XTERNAL

char *mp_verstr="33t08";

/********************************/
/* global variable declarations */
/********************************/
char gstring2[MAX_STR];                  /* string for label */
char gstring3[MAX_STR];
char hstring1[MAX_STR];
extern int max_workers;

/********************************/
/* extern variable declarations */
/********************************/
extern char *prog_func;		/* function label */
extern char *verstr, *iprompt0, *iprompt1, *iprompt2, *refstr;

/********************************/
/*extern function declarations  */
/********************************/
/* open sequence file (nxgetaa.c) */
#ifndef FASTX
extern int getseq(char *filen, unsigned char *seq, int maxs, char *libstr,
		   int *dnaseq,long *sq0ff);
#else
extern int getntseq(char *filen, unsigned char *seq, int maxs, char *libstr,
		   int *dnaseq,long *sq0ff);
#endif
#ifndef PRSS
extern int openlib(int iln, char *lname, int *lb_types, int ldnaseq);
#else
extern int openlib(char *lname, struct mngmsg m_msg);
#endif

extern void closelib();

 /* check for DNA sequence (nxgetaa.c) */
extern int scanseq(unsigned char *seq, int n, char *str);
extern void revcomp(unsigned char *seq, int n);

/* initialize environment (doinit.c) */
extern void initenv (int argc, char **argv, struct mngmsg *m_msg,
		     struct pstruct *ppst, unsigned char **aa0, int outtty);
/* print timing information */
extern void ptime (FILE *, time_t);
/* reset parameters if DNA sequence (initxx.c) */
extern void resetp (unsigned char *aa0, int n0,
	struct mngmsg *m_msg, struct pstruct *ppst);
/* read a sequence (nxgetaa.c) */
extern int (*getlib) (unsigned char *seq, int maxs,
		      char *libstr,
#ifndef USE_FSEEKO
		      long *libpos,
#else
		      off_t *libpos,
#endif
		      int *lcont);
#define GETLIB (*getlib)

/* functions for getting/sending buffers to threads (thr_sub.c) */
extern void init_thr(int nthreads, struct thr_str *work_info);
extern void start_thr(void);
extern void get_rbuf(struct buf_head **cur_buf, int max_wor_buf);
extern void put_rbuf(struct buf_head *cur_buf, int max_work_buf);
extern void put_rbuf_done(int nthreads, struct buf_head *cur_buf, 
			  int max_work_buf);
#ifndef PRSS
void save_best(struct buf_head *cur_buf, struct pstruct pst, int nitt1,
	       FILE *fdata, int *);
#else
void save_best(struct buf_head *cur_buf, struct pstruct pst, int nitt1,
	       FILE *fdata, int *, int *, int *);
#endif
void selectbestz(struct beststr **, int, int );

/* statistics functions */

extern void process_hist(struct stat_str *sptr, int nstat, struct pstruct pst);
extern double zs_to_E(float zs, int n1, int isdna);
extern float (*find_zp)(int score, int length, double comp);
extern void addhistz(float);

/* function specific calls (initxx.c, dropxx.c) */
extern void init_work (unsigned char *aa0, int n0, struct pstruct *ppst,
		       void **f_arg);
extern void close_work (unsigned char *aa0, int n0, struct pstruct *ppst,
			void **f_arg);
extern void get_param (struct pstruct *pstr, char *pstring1, char *pstring2);

/* display functions */
#ifndef PRSS
extern void showbest (FILE *fp, unsigned char **aa0, unsigned char *aa1,
		       int maxn, struct beststr **bptr, int nbest,
		       int qlib, struct mngmsg *m_msg,struct pstruct pst,
		       void **f_str, char *gstring2);

#else
extern void showbest (FILE *fp, int *s_score, int maxn, 
		       int qlib, struct mngmsg *m_msg,struct pstruct pst,
		       char *gstring2);
#endif

extern void showalign (FILE *fp, unsigned char **aa0, unsigned char *aa1,
		       int maxn, struct beststr **bptr, int nbest,
		       int qlib, struct mngmsg m_msg,struct pstruct pst,
		       void **f_str, char *gstring2, int *lb_current);

/* misc functions */
void h_init(struct pstruct *);		/* doinit.c */
void last_init(struct mngmsg *, struct pstruct *); /* initfa/sw.c */
void s_abort(char *, char *);			/* compacc.c */
void revcomp(unsigned char *, int);	/* nxgetaa.c */
#ifndef FASTX
void resetp(unsigned char *, int, struct mngmsg *, struct pstruct *); /* initfa/sw.c */
#else
void reseta(unsigned char *, int, struct mngmsg *, struct pstruct *); /* initfa/sw.c */
#endif
void gettitle(char *, char *, int);	/* nxgetaa.c */
void libchoice(char *lname, int, struct mngmsg *); /* lib_sel.c */
void libselect(char *lname, struct mngmsg *);	/* lib_sel.c */
void query_parm(struct mngmsg *, struct pstruct *); /* initfa/sw.c */
void addhistz(float zscore);		/* scaleswn.c */
void selectbestz(struct beststr **, int, int);
void sortbestz(struct beststr **, int);
void sortbeste(struct beststr **, int);
void sortbest(struct beststr **, int, int);
void prhist(FILE *, struct mngmsg, struct pstruct, char *);
void printsum(FILE *);

FILE *outfd;			/* Output file */
int outtty=1; 		 	/* request tty input unless told otherwise */

/* this information is global for fsigint() */
time_t tstart, tscan, tprev, tdone, s_time();	/* Timing */
time_t tdstart, tddone, time();

int  nlib = 0;			/* number of sequences scanned */
int nclib=0, qlib = 0;	
struct total_db_str ntt, qtt;

extern int nsfnum;
extern int sfnum[];
#ifdef SUPERFAMNUM
extern int nsfnum_n;
extern int sfnum_n[];
#endif

/***************************************/
/* thread global variable declarations */
/***************************************/

#undef XTERNAL
#include "thr.h"
struct buf_head buf_list[NUM_WORK_BUF];
 
struct beststr *best,		/* array of best scores */
	       *bestptr,
               **bptr;		/* array of pointers for sorts */
int nbest, nres;		/* number of best scores */

struct stat_str *stats;
int nstats;

float zbestcut;	/* cut off value for best scores */
int bestfull;			/* index for selectbest() */
int dohist=0;			/* flag for z-value processing */

main(int argc, char **argv)
{
  unsigned char *aa0[6], *aa1, *aa1ptr, *aa1s;
  int n1;
  int iln, itt, nitt1;
  char lib_label[MAX_FN];
#ifdef PRSS
  int s_score[3];
  int s_n1;
#endif

  long ntbuff;
  struct pstruct pst;
  
  struct mngmsg m_msg;		/* Message from host to manager */
  char rline[MAX_FN];
  int i, t_quiet;

  FILE *fdata=NULL;		/* file for full results */
  char libstr[21];		/* string for labeling full results */
  
  int node, snode;		/* Number of nodes */
  int bufid, numt, tid;

  int leng,type;		/* leng is length of the descriptive line */
  int maxn = MAXTOT;	/* max size of the query sequence */
#ifndef USE_FSEEKO
  long lmark;
#else
  off_t lmark;
#endif
  int lcont, ocont, maxt;	/* continued sequence */
  int counr = 0;		/* count of number of query sequences */
  int var;			/* dummy variable for message */
  char errstr[120];
  int igncnt=0;			/* count for ignoring sequences warning */
  float zscore;			/* tmp value */
  char *bp;
  void *f_str[6];			/* function-specific structure */
  int max_buf_cnt, ave_seq_len, buf_siz;

  void fsigint();

  struct thr_str *work_info;
  int max_work_buf;

  void *exit_value;

  struct buf_head *cur_buf;
  int nseq;
  
#ifdef SGI_BUG
 argv[-1*0x20000] = 0;
#endif

#ifdef UNIX
  outtty=isatty(1);
#else
  outtty=1;
#endif

  h_init(&pst);
  
  ntt.length = qtt.length = 0;
  ntt.carry = qtt.carry = 0;

  aa0[0]=NULL;
  initenv (argc, argv, &m_msg, &pst, &aa0[0], outtty);
  if (m_msg.quiet) outtty = 0;

#ifndef PRSS
  if (m_msg.n1_low > 0 && m_msg.n1_high < BIGNUM) 
    sprintf(lib_label,"library (range: %d-%d)",m_msg.n1_low,m_msg.n1_high);
  else if (m_msg.n1_low > 0) 
    sprintf(lib_label,"library (range: >%d)",m_msg.n1_low);
  else if (m_msg.n1_high < BIGNUM)
    sprintf(lib_label,"library (range: <%d)",m_msg.n1_high);
  else
    strncpy(lib_label,"library",sizeof(lib_label));
#else
  sprintf(lib_label,"shuffled sequence");
#endif

  /*
  fprintf (stderr, "Threaded library processor(%d processes)\n",max_workers);
  fprintf (stderr, "Using %s\n", prog_func);
  */

  tstart = s_time();
  tdstart = time(NULL);
  
  /* allocate space for query sequence */

  if (aa0[0]==NULL) {
    if ((aa0[0] = (unsigned char *) malloc ( (MAXTOT+1) * sizeof (unsigned char)))
	== NULL)
      s_abort ("Unable to allocate query sequence", "");
    *aa0[0]=0;
    aa0[0]++;
  }
  aa0[5] = aa0[4] = aa0[3] =  aa0[2] =  aa0[1] =  aa0[0];

  if (m_msg.markx == 6) {
    fprintf(stdout,"<html>\n<head>\n<title>%s Results\n",
	    prog_func);
    fprintf(stdout,"</title>\n</head>\n<body>\n<pre>\n");
  }

  fputs(iprompt0,stdout);
  fprintf(stdout," %s%s\n",verstr,refstr);

  /* Query library */
  if (m_msg.tname[0] == '\0')
    {
      if (m_msg.quiet == 1)
	s_abort("Query sequence undefined","");
      
    l1:	fputs (iprompt1, stdout);
      fflush  (stdout);
      if (fgets (m_msg.tname, 80, stdin) == NULL)
	s_abort ("Unable to read query library name","");
      if ((bp=strchr(m_msg.tname,'\n'))!=NULL) *bp='\0';
      if (m_msg.tname[0] == '\0') goto l1;
    }
  
  /* read query sequence */

  lcont = 0;
#ifndef FASTX
  m_msg.n0 = getseq (m_msg.tname,aa0[0], MAXTOT, m_msg.libstr,
		     &pst.dnaseq,&m_msg.sq0off);
#else
  m_msg.n0 = getntseq (m_msg.tname,aa0[0], MAXTOT, m_msg.libstr,
		     &pst.dnaseq,&m_msg.sq0off);
#endif

#ifdef SUPERFAMNUM
  m_msg.nqsfnum = nsfnum;
  for (i=0; i<nsfnum; i++) m_msg.qsfnum[i] = sfnum[i];
  m_msg.nqsfnum_n = nsfnum_n;
  for (i=0; i<nsfnum_n; i++) m_msg.qsfnum_n[i] = sfnum_n[i];
#endif

  if (m_msg.n0 > MAXTST) {
    fprintf(stderr," sequence truncated to %d\n %s\n",MAXTST,sqnam);
    fprintf(stdout," sequence truncated to %d\n %s\n",MAXTST,sqnam);
    aa0[0][MAXTST]=EOSEQ;
    m_msg.n0=MAXTST;
  }
  pst.n0 = m_msg.n0;

#ifdef SUPERFAMNUM
  m_msg.nqsfnum = nsfnum;
  for (i=0; i <= nsfnum & i<10; i++) m_msg.qsfnum[i] = sfnum[i];
  m_msg.nqsfnum_n = nsfnum_n;
  for (i=0; i <= nsfnum_n & i<10; i++) m_msg.qsfnum_n[i] = sfnum_n[i];
#endif

  if (m_msg.n0 <= 0)
    s_abort ("Query sequence length <= 0: ", m_msg.tname);

#if !defined(TFASTX) && !defined(TFASTA)
  /* a sanity check */
   if (m_msg.revcomp && !(pst.dnaseq==1)) {
     fprintf(stderr," cannot reverse complement %s\n",sqtype);
     m_msg.revcomp = 0;
   }
   /*  if (m_msg.revcomp) revcomp(aa0[0],m_msg.n0); */
#endif

#ifndef FASTX
  resetp (aa0[0], m_msg.n0, &m_msg, &pst);
#else
  reseta (aa0[0], m_msg.n0, &m_msg, &pst);
#endif
  gettitle(m_msg.tname,m_msg.qtitle,sizeof(m_msg.qtitle));
  if (strcmp(m_msg.tname,"-")==0) {
    strncmp(m_msg.tname,m_msg.qtitle,sizeof(m_msg.tname));
    if ((bp=strchr(m_msg.tname,' '))!=NULL) *bp='\0';
    if ((bp=strchr(&m_msg.tname[4],'|'))!=NULL) *bp='\0';
  }

  maxn = MAXTOT - m_msg.n0 -2;
  if (m_msg.maxn > 0 && m_msg.maxn < maxn) maxn = m_msg.maxn;

#ifndef FASTX
  if (pst.dnaseq==1) m_msg.ldnaseq = 1;
#else
  aainit(pst.tr_type, pst.debug_lib);
#endif


#if !defined(TFASTA) && !defined(TFASTX)
  if (pst.dnaseq==1 && m_msg.n0>(MAXTST+MAXLIB)/3) {
    fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
    exit(1);
  }
  else if (pst.dnaseq==0 && m_msg.n0>(MAXTST+MAXLIB)/2) {
    fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
    exit(1);
  }

  m_msg.loff = m_msg.n0;
  m_msg.maxt3 = maxn-m_msg.loff;
#else        	
  if (sqtype[0]=='D' || pst.dnaseq == 1) {
    fprintf(stderr," %s compares a protein to a translated\n\
DNA sequence library.  Do not use a DNA query sequence/matrix.\n",prog_func);
    exit(1);
  }
  aainit(pst.tr_type,pst.debug_lib);
  m_msg.ldnaseq = 1;

  if (m_msg.n0 > MAXTST) {
    fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
    exit(1);
  }
  if (m_msg.n0*6 > maxn ) {	/* n0 for the complement + n0*3 for
                                   the three frames */
    if (m_msg.n0*7+2 < (MAXTST+MAXLIB)) { /* n0*6 + n0 */
      fprintf(stderr,
	      " query sequence too long for library segment: %d - resetting to %d\n",
	      maxn, m_msg.n0*6 );
      maxn = m_msg.maxn = m_msg.n0*6;
    }
    else {
      fprintf(stderr," query sequence too long for translated search: %d %s\n",
	      m_msg.n0,sqnam);
      exit(1);
    }
  }

  /* set up some constants for overlaps */
  m_msg.loff = 3*m_msg.n0;
  m_msg.maxt3 = maxn-m_msg.loff-3; m_msg.maxt3 -= m_msg.maxt3%3; m_msg.maxt3++;
  maxn = maxn-3; maxn -= maxn%3; maxn++; 
#endif

  pst.maxlen = maxn;

  /* set up some space and constants for overlaps */
  if ( 2*(m_msg.n0+1) > MAXTOT) {
    if ((aa1s=(unsigned char*)calloc((size_t)(m_msg.n0+1),sizeof(unsigned char)))==NULL) {
      fprintf(stderr," cannot allocate aa1s array %d\n",m_msg.n0+1);
      exit(1);
    }
  }
  else aa1s = aa0[1] + m_msg.n0 + 2;

  if (pst.dnaseq==1) m_msg.ldnaseq = 1;

  counr = 0;
  leng = strlen (m_msg.libstr);
  sprintf (&m_msg.libstr[leng], " %d %s", m_msg.n0, sqnam);
  m_msg.seqnm = counr;

  /* get library file names */

#ifndef PRSS
  if (strlen (m_msg.lname) == 0) {
    if (m_msg.quiet == 1) s_abort("library name undefined","");
    libchoice(m_msg.lname, sizeof(m_msg.lname), &m_msg);
  }

  libselect(m_msg.lname, &m_msg);
#else
  if (strlen (m_msg.lname) == 0) {
    if (m_msg.quiet == 1) s_abort("Shuffle sequence undefined","");
l2:    fputs(iprompt2,stdout);
    fflush(stdout);
    if (fgets (m_msg.lname, MAX_FN, stdin) == NULL)
      s_abort ("Unable to read shuffle file name","");
    if ((bp=strchr(m_msg.lname,'\n'))!=NULL) *bp='\0';
    if (m_msg.lname[0] == '\0') goto l2;
  }
  m_msg.lbnames[0]= m_msg.lname;
  m_msg.nln = 1;
  m_msg.nshow = 0;
#endif

  /* Get additional parameters here */
  if (!m_msg.quiet) query_parm (&m_msg, &pst);

  last_init(&m_msg, &pst);

#if !defined(TFASTA) && !defined(TFASTX)
  nitt1 = m_msg.nitt1 = m_msg.qframe-1;
#else
  nitt1 = m_msg.nitt1 = m_msg.nframe-1;
#endif

  /* Allocate space for saved scores */
  if ((best = 
       (struct beststr *)malloc((MAXBEST+1)*sizeof(struct beststr)))==NULL)
    s_abort ("Cannot allocate best struct","");
  if ((bptr = 
       (struct beststr **)malloc((MAXBEST+1)*sizeof(struct beststr *)))==NULL)
    s_abort ("Cannot allocate bptr","");
  
  /* Initialize bptr */
  for (nbest = 0; nbest < MAXBEST+1; nbest++)
    bptr[nbest] = &best[nbest];
  best++; bptr++;
  best[-1].score[0]=best[-1].score[1]=best[-1].score[2]=MAXINT;
  best[-1].zscore=MAXFLOAT;
  nbest = 0;
  zbestcut = -MAXFLOAT;

  if ((stats =
       (struct stat_str *)malloc((MAXSTATS)*sizeof(struct stat_str)))==NULL)
    s_abort ("Cannot allocate stats struct","");
  nstats = 0;

  outfd = stdout;

  /* set up the thr_str to pass info to threads */
  if ((work_info=(struct thr_str *)calloc(max_workers,
					  sizeof(struct thr_str)))==NULL) {
    fprintf(stderr, " cannot allocate work_info[%d]\n",max_workers);
    exit(1);
  }

  for (i=0; i<max_workers; i++) {
    work_info[i].n0 = m_msg.n0;
    work_info[i].qframe = m_msg.qframe;
    work_info[i].pst = &pst;
    work_info[i].aa0 = aa0[0];
    work_info[i].worker=0;
  }

  /* Set up buffer for reading the library:

     We will start by using a 2 Mbyte buffer for each worker.  For
     proteins, that means 5,000 sequences of length 400 (average).
     For DNA, that means 2,000 sequences of length 1000.  At the moment,
     those are good averages.

     */

  if (m_msg.ldnaseq==1) {
    max_buf_cnt = MAX_NT_BUF;
    ave_seq_len = AVE_NT_LEN;
  }
  else {
    max_buf_cnt = MAX_AA_BUF;
    ave_seq_len = AVE_AA_LEN;
  }

  /* however - buffer sizes should be a function of the number of
     workers so that all the workers are kept busy.  Assuming a 10,000
     entry library is the smallest we want to schedule, then
     */

  if (max_buf_cnt > 10000/max_workers) 
    max_buf_cnt = 10000/(2*max_workers);

  max_buf_cnt /= m_msg.thr_fact;

  /* finally, max_work_buf should be %6 for tfasta */
  max_buf_cnt += 6 - (max_buf_cnt % 6);

  max_work_buf = 2*max_workers;
  for (i=0; i<max_workers; i++)
    work_info[i].max_work_buf = max_work_buf;

  /* allocate space for library buffers and results */

  buf_siz=max_buf_cnt*ave_seq_len;
  if (buf_siz < 3*maxn) buf_siz = 3*maxn;
  for (i=0; i<max_work_buf; i++) {
    if ((buf_list[i].buf =
         (struct buf_str *)calloc((size_t)(max_buf_cnt+1),
				  sizeof(struct buf_str)))
        ==NULL) {
      fprintf(stderr," cannot allocate buffer struct %d %d\n",i,max_buf_cnt+1);
      exit(1);
    }
    buf_list[i].buf_cnt=0;
    buf_list[i].have_results=0;
    if ((buf_list[i].start =
         (unsigned char *)calloc((size_t)(buf_siz),sizeof(unsigned char)))
        ==NULL) {
      fprintf(stderr," cannot allocate buffer %d\n",i);
      exit(1);
    }
    reader_buf[i] = &buf_list[i];
  }

  /* initialization of global variables for threads */

  num_worker_bufs = 0;
  num_reader_bufs = max_work_buf;
  reader_done = 0;
  worker_buf_workp = 0;
  worker_buf_readp = 0;
  reader_buf_workp = 0;
  reader_buf_readp = 0;
  start_thread = 1;

  init_thr(max_workers, work_info);

  /* Label the output */
  if ((bp = (char *) strchr (m_msg.lname, ' ')) != NULL) *bp = '\0';
  
  if (m_msg.ltitle[0] == '\0')
    strncpy(m_msg.ltitle,m_msg.lname,sizeof(m_msg.ltitle));

#if !defined(TFASTX) && !defined(TFASTA) && !defined(PRSS)
  printf("%.50s: %d %s%s\n %s\n vs  %.60s %s\n",
	 m_msg.tname, m_msg.n0, m_msg.sqnam, (m_msg.revcomp)?
	 " (reverse-complement)":"\0", m_msg.qtitle,m_msg.ltitle,
	 lib_label);

  if (outfd != stdout)
    fprintf (outfd, "Query sequence: %.50s (%d %s)%s\n vs  %.60s %s\n",
	     m_msg.tname, m_msg.n0, m_msg.sqnam, (m_msg.revcomp)?
	     " (reverse-complement)":"\0",m_msg.lname,
	     lib_label);
#else
#ifndef PRSS
  printf("%.50s: %d %s%s\n %s\n vs  %.60s %s\n",
	 m_msg.tname, m_msg.n0, m_msg.sqnam, 
	 (m_msg.revcomp ? " (reverse complement)" : (m_msg.nframe==1)?
	  " (forward-only)" : "\0"),
	 m_msg.qtitle,m_msg.ltitle, lib_label);

  if (outfd != stdout)
    fprintf (outfd, "Query sequence: %.50s (%d %s)%s\n vs  %.60s %s\n",
	     m_msg.tname, m_msg.n0, m_msg.sqnam,
	     (m_msg.revcomp ? " (reverse complement)" : (m_msg.nframe==1)?
	      " (forward-only)" : "\0"),m_msg.lname,
	     lib_label);
#endif
#endif
  fflush (outfd);

  tprev = s_time();
  
  if (m_msg.dfile[0] && (fdata=fopen(m_msg.dfile,"w"))!=NULL)
    fprintf(fdata,"%3d%.50s\n",qlib,m_msg.qtitle);

  qtt.length += m_msg.n0;
  qlib++;

  start_thr();

  /* now open the library and start reading */
  /* get a buffer and fill it up */
  get_rbuf(&cur_buf,max_work_buf);

  cur_buf->buf_cnt = 0;
  cur_buf->have_results = 0;
  cur_buf->buf[0].aa1b = cur_buf->start;
  ntbuff = 0;
  nseq = 0;

  /* iln counts through each library */
  for (iln = 0; iln < m_msg.nln; iln++) {
    m_msg.lb_current = iln;
#ifndef PRSS
    if (openlib(iln, m_msg.lbnames[iln], m_msg.lb_types, m_msg.ldnaseq)!=1)
#else
    if (openlib(m_msg.lbnames[0], m_msg)!=1)
#endif
      {
	fprintf(stderr," cannot open library %s\n",m_msg.lbnames[iln]);
	continue;
      }
#ifndef PRSS
    else {
      printf ("searching %s library\n",m_msg.lbnames[iln]);
    }
#endif
    lcont = 0;
    ocont = 0;

    /* get next buffer to read into */
    maxt = maxn;

    /* read sequence directly into buffer */
    aa1ptr = aa1 = cur_buf->buf[nseq].aa1b;

    while ((n1=GETLIB(aa1ptr,maxt,libstr,&lmark,&lcont))>= 0) {
      if (n1 < m_msg.n1_low || n1 > m_msg.n1_high) goto loop1;
      nlib++;		/* inc number of sequences */
      ntt.length += n1;	/* update total library length */
      if (ntt.length > MAXLONG) {ntt.length -= MAXLONG; ntt.carry++;}

#ifdef DEBUG
      /* This discovers most reasons for core dumps */
      if (pst.debug_lib)
	for (i=0; i<n1; i++)
	  if (aa1[i]>pst.nsq) {
	    fprintf(stderr,
		    "%s residue[%d/%d] %d range (%d) lcont/ocont: %d/%d\n%s\n",
		    libstr,i,n1,aa1[i],pst.nsq,lcont,ocont,aa1ptr+i);
	    aa1[i]=0;
	    n1=i-1;
	    break;
	  }
#endif
      /* don't count long sequences more than once */
      if (aa1!=aa1ptr) {n1 += m_msg.loff; nlib--;}

      if (n1>1) {
	ntbuff += n1+1;

	for (itt=m_msg.revcomp; itt<=nitt1; itt++) {

	  cur_buf->buf_cnt++;
	  cur_buf->buf[nseq].n1  = n1;
	  cur_buf->buf[nseq].lseek = lmark;
	  cur_buf->buf[nseq].cont = ocont;
	  cur_buf->buf[nseq].lib = iln;
	  cur_buf->buf[nseq].frame = itt;
	  strncpy(cur_buf->buf[nseq].libstr,libstr,12);
	  cur_buf->buf[nseq].libstr[12]='\0';
	  cur_buf->buf[nseq].nsfnum = nsfnum;
#ifdef SUPERFAMNUM
	  if ((cur_buf->buf[nseq].sfnum[0]=sfnum[0])>0 &&
	      (cur_buf->buf[nseq].sfnum[1]=sfnum[1])>0 &&
	      (cur_buf->buf[nseq].sfnum[2]=sfnum[2])>0 &&
	      (cur_buf->buf[nseq].sfnum[3]=sfnum[3])>0 &&
	      (cur_buf->buf[nseq].sfnum[4]=sfnum[4])>0 &&
	      (cur_buf->buf[nseq].sfnum[5]=sfnum[5])>0 &&
	      (cur_buf->buf[nseq].sfnum[6]=sfnum[6])>0 &&
	      (cur_buf->buf[nseq].sfnum[7]=sfnum[7])>0 &&
	      (cur_buf->buf[nseq].sfnum[8]=sfnum[8])>0 &&
	      (cur_buf->buf[nseq].sfnum[9]=sfnum[9])>0) ;
#endif
	  nseq++;

	  /* this assumes that max_buf_cnt is guaranteed %6=0 so that
	     additional pointers to the same buffer can be used */
	  cur_buf->buf[nseq].aa1b = cur_buf->buf[nseq-1].aa1b;
	} /* for (itt .. */

	/* make a copy of the overlap */
	if (lcont) {
	  memcpy(aa1s,&aa1[n1-m_msg.loff],m_msg.loff);
	}

	/* if the buffer is filled */
	if (nseq >= max_buf_cnt || ntbuff >= buf_siz - maxn) {
	  /* provide filled buffer to workers */
	  put_rbuf(cur_buf,max_work_buf);

	  /* get an empty buffer to fill */
	  get_rbuf(&cur_buf,max_work_buf);

	  /* "empty" buffers have results that must be processed */
	  if (cur_buf->buf_cnt && cur_buf->have_results)
#ifndef PRSS
	    save_best(cur_buf,pst,nitt1,fdata,m_msg.qsfnum);
#else
	    save_best(cur_buf,pst,nitt1,fdata,m_msg.qsfnum,
		      s_score,&s_n1);
#endif
	  /* now the buffer is truly empty, fill it up */
	  cur_buf->buf_cnt = 0;
	  cur_buf->have_results = 0;
	  /* point the first aa1 ptr to the buffer start */
	  aa1=cur_buf->buf[0].aa1b = cur_buf->start;
	  ntbuff = 0;
	  nseq=0;
	}
	else {	/* room left in current buffer, increment ptrs */
	  aa1=cur_buf->buf[nseq].aa1b = cur_buf->buf[nseq-1].aa1b+n1+1;
	}


      loop1:
	if (lcont) {
	  /* put the copy back */
	  memcpy(aa1,aa1s,m_msg.loff);
	  aa1ptr = &aa1[m_msg.loff];
	  ocont = lcont;
	  maxt = m_msg.maxt3;
	}
	else {
	  ocont = 0;
	  aa1ptr = aa1;
	  maxt = maxn;
	}

#ifdef PROGRESS
#ifdef UNIX
	if (!m_msg.quiet) 
#endif
	  if (nlib % 200 == 199) {
	    fputc('.',stderr);
	    if (nlib % 10000 == 9999) fputc('\n',stderr);
	    else if (nlib % 1000 == 999) fputc(' ',stderr);
	  }
#endif
      }
    }
  }	/* for (iln < nln) */

  put_rbuf_done(max_workers,cur_buf,max_work_buf);

  /* check last buffers for any results */    /***********************/

  for (i=0; i < num_reader_bufs; i++) {
    reader_buf_readp = (reader_buf_readp+1)%(max_work_buf);
    if (reader_buf[reader_buf_readp]->buf_cnt > 0 && 
	reader_buf[reader_buf_readp]->have_results) {
#ifndef PRSS
      save_best(reader_buf[reader_buf_readp],pst,nitt1,fdata,m_msg.qsfnum);
#else
      save_best(reader_buf[reader_buf_readp],pst,nitt1,fdata,m_msg.qsfnum,s_score,&s_n1);
#endif
    }
  }

  /* sort the results and present them */

#if PROGRESS
#ifdef UNIX
  if (!m_msg.quiet) 
#endif
    if (nlib >= 200) fprintf(stderr," Done!\n");
#endif

  m_msg.nbr_seq = nlib;
  get_param(&pst, gstring2, gstring3);

/* *************************** */
/* analyze the last results    */
/* *************************** */

  if (!dohist) {
    if ((nstats < 20 || pst.zsflag <= 0) && pst.zsflag != 3) {
      pst.zsflag = -1;
      m_msg.e_cut = MAXFLOAT;
    }
    else {
      process_hist(stats,nstats,pst);
      for (i=0; i<nbest; i++)
	bptr[i]->zscore = 
	  (*find_zp)(bptr[i]->score[pst.score_ix],bptr[i]->n1,bptr[i]->comp);
      dohist = 1;
    }
  }

  if (!m_msg.noshow) {
    if (pst.zsflag>=0) {
      sortbestz (bptr, nbest);
      if (pst.zdb_size > 1) set_db_size(pst.zdb_size);
      for (i=0; i<nbest; i++)
	bptr[i]->escore=zs_to_E(bptr[i]->zscore,bptr[i]->n1,pst.dnaseq);
      sortbeste(bptr, nbest);
    }
    else sortbest(bptr,nbest,pst.score_ix);

    prhist (stdout,m_msg,pst,gstring2);

    tscan = s_time();
    printf (" Scan time: ");
    ptime(stdout,tscan-tprev);
    printf ("\n");

  l3: 
    if (m_msg.outfile[0]=='\0' && !m_msg.quiet) {
      printf("Enter filename for results: "); fflush(stdout);
    }
    rline[0]='\0';

    if (!m_msg.quiet && fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
    if ((bp=strchr(rline,'\n'))!=NULL) *bp = '\0';
    if (rline[0]!='\0') strncpy(m_msg.outfile,rline,sizeof(m_msg.outfile));
    if (m_msg.outfile[0]!='\0') {
      if ((outfd=fopen(m_msg.outfile,"w"))==0) {
	printf(" could not open %s\n",m_msg.outfile);
	if (!m_msg.quiet) goto l3; else goto l4;
      }

      if (m_msg.markx == 6) {
    	fprintf(outfd,"<html>\n<head>\n<title>%s Results\n",
		    prog_func);
    	fprintf(outfd,"</title>\n</head>\n<body>\n<pre>\n");
      }

      fputs(iprompt0,outfd);
      fprintf(outfd," %s%s\n",verstr,refstr);

      fprintf(outfd," %s%s, %d %s\n vs %s library\n",
	      m_msg.tname, (m_msg.revcomp ? "-" : "\0"), m_msg.n0,
	      m_msg.sqnam, m_msg.ltitle);
      prhist(outfd,m_msg,pst,gstring2);
    }
  
  l4:
    if (m_msg.markx == 6) {
	fprintf(outfd,"</pre>\n<p>\n<hr>\n<p>\n");
      }

    /* before I call showbest/showalign, I need init_work() to get an f_str */

#ifndef PRSS
    init_work(aa0[0],m_msg.n0,&pst,&f_str[0]);
    f_str[5] = f_str[4] = f_str[3] = f_str[2] = f_str[1] =  f_str[0];
    if (m_msg.qframe == 2) {
      if ((aa0[1]=(unsigned char *)calloc((size_t)m_msg.n0+2,
					  sizeof(unsigned char)))==NULL) {
	fprintf(stderr," cannot allocate aa0[1][%d] for alignments\n",
		m_msg.n0+2);
      }
      *aa0[1]='\0';
      aa0[1]++;
      memcpy(aa0[1],aa0[0],m_msg.n0+1);
      revcomp(aa0[1],m_msg.n0);
      init_work(aa0[1],m_msg.n0,&pst,&f_str[1]);
      /*
      for (i=0; i<m_msg.n0; i++) {
	fputc(pst.sq[aa0[1][i]],stderr);
	if (i%60==59) fputc('\n',stderr);
      }
      fputc('\n',stderr);
      */
    }

    /* I also need a "real" aa1 */
    aa1 = buf_list[0].start;

    showbest (stdout, aa0, aa1, maxn, bptr, nbest, qlib, &m_msg, pst,
	      f_str, gstring2);
#else
    showbest (stdout, s_score, s_n1, qlib, &m_msg, pst, gstring2);
#endif

    if (outfd != stdout) {
      t_quiet = m_msg.quiet;
      m_msg.quiet = -1;	/* should guarantee 1..nbest shown */
#ifndef PRSS
      showbest (outfd, aa0, aa1, maxn, bptr, nbest, qlib, &m_msg, pst,
		f_str, gstring2);
#else
    showbest (stdout, s_score, s_n1, qlib, &m_msg, pst, gstring2);
#endif
      m_msg.quiet = t_quiet;
    }

    rline[0]='N';
    if (!m_msg.quiet) {
      printf(" Display alignments also? (y/n) [n] "); fflush(stdout);
      if (fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
    }
    else rline[0]='Y';

    if (toupper(rline[0])=='Y') {
      if (!m_msg.quiet) {
	printf(" number of alignments [%d]? ",m_msg.nshow);
	fflush(stdout);
	if (fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
	if (rline[0]!=0) sscanf(rline,"%d",&m_msg.nshow);
	m_msg.ashow=m_msg.nshow;
      }

      if (m_msg.markx == 5 || m_msg.markx==6 || m_msg.markx==9) {
	fprintf(outfd,"\n>>>%s%s, %d %s vs %s %s\n",
		m_msg.tname,(m_msg.revcomp ? "_rev":"\0"), m_msg.n0,
		m_msg.sqnam,m_msg.lname,lib_label);
      }

      if (m_msg.markx==10) {
	fprintf(outfd,"\n>>>%s%s, %d %s vs %s %s\n",
		m_msg.tname,(m_msg.revcomp ? "-":"\0"), m_msg.n0, m_msg.sqnam,
		m_msg.lname,lib_label);
	fprintf(outfd,"; mp_name: %s\n",argv[0]);
	fprintf(outfd,"; mp_ver: %s\n",mp_verstr);
	fprintf(outfd,"; mp_argv:");
	for (i=0; i<argc; i++)
	  fprintf(outfd," %s",argv[i]);
	fputc('\n',outfd);
	fputs(gstring3,outfd);
	fputs(hstring1,outfd);
      }

#ifndef PRSS
      showalign (outfd, aa0, aa1, maxn, bptr, nbest, qlib, m_msg, pst,
		 f_str, gstring2, &m_msg.lb_current);

      close_work (aa0[0],m_msg.n0, &pst, &f_str[0]);
      if (m_msg.qframe == 2) {
	close_work (aa0[1],m_msg.n0, &pst, &f_str[1]);
      }
#endif
    }

    tddone = time(NULL);
    tdone = s_time();
    fflush(outfd);
  }
    
  if (fdata) {
    fprintf(fdata,"/** %3d%-50s done **/\n",qlib-1,m_msg.libstr);
    fprintf(fdata,"/** %s **/\n",gstring2);
    fflush(fdata);
  }
    
  if (m_msg.markx==10)
    fprintf(outfd,">>><<<\n");

  tdone = s_time();
#ifndef PRSS
  if ( m_msg.markx == 6) fprintf(outfd,"<p><pre>\n"); 
  printsum(outfd);
#endif
  if ( m_msg.markx == 6) fprintf(outfd,"</pre>\n</body></html>\n");

  if (outfd!=stdout) printsum(stdout);
/*  printsum(stderr); */
  exit(0);
  /* done */
  }

void
printsum(fd)
  FILE *fd;
{
  double db_tt;
  char tstr1[26], tstr2[26];

  strncpy(tstr1,ctime(&tdstart),sizeof(tstr1));
  strncpy(tstr2,ctime(&tddone),sizeof(tstr1));
  tstr1[24]=tstr2[24]='\0';

  /* Print timing to output file as well */
  /* Print timing to output file as well */

  if (qtt.carry==0) {
    fprintf(fd, "\n\n%ld residues in %d query   sequences\n", qtt.length, qlib);
  }
  else {
    db_tt = (double)qtt.carry*(double)MAXLONG + (double)qtt.length;
    fprintf(fd, "\n\n%.0g residues in %d query   sequences\n", db_tt, qlib);
  }

  if (ntt.carry==0) {
    fprintf(fd, "%ld residues in %d library sequences\n", ntt.length, nlib);
  }
  else {
    db_tt = (double)ntt.carry*(double)MAXLONG + (double)ntt.length;
    fprintf(fd, "%.6f residues in %d library sequences\n", db_tt, nlib);
  }

  fprintf(fd," Tcomplib (%d proc)[%s]\n start: %s done: %s\n",
	  max_workers,mp_verstr,tstr1,tstr2);
  fprintf(fd," Scan time: ");
  ptime(fd, tscan - tprev);
  fprintf (fd," Display time: ");
  ptime (fd, tdone - tscan);
  fprintf (fd,"\n");
  fprintf (fd, "\nFunction used was %s [%s]\n", prog_func,verstr);
}

#ifndef PRSS
void save_best(struct buf_head *cur_buf, struct pstruct pst, int nitt1,
	       FILE *fdata, int *qsfnum)
#else
void save_best(struct buf_head *cur_buf, struct pstruct pst, int nitt1,
	       FILE *fdata, int *qsfnum, int *s_score, int *s_n1)
#endif

{
  float zscore;
  register int i_score;
  struct buf_str *p_rbuf, *cur_buf_p;
  int i, t_best, t_rbest, tm_best, t_n1, sc_ix;

  sc_ix = pst.score_ix;

  cur_buf_p = cur_buf->buf;
  
  t_best = t_rbest = -1;

  while (cur_buf->buf_cnt--) { /* count down the number of results */
    p_rbuf = cur_buf_p++;	/* step through the results buffer */

    i_score = p_rbuf->rst.score[sc_ix];
    /* need to look for frame 0 if TFASTA, then save stats at frame 6 */
    if (fdata) {
      fprintf(fdata,"%-12s %5d %6d %d %.5lf %.5lf %4d %4d %4d %8ld\n",
	      p_rbuf->libstr,sfn_cmp(qsfnum,p_rbuf->sfnum),
	      p_rbuf->n1,p_rbuf->frame,p_rbuf->rst.comp,p_rbuf->rst.H,
	      p_rbuf->rst.score[0],p_rbuf->rst.score[1],p_rbuf->rst.score[2],
	      p_rbuf->lseek);
    }

#ifdef PRSS
	if (p_rbuf->lseek==0) {
	  s_score[0] = p_rbuf->rst.score[0];
	  s_score[1] = p_rbuf->rst.score[1];
	  s_score[2] = p_rbuf->rst.score[2];
	  *s_n1 = p_rbuf->n1;
	  continue;
	}
#endif

    if (i_score > t_best) {
      tm_best = t_best = i_score;
      t_n1 = p_rbuf->n1;
    }
    if (pst.zsflag >= 10 && p_rbuf->r_score > t_rbest)
      t_rbest = p_rbuf->r_score;

    /* statistics done for best score of set */

    if (nstats < MAXSTATS ) {
      if (p_rbuf->frame == nitt1) {
	stats[nstats].n1 = t_n1;
	stats[nstats].comp = p_rbuf->rst.comp;
	stats[nstats].H = p_rbuf->rst.H;
	if (pst.zsflag >= 10) {
	  tm_best = t_rbest;
	  t_rbest = -1;
	}
	stats[nstats++].score = tm_best;
	t_best = -1;
      }
    }
    else if (!dohist && pst.zsflag >= 0) {
      process_hist(stats,nstats,pst);
      dohist = 1;
      for (i=0; i<nbest; i++)
	  bptr[i]->zscore = 
	    (*find_zp)(bptr[i]->score[pst.score_ix],bptr[i]->n1,bptr[i]->comp);
    }

    /* best saved for every score */
    if (dohist) {

      zscore=(*find_zp)(i_score, p_rbuf->n1,(float)p_rbuf->rst.comp);

      if (p_rbuf->frame == nitt1) {
	addhistz((*find_zp)(t_best, p_rbuf->n1, (float) p_rbuf->rst.comp));
	t_best = t_rbest = -1;
      }
    }
    else zscore = (float) i_score;

    if (zscore > zbestcut) {
      if (nbest >= MAXBEST) {
	bestfull = nbest-MAXBEST/4;
	selectbestz(bptr,bestfull-1,nbest);
	zbestcut = bptr[bestfull-1]->zscore;
	nbest = bestfull;
	/*	fprintf(stderr," zbestcut (%d): %f\n",nlib,zbestcut);*/
      }
      bestptr = bptr[nbest];
      bestptr->score[0] = p_rbuf->rst.score[0];
      bestptr->score[1] = p_rbuf->rst.score[1];
      bestptr->score[2] = p_rbuf->rst.score[2];
      bestptr->comp = (float) p_rbuf->rst.comp;
      bestptr->H = (float) p_rbuf->rst.H;
      bestptr->nsfnum = p_rbuf->nsfnum;
      bestptr->sfnum[0] = p_rbuf->sfnum[0];
#ifdef SUPERFAMNUM
      if (p_rbuf->nsfnum > 0) {
	bestptr->sfnum[1] = p_rbuf->sfnum[1];
	bestptr->sfnum[2] = p_rbuf->sfnum[2];
	bestptr->sfnum[3] = p_rbuf->sfnum[3];
	bestptr->sfnum[4] = p_rbuf->sfnum[4];
	if (p_rbuf->nsfnum > 4) 
	  for (i=5; i <= p_rbuf->nsfnum; i++)
	    bestptr->sfnum[i] = p_rbuf->sfnum[i];
      }
#endif
      bestptr->zscore = zscore;
      bestptr->lseek = p_rbuf->lseek;
      bestptr->lib = p_rbuf->lib;
      bestptr->cont = p_rbuf->cont;
      bestptr->frame = p_rbuf->frame;
      bestptr->n1 = p_rbuf->n1;
      nbest++;
    }
  }
}
