/* copyright (c) 1996, 1997, 1998, 1999 William R. Pearson and the
   U. of Virginia */

/* $Name: fa33t09d0 $ - $Id: complib.c,v 1.20.2.9.2.2 2001/03/29 22:29:43 wrp Exp $ */

/*
 * Concurrent read version
 * 	    Options :	-q		quiet (will not prompt)
 *			-Q		do not show output (default is display)
 *			-b number	show 'number' number of scores (all if $)
 *			-i		invert (reverse complement queries)
 *			-r		results file name
 *
 *			-s file		pam matrix file name
 *
 *	In addition the test sequence file name may be specified as the first
 *	argument and the library file name as the second argument, both without
 *	any -flags.
 *
 *	Feb 20, 1998 modifications for prss3
 *
 *	December, 1998 - DNA searches are now down with forward and reverse
 *			 strands
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#ifdef UNIX
#if !defined(__NetBSD__) && !defined(MacOSX)
#include <values.h>
#else
#include <limits.h>
#define MAXLONG LONG_MAX
#endif
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#else
#define MAXFLOAT 1E+37
#endif

#include <math.h>

#ifdef __MWERKS__
#include <limits.h>
#define MAXLONG LONG_MAX
#include <Types.h>
#include <StandardFile.h>
StandardFileReply freply;
Point wpos;
int tval;
char prompt[256];
#include <sioux.h>
#endif

#include "mw.h"
#include "defs.h"
#include "structs.h"
#include "param.h"

#define XTERNAL
#include "upam.h"
#include "uascii.h"
#undef XTERNAL

char *mp_verstr="33t08";

/********************************/
/* global variable declarations */
/********************************/
char gstring2[MAX_STR];                  /* string for label */
char gstring3[MAX_STR];
char hstring1[MAX_STR];

extern int nsfnum;
extern int sfnum[];
#ifdef SUPERFAMNUM
extern int nsfnum_n;
extern int sfnum_n[];
#endif

#ifdef __MWERKS__
/* short ouvRef, q0vRef, q1vRef; */
FSSpec ouSpec, q0Spec, q1Spec;
OSErr error;
extern int PgmDID;
int ccommand(char ***);	/* MSL */
void InitEvent();	/* checkevent.c */
void ChkEvent();	/* checkevent.c */
#endif

/********************************/
/* extern variable declarations */
/********************************/
extern char *prog_func;		/* function label */
extern char *verstr, *iprompt0, *iprompt1, *iprompt2, *refstr;

/********************************/
/*extern function declarations  */
/********************************/
#ifndef FASTX
extern int getseq(char *filen, unsigned char *seq, int maxs, char *libstr,
		   int *dnaseq,long *sq0ff);
#else
extern int getntseq(char *filen, unsigned char *seq, int maxs, char *libstr,
		   int *dnaseq,long *sq0ff);
#endif

#ifndef PRSS
extern int openlib(int iln, char *lname, int *lb_types, int ldnaseq);
#else
extern int openlib(char *lname, struct mngmsg m_msg);
#endif

extern int scanseq(unsigned char *seq, int n, char *str);
extern void revcomp(unsigned char *seq, int n);

extern int (*getlib) (unsigned char *seq, int maxs,
		      char *libstr,
#ifndef USE_FSEEKO
		      long *libpos,
#else
		      off_t *libpos,
#endif
		      int *lcont);

#define GETLIB (*getlib)

extern void initenv (int argc, char **argv, struct mngmsg *m_msg, 
		     struct pstruct *ppst, unsigned char **aa0, int outtty);
extern void ptime (FILE *fp, time_t time);		/* prints the time */
extern void reseq ();		/* reset file pointer if DNA sequence */

#if defined(TFASTA) || defined(FASTX) || defined(TFASTX)
extern void aainit(int tr_type, int debug);
#endif

/* calculation functions */
extern void init_work(unsigned char *aa0, int n0, 
		      struct pstruct *ppst, void **f_arg);
extern void do_work(unsigned char *aa0, int n0, unsigned char *aa1, int n1,
		    int frame, 
		    struct pstruct *ppst, void *f_str, struct rstruct *rst);

extern void close_work(unsigned char *aa0, int n0,
		       struct pstruct *ppst, void **f_arg);
extern void get_param (struct pstruct *pstr, char *pstring1, char *pstring2);

/* display functions */
#ifndef PRSS
extern void showbest (FILE *fp, unsigned char **aa0, unsigned char *aa1,
		       int maxn, struct beststr **bptr, int nbest,
		       int qlib, struct mngmsg *m_msg,struct pstruct pst,
		       void **f_str, char *gstring2);

#else
extern void showbest (FILE *fp, int *s_score, int maxn, 
		       int qlib, struct mngmsg *m_msg,struct pstruct pst,
		       char *gstring2);
#endif
extern void showalign (FILE *fp, unsigned char **aa0, unsigned char *aa1,
		       int maxn, struct beststr **bptr, int nbest,
		       int qlib, struct mngmsg m_msg,struct pstruct pst,
		       void **f_str, char *gstring2, int *lb_current);

/* statistical functions */
extern void process_hist(struct stat_str *sptr, int nstat, struct pstruct pst);
extern double zs_to_E(float zs, int n1, int isdna);
extern float (*find_zp)(int score, int length, double comp);
extern void irand(int);
extern int shuffle(unsigned char *, unsigned char *, int);
extern int wshuffle(unsigned char *, unsigned char *, int, int, int *);
extern void set_db_size(int);

/* misc functions */
void h_init(struct pstruct *);		/* doinit.c */
void last_init(struct mngmsg *, struct pstruct *); /* initfa/sw.c */
void s_abort(char *, char *);			/* compacc.c */
void revcomp(unsigned char *, int);	/* nxgetaa.c */
#ifndef FASTX
void resetp(unsigned char *, int, struct mngmsg *, struct pstruct *); /* initfa/sw.c */
#else
void reseta(unsigned char *, int, struct mngmsg *, struct pstruct *); /* initfa/sw.c */
#endif
void gettitle(char *, char *, int);	/* nxgetaa.c */
void libchoice(char *lname, int, struct mngmsg *); /* lib_sel.c */
void libselect(char *lname, struct mngmsg *);	/* lib_sel.c */
void query_parm(struct mngmsg *, struct pstruct *); /* initfa/sw.c */
void addhistz(float zscore);		/* scaleswn.c */
void selectbestz(struct beststr **, int, int);
void sortbestz(struct beststr **, int);
void sortbeste(struct beststr **, int);
void sortbest(struct beststr **, int, int);
void prhist(FILE *, struct mngmsg, struct pstruct, char *);
void printsum(FILE *);

FILE *outfd;			/* Output file */
int outtty=1;

extern time_t s_time ();                 /* fetches time for timing */

/* this information is global for fsigint() */
time_t tstart, tscan, tprev, tdone, s_time();	/* Timing */
time_t tdstart, tddone, time();

int  nlib = 0, nclib=0, qlib = 0;	/* number of sequences scanned */
struct total_db_str ntt, qtt;

main (argc, argv)
int argc;
char *argv [];
{
  unsigned char *aa0[6], *aa1, *aa1ptr, *aa1s;
  int n1;
  int t_best, t_rbest;		/* best score of two/six frames */
  char lib_label[MAX_FN];
#ifdef PRSS
  int s_score[3];
  int s_n1;
#endif
  int i_score;
  struct pstruct pst;
  void *f_str[6];
  
  struct mngmsg m_msg;		/* Message from host to manager */
  int iln;			/* index into library names */
  char rline[MAX_FN];
  int t_quiet;
  struct beststr *best,		/* array of best scores */
                 *bestptr,
                 **bptr;	/* array of pointers */
  struct stat_str *stats;	/* array of scores for statistics */
  int nstats;

  struct rstruct  rst;		/* results structure */
  struct rstruct  rrst;		/* results structure for shuffle*/
  int nbest;		/* number of best scores */
  float zbestcut = -BIGNUM;	/* cut off for best z-score */
  int bestfull;			/* index for selectbest() */
  int i;

  FILE *fdata=NULL;		/* file for full results */
  char libstr[21];		/* string for labeling full results */
  
  int leng;		/* leng is length of the descriptive line */
  int maxn = MAXTOT;		/* max size of the query sequence */
  int maxt;
#ifndef USE_FSEEKO
  long lmark;			/* seek into library of current sequence */
#else
  off_t lmark;			/* seek into library of current sequence */
#endif
  int itt, nitt1, lcont, ocont;	/* continued sequence */
  int counr = 0;		/* count of number of query sequences */
  int dohist=0;			/* flag for z-value processing */
  int igncnt=0;			/* count for ignoring sequences warning */
  int ieven=0;			/* tmp for wshuffle */
  float zscore;			/* tmp value */
  FILE *fopen ();		/* file pointer */
  char *bp;			/* general purpose string ptr */
  void fsigint();
  
  /* Initialization */

#ifdef __MWERKS__
  SIOUXSettings.showstatusline=FALSE;
  SIOUXSettings.asktosaveonclose=TRUE;
  SIOUXSettings.autocloseonquit=TRUE;
  
  argc = ccommand(&argv);

/* 
   if (GetResource('DLOG',PgmDID)==(Handle)0 && 
      OpenResFile("\pFASTA.rsrc")<0) {
    SysBeep(100);
    fprintf(stderr," WARNING FASTA.rsrc file could not be found\n");
  }
  wpos.h=50; wpos.v=100;
*/
  InitEvent();
  /* GetVol((unsigned char *)prompt,&ouvRef); */
  error=HGetVol(NULL,&ouSpec.vRefNum, &ouSpec.parID);
  if (error != noErr) {
  	fprintf(stderr," cannot get current directory\n");
  	exit(1);
  	}
  wpos.h=50; wpos.v=100;
#endif

#ifdef UNIX
  outtty=isatty(1);
#else
  outtty=1;
#endif

  h_init( &pst );
  
  ntt.length = qtt.length = 0l;
  ntt.carry = qtt.carry = 0;

  aa0[0] = NULL;
  initenv (argc, argv, &m_msg, &pst,&aa0[0],outtty);

#ifndef PRSS
  if (m_msg.n1_low > 0 && m_msg.n1_high < BIGNUM) 
    sprintf(lib_label,"library (range: %d-%d)",m_msg.n1_low,m_msg.n1_high);
  else if (m_msg.n1_low > 0) 
    sprintf(lib_label,"library (range: >%d)",m_msg.n1_low);
  else if (m_msg.n1_high < BIGNUM)
    sprintf(lib_label,"library (range: <%d)",m_msg.n1_high);
  else
    strncpy(lib_label,"library",sizeof(lib_label));
#else
  sprintf(lib_label,"shuffled sequence");
#endif

  if (m_msg.quiet) {
    outtty = 0;
#ifdef __MWERKS__
    SIOUXSettings.asktosaveonclose=FALSE;
#endif
  }

  tstart = tscan = s_time();
  tdstart = time(NULL);
  
#if defined(FASTX) || defined(TFASTX) 
  /* set up translation tables: faatran.c */
  aainit(pst.tr_type,pst.debug_lib);
#endif

#ifdef TFASTA
  aainit(pst.tr_type,pst.debug_lib);
  pst.dnaseq = -1;	/* force to protein */
  m_msg.ldnaseq = 1;
  if (sqtype[0]=='D') {
    fprintf(stderr," %s compares a protein to a translated\n\
DNA sequence library.  Do not use a DNA query/scoring matrix.\n",prog_func);
    exit(1);
  }
#endif

  /* Allocate space for the query and library sequences */
  if (aa0[0]==NULL) {
    if ((aa0[0] = (unsigned char *)malloc((MAXTOT+1) * sizeof (char))) == NULL)
      s_abort ("Unable to allocate query sequence", "");
    *aa0[0]=0;
    aa0[0]++;
  }
  aa0[5]=aa0[4]=aa0[3]=aa0[2]=aa0[1]=aa0[0];

  /* make room for random sequence */
  if ((aa1s = (unsigned char *)malloc((MAXTOT+1) * sizeof (char))) == NULL) {
      s_abort ("Unable to allocate shuffled library sequence", "");
      *aa1s=0;
      aa1s++;
  }
  irand(0);

  if (m_msg.markx == 6) {
    fprintf(stdout,"<html>\n<head>\n<title>%s Results\n",
	    prog_func);
    fprintf(stdout,"</title>\n</head>\n<body>\n<pre>\n");
  }

  fputs(iprompt0,stdout);
  fprintf(stdout," %s%s\n",verstr,refstr);

  /* Query library */
  if (m_msg.tname[0] == '\0')
    {
      if (m_msg.quiet == 1)
	s_abort("Query sequence undefined","");
#ifndef __MWERKS__           
    l1:	fputs (iprompt1, stdout);
      fflush  (stdout);
      if (fgets (m_msg.tname, MAX_FN, stdin) == NULL)
	s_abort ("Unable to read query library name","");
      if ((bp=strchr(m_msg.tname,'\n'))!=NULL) *bp='\0';
      if (m_msg.tname[0] == '\0') goto l1;
#else
    NIntroDlog(PgmDID,verstr,"\0","\0","\0");	

  l1:	SFileDlog(iprompt1,&freply);
    if (freply.sfGood==TRUE) {
      PtoCstr((unsigned char *)freply.sfFile.name);
      strcpy(m_msg.tname,(char *)freply.sfFile.name);
	  q1Spec.vRefNum = q0Spec.vRefNum = freply.sfFile.vRefNum;
	  q1Spec.parID = q0Spec.parID = freply.sfFile.parID;
	  HSetVol(NULL,q0Spec.vRefNum,q0Spec.parID);
    }
    else exit(0);
#endif
    }
  
  /* Fetch first sequence */
#ifndef FASTX
  m_msg.n0 = getseq (m_msg.tname,aa0[0], MAXTOT, m_msg.libstr,&pst.dnaseq,
		     &m_msg.sq0off);
#else
  m_msg.n0 = getntseq (m_msg.tname,aa0[0], MAXTOT, m_msg.libstr,&pst.dnaseq,
		     &m_msg.sq0off);
#endif

  if (m_msg.n0 > MAXTST) {
    fprintf(stderr," sequence truncated to %d\n %s\n",MAXTST,sqnam);
    fprintf(stdout," sequence truncated to %d\n %s\n",MAXTST,sqnam);
    aa0[0][MAXTST]=EOSEQ;
    m_msg.n0=MAXTST;
  }

  if (m_msg.n0 <= 0)
    s_abort ("Query sequence length <= 0: ", m_msg.tname);

#ifdef SUPERFAMNUM
  m_msg.nqsfnum = nsfnum;
  for (i=0; i <= nsfnum & i<10; i++) m_msg.qsfnum[i] = sfnum[i];
  m_msg.nqsfnum_n = nsfnum_n;
  for (i=0; i <= nsfnum_n & i<10; i++) m_msg.qsfnum_n[i] = sfnum_n[i];
#endif

  pst.n0 = m_msg.n0;

/* a sanity check */
#if !defined(TFASTX) && !defined(TFASTA)
   if (m_msg.revcomp && !(pst.dnaseq==1)) {
     fprintf(stderr," cannot reverse complement %s\n",sqtype);
     m_msg.revcomp = 0;
   }
#endif

#ifndef FASTX
  resetp (aa0[0], m_msg.n0, &m_msg, &pst);
#else
  reseta (aa0[0], m_msg.n0, &m_msg, &pst);
#endif

  gettitle(m_msg.tname,m_msg.qtitle,sizeof(m_msg.qtitle));
  if (strcmp(m_msg.tname,"-")==0) {
    strncmp(m_msg.tname,m_msg.qtitle,sizeof(m_msg.tname));
    if ((bp=strchr(m_msg.tname,' '))!=NULL) *bp='\0';
    if ((bp=strchr(&m_msg.tname[4],'|'))!=NULL) *bp='\0';
  }

  maxn = MAXTOT - m_msg.n0 -2;	/* maxn = max library sequence space */

  if (m_msg.qframe == 2) {
    aa0[1] = aa0[0] + m_msg.n0 + 2;
    *aa0[1] = '\0';
    aa0[1]++;
    maxn -= m_msg.n0+2;
    memcpy(aa0[1],aa0[0],m_msg.n0);
    revcomp(aa0[1],m_msg.n0);
/*
    for (i=0; i<m_msg.n0; i++) {
      fputc(nt[aa0[1][i]],stderr);
      if (i%60==59) fputc('\n',stderr);
    }
    fputc('\n',stderr);
*/
  }
  aa1 = aa0[1] + m_msg.n0+2;

  /* reduce maxn if requested */
  if (m_msg.maxn > 0 && m_msg.maxn < maxn) maxn = m_msg.maxn;

#if !defined(TFASTA) && !defined(TFASTX)
    if (pst.dnaseq==1 && m_msg.n0>(MAXTST+MAXLIB)/3) {
      fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
      exit(1);
    }
    else if (pst.dnaseq==0 && m_msg.n0>(MAXTST+MAXLIB)/2) {
      fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
      exit(1);
    }

    m_msg.loff = m_msg.n0;
    m_msg.maxt3 = maxn-m_msg.loff;
#else        	
    if (sqtype[0]=='D' || pst.dnaseq == 1) {
      fprintf(stderr," tfasta/x/y compares a protein to a translated\n\
DNA sequence library.  Do not use a DNA query/scoring matrix.\n");
      exit(1);
    }
    aainit(pst.tr_type,pst.debug_lib);
    m_msg.ldnaseq = 1;
    if (m_msg.n0 > MAXTST) {
      fprintf(stderr," query sequence is too long %d %s\n",m_msg.n0,sqnam);
      exit(1);
    }

    if (m_msg.n0*6 > maxn ) {	/* n0 for the complement + n0*3 for
                                   the three frames */
      if (m_msg.n0*7+2 < MAXTST+MAXLIB) { /* m_msg0*6 + m_msg0 */
	fprintf(stderr,
		" query sequence too long for library segment: %d - resetting to %d\n",
	      maxn,m_msg.n0*6);
	maxn = m_msg.maxn = m_msg.n0*6;
      }
      else {
	fprintf(stderr," query sequence too long for translated search: %d %s\n",
	      m_msg.n0,sqnam);
	exit(1);
      }
    }

    /* set up some constants for overlaps */
    m_msg.loff = 3*m_msg.n0;
    m_msg.maxt3 = maxn-m_msg.loff-3; m_msg.maxt3 -= m_msg.maxt3%3; m_msg.maxt3++;
    maxn = maxn - 3; maxn -= maxn%3; maxn++;
#endif

#if defined(TFASTA) || defined(TFASTX)
  pst.maxlen = maxn;
#endif
  
#ifndef FASTX
    if (pst.dnaseq==1) m_msg.ldnaseq = 1;
#endif

  counr = 0;
  leng = (int)strlen (m_msg.libstr);
  sprintf (&m_msg.libstr[leng], " %d %s", m_msg.n0, sqnam);
  m_msg.seqnm = counr;
  
  /* Library */

#ifdef __MWERKS__
  HSetVol(NULL,ouSpec.vRefNum,ouSpec.parID);
#endif

#ifndef PRSS
  if (strlen (m_msg.lname) == 0) {
    if (m_msg.quiet == 1) s_abort("Library name undefined","");
    libchoice(m_msg.lname,sizeof(m_msg.lname),&m_msg);
  }
  
  libselect(m_msg.lname, &m_msg);
#else
  if (strlen (m_msg.lname) == 0) {
    if (m_msg.quiet == 1) s_abort("Shuffle sequence undefined","");
l2:    fputs(iprompt2,stdout);
    fflush(stdout);
    if (fgets (m_msg.lname, MAX_FN, stdin) == NULL)
      s_abort ("Unable to read shuffle file name","");
    if ((bp=strchr(m_msg.lname,'\n'))!=NULL) *bp='\0';
    if (m_msg.lname[0] == '\0') goto l2;
  }
  m_msg.lbnames[0]= m_msg.lname;
  m_msg.lb_current = 0;
  m_msg.nln = 1;
  m_msg.nshow = 0;
#endif

  /* Get additional parameters here */
  if (!m_msg.quiet) query_parm (&m_msg, &pst);
  
  last_init(&m_msg, &pst);
#if !defined(TFASTA) && !defined(TFASTX)
  /* for fasta/fastx searches, itt iterates the the query strand */
  nitt1 = m_msg.nitt1 = m_msg.qframe-1;
#else
  /* for tfasta/tfastxy searches, itt iterates the library frames */
  nitt1 = m_msg.nitt1 = m_msg.nframe-1;
#endif
  
  /* Allocate space for saved scores */
  if ((best = 
       (struct beststr *)malloc((MAXBEST+1)*sizeof(struct beststr)))==NULL)
    s_abort ("Cannot allocate best struct","");
  if ((bptr = 
       (struct beststr **)malloc((MAXBEST+1)*sizeof(struct beststr *)))==NULL)
    s_abort ("Cannot allocate bptr","");
  
  /* Initialize bptr */
  for (nbest = 0; nbest < MAXBEST+1; nbest++)
    bptr[nbest] = &best[nbest];
  best++; bptr++;
  best[-1].score[0]=best[-1].score[1]=best[-1].score[2]= MAXINT;
  best[-1].zscore=MAXFLOAT;
  
  if ((stats =
       (struct stat_str *)malloc((MAXSTATS)*sizeof(struct stat_str)))==NULL)
    s_abort ("Cannot allocate stats struct","");
  nstats = 0;

  outfd = stdout;

#ifdef UNIX
  /* set up signals now that input is done */
  signal(SIGHUP,SIG_IGN);
#endif

  /* open the library */

  /* Label the output */
  if ((bp = (char *) strchr (m_msg.lname, ' ')) != NULL) *bp = '\0';
  
  if (m_msg.ltitle[0] == '\0')
    strncpy(m_msg.ltitle,m_msg.lname,sizeof(m_msg.ltitle));

#if !defined(TFASTX) && !defined(TFASTA) && !defined(PRSS)
  printf("%.50s: %d %s%s\n %s\n vs  %.60s %s\n",
	 m_msg.tname, m_msg.n0, m_msg.sqnam,
	 (m_msg.revcomp ? " (reverse complement)" : "\0"),
	 m_msg.qtitle,m_msg.ltitle,lib_label);
#else
#ifndef PRSS
  printf("%.50s: %d %s%s\n %s\n vs  %.60s %s\n",
	 m_msg.tname, m_msg.n0, m_msg.sqnam,
	 (m_msg.revcomp ? " (reverse complement)" : (m_msg.nframe==1)? " (forward-only)" : "\0"),
	 m_msg.qtitle,m_msg.ltitle,lib_label);
#endif
#endif

  tprev = s_time();
  
  if (m_msg.dfile[0] && (fdata=fopen(m_msg.dfile,"w"))!=NULL)
    fprintf(fdata,"%3d\t%-50s\n",m_msg.n0,m_msg.qtitle);

  qtt.length += m_msg.n0;
  qlib++;

  nbest = 0;

  init_work (aa0[0], m_msg.n0, &pst, &f_str[0]);

  f_str[5] = f_str[4] = f_str[3] = f_str[2] = f_str[1] = f_str[0];
  if (m_msg.qframe == 2) {
    init_work ( aa0[1], m_msg.n0, &pst, &f_str[1]);
  }

  for (iln = 0; iln < m_msg.nln; iln++) {
    m_msg.lb_current = iln;
#ifndef PRSS
    if (openlib(iln, m_msg.lbnames[iln], m_msg.lb_types, m_msg.ldnaseq)!=1)
#else
    if (openlib(m_msg.lbnames[0], m_msg)!=1)
#endif
      {
	fprintf(stderr," cannot open library %s\n",m_msg.lbnames[iln]);
	continue;
      }
#ifndef PRSS
    else
      printf ("searching %s %s\n",m_msg.lbnames[iln],lib_label);
#endif

    lcont = 0;
    ocont = 0;
    maxt = maxn;
    aa1ptr = aa1;
    while ((n1=GETLIB(aa1ptr,maxt,libstr,&lmark,&lcont))>=0) {

      if (n1 < m_msg.n1_low || n1 > m_msg.n1_high) goto loop2;

      nlib++;
      ntt.length += n1;
      if (ntt.length > MAXLONG) {ntt.length -= MAXLONG; ntt.carry++;}

      /*    fprintf(stderr," nlib: %d n1: %d\n",nlib,n1); */
      if (aa1!=aa1ptr) {n1 += m_msg.loff; nlib--;}
      if (n1<=1) {
	/*	if (igncnt++ <10)
		fprintf(stderr,"Ignoring: %s\n",libstr);
	*/
	goto loop2;
      }
      /* This finds most reasons for core dumps */
#ifdef DEBUG
      if (pst.debug_lib)
	for (i=0; i<n1; i++)
#ifndef TFASTA
	  if (aa1[i]>=pst.nsqx) 
#else 
          if (aa1[i]>=nnt)
#endif
	      {fprintf(stderr,
		       "%s residue[%d/%d] %d range (%d) lcont/ocont: %d/%d\n%s\n",
		       libstr,i,n1,aa1[i],pst.nsqx,lcont,ocont,aa1ptr+i);
	      aa1[i]=0;
	      n1=i-1;
	      break;
	      }
#endif
#ifdef PROGRESS
#ifdef UNIX
      if (!m_msg.quiet) 
#endif
	if (nlib % 200 == 199) {
	  fputc('.',stderr);
	  if (nlib % 10000 == 9999) fputc('\n',stderr);
	  else if (nlib % 1000 == 999) fputc(' ',stderr);
#ifdef __MWERKS__
	  fflush(stderr);	/* this is a bug */
#endif
	}
#endif

#ifdef __MWERKS__
      ChkEvent();
#endif

      /* t_best and t_rbest are used to save the best score or shuffled
	 score from all the frames */

      t_best = t_rbest = -1;
      for (itt=m_msg.revcomp; itt<=nitt1; itt++) {

	rst.score[0] = rst.score[1] = rst.score[2] = 0;
	do_work (aa0[itt], m_msg.n0, aa1, n1, itt, &pst, f_str[itt], &rst);

	if (rst.score[pst.score_ix] > t_best) {
	  t_best = rst.score[pst.score_ix];
	}

	if (fdata) {
	  fprintf(fdata,"%-12s %5d %6d %d %.5lf %.5lf %4d %4d %4d %8ld\n",
		  libstr,sfn_cmp(m_msg.qsfnum,sfnum),n1,itt,
		  rst.comp,rst.H,
		  rst.score[0],rst.score[1],rst.score[2],lmark);
	  fflush(fdata);
	}

#ifdef PRSS
	if (lmark==0) {
	  s_score[0] = rst.score[0];
	  s_score[1] = rst.score[1];
	  s_score[2] = rst.score[2];
	  s_n1 = n1;
	  continue;
	}
#endif
	if (pst.zsflag >= 10) {
	  if (pst.zs_win > 0) wshuffle(aa1,aa1s,n1,pst.zs_win,&ieven);
	  else shuffle(aa1,aa1s,n1);
#ifdef DEBUG
      /* This finds most reasons for core dumps */
      if (pst.debug_lib)
	for (i=0; i<n1; i++)
	  if (aa1s[i]>=pst.nsq || aa1s[i]<=0) 
	      {fprintf(stderr,
		       "%s residue[%d/%d] %d range (%d)\n",
		       libstr,i,n1,aa1s[i],pst.nsq);
	      aa1s[i]=0;
	      n1=i-1;
	      break;
	      }
#endif
	  do_work (aa0[itt], m_msg.n0, aa1s, n1, itt, &pst, f_str[itt], &rrst);
	  i_score = rrst.score[pst.score_ix];
	  if (i_score > t_rbest) t_rbest = i_score;
	}
	i_score = rst.score[pst.score_ix];

	if (nstats < MAXSTATS) {
	  if (itt==nitt1) {
	    stats[nstats].n1 = n1;
	    stats[nstats].comp =  rst.comp;
	    stats[nstats].H = rst.H;
	    if (pst.zsflag >=10) t_best = t_rbest;
	    stats[nstats++].score = t_best;
	    t_best = t_rbest = -1;	/* reset t_rbest, t_best */
	  }
	}
	else if (!dohist && pst.zsflag > 0) {
	  process_hist(stats,nstats,pst);
	  for (i=0; i<nbest; i++) {
	    bptr[i]->zscore = 
	      (*find_zp)(bptr[i]->score[pst.score_ix],
			 bptr[i]->n1, bptr[i]->comp);
	  }
	  dohist = 1;
	}

	if (dohist) { /* dohist > 0 => nstats >= MAXSTATS */
	  zscore=(*find_zp)(i_score,n1, rst.comp);

	  if (itt == nitt1) {
	    if (pst.zsflag >= 10) t_best = t_rbest;
	    
	    addhistz((*find_zp)(t_best,n1, rst.comp));

	    t_best = t_rbest = -1;
	  }
	}
	else zscore = (float) i_score;

	if (zscore > zbestcut ) {
	  if (nbest >= MAXBEST) {
	    bestfull = nbest-MAXBEST/4;
	    selectbestz(bptr,bestfull-1,nbest);
	    zbestcut = bptr[bestfull-1]->zscore;
	    nbest = bestfull;
	  }
	  bestptr = bptr[nbest];
	  bestptr->score[0] = rst.score[0];
	  bestptr->score[1] = rst.score[1];
	  bestptr->score[2] = rst.score[2];
	  bestptr->comp =  rst.comp;
	  bestptr->H = rst.H;
	  bestptr->zscore = zscore;
	  bestptr->lseek = lmark;
	  bestptr->cont = ocont;
	  bestptr->lib = iln;
	  bestptr->n1 = n1;
	  bestptr->nsfnum = nsfnum;
#ifdef SUPERFAMNUM
	  if ((bestptr->sfnum[0]=sfnum[0])>0 &&
	      (bestptr->sfnum[1]=sfnum[1])>0 &&
	      (bestptr->sfnum[2]=sfnum[2])>0 &&
	      (bestptr->sfnum[3]=sfnum[3])>0 &&
	      (bestptr->sfnum[4]=sfnum[4])>0 &&
	      (bestptr->sfnum[5]=sfnum[5])>0 &&
	      (bestptr->sfnum[6]=sfnum[6])>0 &&
	      (bestptr->sfnum[7]=sfnum[7])>0 &&
	      (bestptr->sfnum[8]=sfnum[8])>0 &&
	      (bestptr->sfnum[9]=sfnum[9])>0) ;
#endif
	  bestptr->frame = itt;
	  nbest++;
	}
      }

    loop2: 
      if (lcont) {
	maxt = m_msg.maxt3;
	memcpy(aa1,&aa1[n1-m_msg.loff],m_msg.loff);
	aa1ptr= &aa1[m_msg.loff];
	ocont = lcont;
      }
      else {
	maxt = maxn;
	aa1ptr=aa1;
	ocont = 0;
      }
    } /* end while((n1=getlib())) */
  } /* end iln=1..nln */

#ifdef PROGRESS
#ifdef UNIX
  if (!m_msg.quiet) 
#endif
    if (nlib >= 200) fprintf(stderr," Done!\n");
/* this is a bug */
#ifdef __MWERKS__
	fflush(stderr);
#endif
#endif

  m_msg.nbr_seq = nlib;
  get_param(&pst, gstring2,gstring3);

/* *************************** */
/* analyze the last results    */
/* *************************** */
    
#ifndef PRSS
  if (!dohist) {
    if ((nstats < 20 || pst.zsflag <= 0) && pst.zsflag != 3) {
      pst.zsflag = -1;
    }
    else {
      process_hist(stats,nstats,pst);
      for (i = 0; i < nbest; i++) {
	  bptr[i]->zscore = (*find_zp)(bptr[i]->score[pst.score_ix],
					bptr[i]->n1, bptr[i]->comp);
      }
    }
  }
#else
  process_hist(stats,nstats,pst);
  for (i = 0; i < nbest; i++) {
      bptr[i]->zscore = (*find_zp)(bptr[i]->score[pst.score_ix],
				   bptr[i]->n1,bptr[i]->comp);
  }
#endif

  if (!m_msg.noshow) {
    if (pst.zsflag >=0 ) {
      sortbestz (bptr, nbest);
      if (pst.zdb_size > 1) set_db_size(pst.zdb_size);
      for (i=0; i<nbest; i++) {
	bptr[i]->escore=zs_to_E(bptr[i]->zscore,bptr[i]->n1,pst.dnaseq);
      }
      sortbeste(bptr, nbest);
    }
    else sortbest(bptr,nbest,pst.score_ix);

#ifdef PRSS
    gettitle(m_msg.lname,m_msg.ltitle,sizeof(m_msg.ltitle));
    printf("%.50s - %s %d %s%s\n vs %.60s - %s shuffled sequence\n",
	   m_msg.tname, m_msg.qtitle,m_msg.n0, m_msg.sqnam,
	   (m_msg.revcomp ? " (reverse complement)" : "\0"),
	   m_msg.lname,m_msg.ltitle);
#endif

    prhist (stdout,m_msg,pst,gstring2);

    tscan = s_time();
    printf (" Scan time: ");
    ptime(stdout,tscan-tprev);
    printf ("\n");

  l3:
    if (!m_msg.quiet) {
      printf("Enter filename for results [%s]: ",m_msg.outfile); fflush(stdout);
    }

    rline[0]='\0';
    if (!m_msg.quiet && fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
    if ((bp=strchr(rline,'\n'))!=NULL) *bp = '\0';
    if (rline[0]!='\0') strncpy(m_msg.outfile,rline,sizeof(m_msg.outfile));
    if (m_msg.outfile[0]!='\0') {
      if ((outfd=fopen(m_msg.outfile,"w"))==NULL) {
	fprintf(stderr," could not open %s\n",m_msg.outfile);
	if (!m_msg.quiet) goto l3; else goto l4;
	}
#ifdef __MWERKS__
       HSetVol(NULL,q0Spec.vRefNum,q0Spec.parID);
#endif
      /* label the output */
      if (m_msg.markx == 6) {
    	fprintf(outfd,"<html>\n<head>\n<title>%s Results\n",
		    prog_func);
    	fprintf(outfd,"</title>\n</head>\n<body>\n<pre>\n");
      }

      fputs(iprompt0,outfd);
      fprintf(outfd," %s%s\n",verstr,refstr);

      fprintf(outfd," %s%s, %d %s\n vs %s %s\n",
	      m_msg.tname, (m_msg.revcomp ? "-" : "\0"), m_msg.n0,
	      m_msg.sqnam, m_msg.ltitle, lib_label);

      prhist(outfd,m_msg,pst,gstring2);
    }

  l4:   
    if (m_msg.markx == 6) {
      fprintf(outfd,"</pre>\n<p>\n<hr>\n<p>\n");
    }

#ifndef PRSS
    showbest (stdout, aa0, aa1, maxn, bptr, nbest, qlib, &m_msg, pst,
	      f_str, gstring2);
#else
    showbest (stdout, s_score, s_n1, qlib, &m_msg, pst, gstring2);
#endif

    if (outfd != stdout) {
      t_quiet = m_msg.quiet;
      m_msg.quiet = -1;	/* should guarantee 1..nbest shown */
#ifndef PRSS
      showbest (outfd, aa0, aa1, maxn, bptr, nbest, qlib, &m_msg, pst,
		  f_str, gstring2);
#else
      showbest (outfd, s_score, s_n1, qlib, &m_msg, pst, gstring2);
#endif
      m_msg.quiet = t_quiet;
    }

    if (m_msg.nshow > 0) {
      rline[0]='N';
      if (!m_msg.quiet){
	printf(" Display alignments also? (y/n) [n] "); fflush(stdout);
	if (fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
      }
      else rline[0]='Y';

      if (toupper((int)rline[0])=='Y') {
	if (!m_msg.quiet) {
	  printf(" number of alignments [%d]? ",m_msg.nshow);
	  fflush(stdout);
	  if (fgets(rline,sizeof(rline),stdin)==NULL) exit(0);
	  if (rline[0]!=0) sscanf(rline,"%d",&m_msg.nshow);
	  m_msg.ashow=m_msg.nshow;
	}

	if (m_msg.markx == 5 || m_msg.markx==6 || m_msg.markx==9) {
	  fprintf(outfd,"\n>>>%s%s, %d %s vs %s library\n",
		  m_msg.tname,(m_msg.revcomp ? "_rev":"\0"), m_msg.n0,
		  m_msg.sqnam,m_msg.lname);
	}

	if (m_msg.markx==10) {
	  fprintf(outfd,"\n>>>%s%s, %d %s vs %s library\n",
		  m_msg.tname,(m_msg.revcomp ? "-":"\0"), m_msg.n0, m_msg.sqnam,
		  m_msg.lname);
	  fprintf(outfd,"; mp_name: %s\n",argv[0]);
	  fprintf(outfd,"; mp_ver: %s\n",mp_verstr);
	  fprintf(outfd,"; mp_argv:");
	  for (i=0; i<argc; i++)
	    fprintf(outfd," %s",argv[i]);
	  fputc('\n',outfd);
	  fputs(gstring3,outfd);
	  fputs(hstring1,outfd);
	}

	showalign (outfd, aa0, aa1, maxn, bptr, nbest, qlib, m_msg, pst,
		   f_str, gstring2, &m_msg.lb_current);

	fflush(outfd);
      }
    }

    if (f_str[1]!=f_str[0]) 
      close_work (aa0[1], m_msg.n0, &pst, &f_str[1]);
    close_work (aa0[0], m_msg.n0, &pst, &f_str[0]);

    tddone = time(NULL);
    tdone = s_time();
    fflush(outfd);
  }
    
  if (fdata) {
    fprintf(fdata,"/** %s **/\n",gstring2);
    fprintf(fdata,"%3d%-50s\n",qlib-1,m_msg.libstr);
    fflush(fdata);
  }
    
  if (m_msg.markx==10)
    fprintf(outfd,">>><<<\n");

  tdone = s_time();
#ifndef PRSS
  if ( m_msg.markx == 6) fprintf(outfd,"<p><pre>\n"); 
  printsum(outfd);
  if ( m_msg.markx == 6) fprintf(outfd,"</pre>\n");
#endif
  if (m_msg.markx == 6) fprintf(outfd,"</body></html>\n");
  if (outfd!=stdout) printsum(stdout);
/*  printsum(stderr); */
  exit(0);
}   /* End of main program */

void
printsum(fd)
  FILE *fd;
{
  double db_tt;
  char tstr1[26], tstr2[26];

  strncpy(tstr1,ctime(&tdstart),sizeof(tstr1));
  strncpy(tstr2,ctime(&tddone),sizeof(tstr1));
  tstr1[24]=tstr2[24]='\0';

  /* Print timing to output file as well */
  fprintf(fd, "\n\n%ld residues in %d query   sequences\n", qtt.length, qlib);
  if (ntt.carry == 0) 
    fprintf(fd, "%ld residues in %d library sequences\n", ntt.length, nlib);
  else {
    db_tt = (double)ntt.carry*(double)MAXLONG + (double)ntt.length;
    fprintf(fd, "%.0f residues in %d library sequences\n", db_tt, nlib);
  }

  fprintf(fd," Scomplib [%s]\n start: %s done: %s\n",mp_verstr,tstr1,tstr2);
  fprintf(fd," Scan time: ");
  ptime(fd, tscan - tprev);
  fprintf (fd," Display time: ");
  ptime (fd, tdone - tscan);
  fprintf (fd,"\n");
  fprintf (fd, "\nFunction used was %s [%s]\n", prog_func,verstr);
}

void fsigint()
{
  tdone = s_time();
  tddone = time(NULL);

  printf(" /*** interrupted ***/\n");
  if (outfd!=stdout) fprintf(outfd,"/*** interrupted ***/\n");
  fprintf(stderr,"/*** interrupted ***/\n");

  printsum(stdout);
  if (outfd!=stdout) printsum(outfd);

  exit(1);
}
