/*	initfa.c	*/

/* $Name: fa33t09d0 $ - $Id: initfa.c,v 1.17.2.13.2.1 2001/05/18 20:50:11 wrp Exp $ */

/* copyright (c) 1996, 1997, 1998  William R. Pearson and the U. of Virginia */

/* init??.c files provide function specific initializations */

/* h_init()	- called from comp_lib.c, comp_thr.c to initialize pstruct ppst
   		  which includes the alphabet, and pam matrix

   alloc_pam()	- allocate pam matrix space
   initpam2()	- convert from 1D to 2D pam

   f_initenv()	- set up mngmsg and pstruct defaults
   f_getopt()	- read fasta specific command line options
   f_getarg()	- read ktup

   resetp()	- reset the parameters, scoring matrix for DNA-DNA
   reseta()	- reset the parameters, scoring matrix, for prot-DNA

   query_parm()	- ask for ktup
   last_init()	- some things must be done last

   f_initpam()	- set some parameters based on the pam matrix
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "defs.h"
#include "param.h"

#include "structs.h"

#define XTERNAL
#include "upam.h"
#include "uascii.h"
#undef XTERNAL

#define MAXWINDOW 32

#ifdef FASTA
char *prog_func = "FASTA";
char *iprompt0=" FASTA searches a protein or DNA sequence data bank\n";
int PgmDID=401;
#endif
#ifdef TFASTX
char *prog_func = "TFASTXY";
char *iprompt0=" TFASTXY compares a protein to a translated DNA data bank\n";
int PgmDID=406;
#endif
#ifdef FASTX
char *prog_func = "FASTXY";
char *iprompt0=" FASTXY compares a DNA sequence to a protein sequence data bank\n";
int PgmDID=405;
#endif
#ifdef FASTF
char *prog_func = "FASTF";
char *iprompt0=" FASTF compares mixed peptides to a protein databank\n";
int PgmDID=400;
#endif
#ifdef TFASTF
char *prog_func = "TFASTF";
char *iprompt0=" TFASTF compares mixed peptides to a translated DNA data bank\n";
int PgmDID=400;
#endif
#ifdef TFASTA
char *prog_func = "TFASTA";
char *iprompt0=" TFASTA translates and searches a DNA sequence data bank\n";
int PgmDID=402;
#endif
#ifdef FASTS
char *prog_func = "FASTS";
char *iprompt0=" FASTS compares linked peptides to a protein databank\n";
int PgmDID=400;
#endif
#ifdef TFASTS
char *prog_func = "TFASTS";
char *iprompt0=" TFASTS compares linked peptides to a protein databank\n";
int PgmDID=400;
#endif

char *iprompt1=" test sequence file name: ";
char *iprompt2=" database file name: ";
#if defined(FASTX) || defined(TFASTX)
char *refstr="\nPlease cite:\n Pearson et al, Genomics (1997) 46:24-36\n";
#else
char *refstr="\nPlease cite:\n W.R. Pearson & D.J. Lipman PNAS (1988) 85:2444-2448\n";
#endif

char *verstr="version 3.3t09 May 18, 2001";

char   *s_optstr = "13Ac:E:f:F:g:h:j:oy:";

static int mktup=2;

extern void s_abort(char *, char *);
extern void init_ascii(char *sq, int nsq, int *sascii, int dnaseq);

/* Sets defaults assuming a protein sequence */
void h_init (struct pstruct *ppst )
{
  int i;

#if defined(FASTF) || defined(TFASTF) || defined(FASTS) || defined(TFASTS)
  aascii[','] = ESS;
#if defined(FASTF) || defined(FASTS)
  pam = a_md20;
  strncpy(ppst->pamfile,"MD20",sizeof(ppst->pamfile));
#else		/* TFASTF */
  pam = a_md10;
  strncpy(ppst->pamfile,"MD10",sizeof(ppst->pamfile));
#endif
#else
  pam = abl50;
  strncpy(ppst->pamfile,"BL50",sizeof(ppst->pamfile));
#endif  

  ppst->nsq = naa;
  ppst->nsqx = naax;
  for (i=0; i<=ppst->nsqx; i++) {
    ppst->sq[i] = aa[i];
    ppst->hsq[i] = haa[i];
    ppst->sqx[i]=aax[i];	/* sq = aa */
    ppst->hsqx[i]=haax[i];	/* hsq = haa */
  }
  sascii = aascii;
  init_ascii(ppst->sq,ppst->nsq,sascii,0);
}

/*
 * alloc_pam(): allocates memory for the 2D pam matrix as well
 * as for the integer array used to transmit the pam matrix
 */
void
alloc_pam (int d1, int d2, struct pstruct *ppst)
{
  int     i, *d2p;
  char err_str[128];

  if ((ppst->pam2[0] = (int **) malloc (d1 * sizeof (int *))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
  }

  if ((ppst->pam2[1] = (int **) malloc (d1 * sizeof (int *))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
  }

  if ((pam12 = (int *) malloc (d1 * d2 * sizeof (int))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
   }

   if ((pam12x = (int *) malloc (d1 * d2 * sizeof (int))) == NULL) {
     sprintf(err_str,"Cannot allocate 2d pam matrix: %d",d2);
     s_abort (err_str,"");
   }

   for (i = 0, d2p = pam12; i < d1; i++, d2p += d2)
      ppst->pam2[0][i] = d2p;

   for (i = 0, d2p = pam12x; i < d1; i++, d2p += d2)
      ppst->pam2[1][i] = d2p;
}

/*
 *  initpam2(struct pstruct pst): Converts 1-D pam matrix to 2-D
 */
void    initpam2 (struct pstruct *ppst)
{
   int     i, j, k, nsq, nsqx, pam_x, sa_x, sa_star;


   nsq = ppst->nsq;
   sa_x = sascii['X'];
   sa_star = sascii['*'];
   ppst->pam2[0][0][0] = -BIGNUM;
   ppst->pam_h = -1; ppst->pam_l = 1;

   k = 0;
   for (i = 1; i <= nsq; i++) {
     ppst->pam2[0][0][i] = ppst->pam2[0][i][0] = -BIGNUM;
     for (j = 1; j <= i; j++) {
       ppst->pam2[0][j][i] = ppst->pam2[0][i][j] = pam[k++] - ppst->pamoff;
       if (ppst->pam_l > ppst->pam2[0][i][j]) ppst->pam_l =ppst->pam2[0][i][j];
       if (ppst->pam_h < ppst->pam2[0][i][j]) ppst->pam_h =ppst->pam2[0][i][j];
     }
   }
   if (ppst->pam_x > 0) ppst->pam_x = ppst->pam2[0][sa_x][sa_x];
   else { ppst->pam2[0][sa_x][sa_x]=ppst->pam_x;
     for (i=1; i<=nsq; i++)
       ppst->pam2[0][sa_x][i] = ppst->pam2[0][i][sa_x]=ppst->pam_x;
       ppst->pam2[0][sa_star][i] = ppst->pam2[0][i][sa_star]=ppst->pam_x;
   }
   pam_x = ppst->pam_x;

   if (ppst->ext_sq_set) {	/* using extended alphabet */
     /* fill in pam2[1] matrix */
     ppst->pam2[1][0][0] = -BIGNUM;
     /* fill in additional parts of the matrix */
     for (i = 1; i <= nsq; i++) {
       ppst->pam2[0][0][i+nsq] = ppst->pam2[0][i+nsq][0] = 
       ppst->pam2[1][0][i+nsq] = ppst->pam2[1][i+nsq][0] = 
	 ppst->pam2[1][0][i] = ppst->pam2[1][i][0] = -BIGNUM;
       for (j = 1; j <= nsq; j++) {
	 ppst->pam2[0][i+nsq][j] = ppst->pam2[0][i][j+nsq] =
	   ppst->pam2[0][i+nsq][j+nsq] = ppst->pam2[1][i][j] =
	   ppst->pam2[0][i][j];

	 ppst->pam2[1][i+nsq][j] = ppst->pam2[1][i][j+nsq] = 
	   ppst->pam2[1][i+nsq][j+nsq] = pam_x;
       }
     }
   }
}

void
f_initenv (m_msg, ppst, aa0)
struct mngmsg *m_msg;
struct pstruct *ppst;
char **aa0;
{
/*  function specific initializations */

   m_msg->stages = 2;

#ifdef FASTA	/* FASTA */
   strncpy (m_msg->f_id0,"fa",3);
   strncpy (m_msg->f_id1,"sw",3);
#endif
#ifdef FASTX		/* FASTX */
   strncpy(m_msg->f_id0,"fx",3);
   strncpy(m_msg->f_id1,"sx",3);
#endif
#ifdef TFASTX		/* TFASTX */
   strncpy (m_msg->f_id0,"tx",3);
   strncpy (m_msg->f_id1,"sx",3);
#endif
#ifdef TFASTA		/* TFASTA */
   strncpy (m_msg->f_id0,"tf",3);
   strncpy (m_msg->f_id1,"tf",3);
#endif
#ifdef TFASTF		/* TFASTF */
   strncpy (m_msg->f_id0,"ft",3);
   strncpy (m_msg->f_id1,"ft",3);
#endif
#ifdef FASTF		/* FASTF */
   strncpy (m_msg->f_id0,"ff",3);
   strncpy (m_msg->f_id1,"ff",3);
#endif
#ifdef FASTS		/* FASTF */
   strncpy (m_msg->f_id0,"fs",3);
   strncpy (m_msg->f_id1,"fs",3);
#endif
#ifdef TFASTS		/* FASTF */
   strncpy (m_msg->f_id0,"ts",3);
   strncpy (m_msg->f_id1,"ts",3);
#endif

   strncpy (m_msg->alab[0],"initn",20);
   strncpy (m_msg->alab[1],"init1",20);
   strncpy (m_msg->alab[2],"opt",20);

   ppst->score_ix = 0;
   ppst->histint = 2;
   m_msg->qframe = 1;
#if !defined(TFASTA) && !defined(TFASTX) && !defined(TFASTF) && !defined(TFASTS)
#ifndef GAP_OPEN
   ppst->gdelval = -12;
#else
   ppst->gdelval = -10;
#endif
   ppst->ggapval = -2;
   ppst->sw_flag = 1;
   m_msg->nframe = -1;
#else
#ifndef GAP_OPEN
   ppst->gdelval = -16;
#else
   ppst->gdelval = -14;
#endif
   ppst->ggapval = -2;
#ifdef TFASTX
   m_msg->nframe = 2;
   ppst->sw_flag = 1;
#else
   m_msg->nframe = 6;
   ppst->sw_flag = 0;
#endif
#endif
   m_msg->nrelv = 3;		/* number of relevant scores */
#if !defined(FASTF) && !defined(TFASTF) && !defined(FASTS) && !defined(TFASTS)
   m_msg->srelv = 1;		/* relevant scores in showbest */
   strncpy (m_msg->label, " opt", sizeof(m_msg->label));
#else
   m_msg->srelv = 2;
   strncpy (m_msg->label, "initn init1", sizeof(m_msg->label));
#endif
   m_msg->arelv = 3;		/* number of relevant scores */

/* see param.h for the definition of all these */

   mktup = 2;
   ppst->param_u.fa.bestscale = 300;
   ppst->param_u.fa.bestoff = 36;
   ppst->param_u.fa.bkfact = 6;
   ppst->param_u.fa.scfact = 3;
   ppst->param_u.fa.bktup = 2;
   ppst->param_u.fa.ktup = 0;
   ppst->param_u.fa.bestmax = 50;
   ppst->param_u.fa.pamfact = 1;
   ppst->param_u.fa.altflag = 0;
   ppst->param_u.fa.optflag = 1;
   ppst->param_u.fa.iniflag = 0;
   ppst->param_u.fa.optcut = 0;
   ppst->param_u.fa.optcut_set = 0;
   ppst->param_u.fa.cgap = 0;
   ppst->param_u.fa.optwid = MAXWINDOW;
   
   alloc_pam (MAXSQ, MAXSQ, ppst);
}

/*  switches for fasta only */

static int gap_set=0;
static int del_set=0;
static int shift_set=0;
static int subs_set=0;
static int e_cut_set=0;
static int sw_flag_set=0;
static int nframe_set=0;
static int wid_set=0;

void
f_getopt (copt, optarg, m_msg, ppst)
char    copt;
char   *optarg;
struct mngmsg *m_msg;
struct pstruct *ppst;
{
  switch (copt) {
  case '1':
     ppst->param_u.fa.iniflag=1;
     break;
  case '3':
    nframe_set = 1;
#ifdef TFASTA
    m_msg->nframe = 3; break;
#else
    m_msg->nframe = 1;	/* for TFASTXY */
    m_msg->qframe = 1;  /* for FASTA, FASTX */
    break;
#endif
  case 'A':
    ppst->sw_flag= 1;
    sw_flag_set = 1;
    break;
  case 'c':
    sscanf (optarg, "%d", &ppst->param_u.fa.optcut);
    ppst->param_u.fa.optcut_set = 1;
    break;
  case 'E':
    sscanf(optarg,"%lf",&m_msg->e_cut);
    e_cut_set = 1;
    break;
  case 'F':
    sscanf(optarg,"%lg",&m_msg->e_low);
    e_cut_set = 1;
    break;
  case 'f':
    sscanf (optarg, "%d", &ppst->gdelval);
    if (ppst->gdelval > 0) ppst->gdelval = -ppst->gdelval;
    del_set = 1;
    break;
  case 'g':
    sscanf (optarg, "%d", &ppst->ggapval);
    if (ppst->ggapval > 0) ppst->ggapval = -ppst->ggapval;
    gap_set = 1;
    break;
  case 'h':
    sscanf (optarg, "%d", &ppst->gshift);
    if (ppst->gshift > 0) ppst->gshift = -ppst->gshift;
    shift_set = 1;
    break;
  case 'j':
    sscanf (optarg, "%d", &ppst->gsubs);
    subs_set = 1;
    break;
  case 'o':
    ppst->param_u.fa.optflag = 0;
    m_msg->nrelv = 2;
    break;
  case 'y':
    sscanf (optarg, "%d", &ppst->param_u.fa.optwid);
    wid_set = 1;
    break;
  }
}

void
f_getarg (argc, argv, optind, m_msg, ppst)
int     argc;
char  **argv;
int     optind;
struct mngmsg *m_msg;
struct pstruct *ppst;
{
   if (argc - optind == 4)
      sscanf (argv[optind + 3], "%d", &ppst->param_u.fa.ktup);
   else
      ppst->param_u.fa.ktup = -ppst->param_u.fa.bktup;
}

void
resetp (unsigned char *aa0, int n0,
	struct mngmsg *m_msg, struct pstruct *ppst)
{
  int i;

  if (ppst->dnaseq == 1) {
    if (!wid_set) ppst->param_u.fa.optwid = 16;
    ppst->histint = 4;
#ifndef GAP_OPEN
    if (!del_set) ppst->gdelval = -16;	/* def. del penalty */
#else
    if (!del_set) ppst->gdelval = -12;	/* def. del penalty */
#endif
    if (!gap_set) ppst->ggapval = -4;	/* def. gap penalty */
    if (!nframe_set) {
      m_msg->qframe = 2;		/* number of query seq directions */
      m_msg->nframe = 1;		/* use frames 0, 1 */
    }
    else m_msg->qframe = m_msg->revcomp+1;

    /* these parameters are used to scale optcut, they should be replaced
       by statistically based parameters */
    ppst->param_u.fa.bestscale = 80;
    ppst->param_u.fa.bkfact = 5;
    ppst->param_u.fa.scfact = 1;
    ppst->param_u.fa.bktup = 6;
    ppst->param_u.fa.bestmax = 80;
    ppst->param_u.fa.bestoff = 45;

    /* largest ktup */
    mktup = 6;

    /* use Ecut = 2.0 for DNA */
    if (!e_cut_set) m_msg->e_cut=2.0;

    ppst->nsq = nnt;
    ppst->nsqx = nntx;
    for (i=0; i<=ppst->nsqx; i++) {
      ppst->hsq[i] = hnt[i];
      ppst->sq[i] = nt[i];
      ppst->hsqx[i] = hntx[i];
      ppst->sqx[i] = ntx[i];
    }

    if (ppst->p_d_set)
      mk_n_pam(npam,nnt,ppst->p_d_mat,ppst->p_d_mis);
    else if (strncmp(ppst->pamfile,"BL50",4)==0) {
      strncpy (ppst->pamfile, "+5/-4", sizeof(ppst->pamfile));
    }

    /* this may not be necessary, as init_ascii fills out sascii */
    sascii = nascii;
    pam = npam;	    /* required, broken in pre-March 3 fa33t08 */

    if (ppst->ext_sq_set) init_ascii(ppst->sqx,ppst->nsqx,sascii,1);
    else init_ascii(ppst->sq,ppst->nsq,sascii,1);

    strncpy (m_msg->sqnam, "nt",sizeof(m_msg->sqnam));
    strncpy (m_msg->sqtype, "DNA",sizeof(m_msg->sqtype));
    if (ppst->param_u.fa.pamfact >= 0)
      ppst->param_u.fa.pamfact = 0;
    if (ppst->param_u.fa.ktup < 0)
      ppst->param_u.fa.ktup = -ppst->param_u.fa.bktup;
    if (!sw_flag_set) ppst->sw_flag = 0;
  }

#if defined(TFASTX)
#ifndef GAP_OPEN
  if (!del_set) ppst->gdelval = -15;
#else
  if (!del_set) ppst->gdelval = -13;
#endif
  if (!gap_set) ppst->ggapval = -2;
  if (!shift_set) ppst->gshift = -20;
  if (!subs_set) ppst->gsubs = -20;
  if (!e_cut_set) m_msg->e_cut=5.0;
#endif
#if defined(TFASTF)
  if (!e_cut_set) m_msg->e_cut=5.0;
#endif
}

/* reseta is called for FASTX/Y to reset parameters for translated DNA
   comparison*/
void
reseta (unsigned char *aa0, int n0,
	struct mngmsg *m_msg, struct pstruct *ppst)
{

  ppst->dnaseq = 0;
  if (!nframe_set) {
    m_msg->qframe = 2;
    m_msg->nframe = 1;
  }
  if (!wid_set) {
    if (ppst->param_u.fa.ktup==1) ppst->param_u.fa.optwid = 32;
    else ppst->param_u.fa.optwid = 16;
  }
#ifndef GAP_OPEN
  if (!del_set) ppst->gdelval = -15;
#else
  if (!del_set) ppst->gdelval = -13;
#endif
  if (!gap_set) ppst->ggapval = -2;
  if (!shift_set) ppst->gshift = -20;
  if (!subs_set) ppst->gsubs = -20;
  if (!e_cut_set) m_msg->e_cut=5.0;

  /* in case someone uses the -n option */
  ppst->dnaseq=0;
  sascii = aascii;

  strncpy (m_msg->sqnam, "aa",sizeof(m_msg->sqnam));
  strncpy (m_msg->sqtype, "protein",sizeof(m_msg->sqtype));
}

/* query_parm()	this function asks for any additional parameters
	that have not been provided.  Could be null. */

void
query_parm (m_msp, ppst)
struct mngmsg *m_msp;
struct pstruct *ppst;
{
   char    qline[40];

   if (ppst->param_u.fa.ktup < 0)
      ppst->param_u.fa.ktup = -ppst->param_u.fa.ktup;

   if (ppst->param_u.fa.ktup == 0) {
      printf (" ktup? (1 to %d) [%d] ", mktup, ppst->param_u.fa.bktup);
      if (fgets (qline, sizeof(qline), stdin) == NULL) exit (0);
      else sscanf(qline,"%d",&ppst->param_u.fa.ktup);
   }

   if (ppst->param_u.fa.ktup == 0)
     ppst->param_u.fa.ktup = ppst->param_u.fa.bktup;

}

void
last_init (m_msg, ppst)
struct mngmsg *m_msg;
struct pstruct *ppst;
{

   if (ppst->param_u.fa.ktup < 0)
      ppst->param_u.fa.ktup = -ppst->param_u.fa.ktup;

   if (ppst->param_u.fa.ktup < 1 || ppst->param_u.fa.ktup > mktup) {
      fprintf(stderr," warning ktup = %d out of range [1..%d], reset to %d\n",
	      ppst->param_u.fa.ktup, mktup, ppst->param_u.fa.bktup);
      ppst->param_u.fa.ktup = ppst->param_u.fa.bktup;
   }

#ifndef TFASTA
   if (!ppst->sw_flag) strncpy(m_msg->f_id1,"fa",3);
#else
   m_msg->revcomp *= 3;
   if (m_msg->nframe == 3) m_msg->nframe += m_msg->revcomp;
#endif

#ifdef TFASTX
   if (m_msg->nframe == 1) m_msg->nframe += m_msg->revcomp;
#endif

   if (ppst->param_u.fa.ktup>=2 && !wid_set) {
     ppst->param_u.fa.optwid=16;
#if !defined(TFASTA) && !defined(TFASTX) && !defined(TFASTF)
#ifndef FASTX
     m_msg->thr_fact = 16;
#else
     m_msg->thr_fact = 8;
#endif
#else
     m_msg->thr_fact = 4;
#endif
   }
   else m_msg->thr_fact = 4;

   if (ppst->param_u.fa.iniflag) {
     ppst->score_ix = 1;
     strncpy (m_msg->label, "initn init1", sizeof(m_msg->label));
   }
   else if (ppst->param_u.fa.optflag) {
     ppst->score_ix = 2;
     m_msg->stages = 1;
   }

#if defined(FASTF) || defined(TFASTF) || defined(FASTS) || defined(TFASTS)
   sascii[','] = aascii[',']= NA;
#endif

   initpam2(ppst);

   /* once we have a complete pam matrix, we can calculate Lambda and K 
      for "average" sequences */
   ppst->kar_p = NULL;
   init_karlin_a(ppst, ppst->aa0_f, &ppst->kar_p);
   do_karlin_a(ppst->pam2[0], ppst,ppst->aa0_f, 
	       ppst->kar_p, &ppst->Lambda, &ppst->K, &ppst->H);

#if defined(FASTF) || defined(TFASTF) || defined(FASTS) || defined(TFASTS)
   if (ppst->ext_sq_set) {
     fprintf(stderr," -S not available on [t]fast[fs]\n");
     ppst->ext_sq_set = 0;
     init_ascii(ppst->sq,ppst->nsq,sascii,0);
   }
#endif
}


void
f_initpam (line, ppst)
char   *line;
struct pstruct *ppst;
{
   if (sscanf (line, " %d %d %d %d %d %d %d", &ppst->param_u.fa.scfact,
	       &ppst->param_u.fa.bestoff, &ppst->param_u.fa.bestscale,
	       &ppst->param_u.fa.bkfact, &ppst->param_u.fa.bktup,
	       &ppst->param_u.fa.bestmax, &ppst->histint) != 7)
   {
      printf ("  bestcut parameters - bad format\n");
      exit (1);
   }
}
