/*	initsw.c	general and function-specific initializations */

/* $Name: fa33t09d0 $ - $Id: initsw.c,v 1.18.2.11.2.1 2001/05/18 20:50:11 wrp Exp $ */

/* copyright (c) 1996, 1997, 1998  William R. Pearson and the U. of Virginia */

/* init??.c files provide function specific initializations */

/* h_init()	- called from comp_lib.c, comp_thr.c to initialize pstruct ppst
   		  which includes the alphabet, and pam matrix

   alloc_pam()	- allocate pam matrix space
   initpam2()	- convert from 1D to 2D pam

   f_initenv()	- set up mngmsg and pstruct defaults
   f_getopt()	- read fasta specific command line options
   f_getarg()	- does nothing for SW - gets # of shuffles for prss3

   resetp()	- reset the parameters, scoring matrix for DNA-DNA
   reseta()	- nothing

   query_parm()	- does nothing for Smith-Waterman, gets number of shuffles,
   		  window shuffle for prss3.

   last_init()	- some things must be done last

   f_initpam()	- set some parameters based on the pam matrix

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef UNIX
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "defs.h"
#include "param.h"

#include "structs.h"

#define XTERNAL
#include "upam.h"
#include "uascii.h"
#undef XTERNAL

char *iprompt1=" query sequence file name: ";
#ifndef PRSS
char *refstr="\nPlease cite:\n T. F. Smith and M. S. Waterman, (1981) J. Mol. Biol. 147:195-197; \n W.R. Pearson (1991) Genomics 11:635-650\n";
char *iprompt0=" SSEARCH searches a sequence database\n\
 using the Smith-Waterman algorithm\n";
char *iprompt2=" database file name: ";
int PgmDID=404;
#else
char *refstr="\nPlease cite:\n W.R. Pearson (1996) Meth. Enzymol. 266:227-258\n";
char *iprompt0=" PRSS compares a query sequence to shuffled sequences\n\
 using the Smith-Waterman algorithm\n";
char *iprompt2=" shuffle file name: ";
int PgmDID=400;
#endif

char *verstr="version 3.3t09 May 18, 2001";

char *prog_func = "Smith-Waterman";
char *s_optstr = "3E:F:f:g:";

extern void init_ascii(char *sq, int nsq, int *sascii, int dnaseq);


/* Sets defaults assuming a protein sequence */
void h_init (struct pstruct *ppst)
{
  int i;

  pam = abl50;
  strncpy(ppst->pamfile,"BL50",sizeof(ppst->pamfile));

  ppst->nsq = naa;
  ppst->nsqx = naax;
  for (i=0; i<=ppst->nsqx; i++) {
    ppst->sq[i]=aa[i];		/* sq = aa */
    ppst->hsq[i]=haa[i];	/* hsq = haa */
    ppst->sqx[i]=aax[i];	/* sq = aax */
    ppst->hsqx[i]=haax[i];	/* hsq = haax */
  }
  sascii = aascii;
  init_ascii(ppst->sq,ppst->nsq,sascii,0);
}

/*
 * alloc_pam(): allocates memory for the 2D pam matrix as well
 * as for the integer array used to transmit the pam matrix
 */
void    alloc_pam (int d1, int d2, struct pstruct *ppst)
{
  int     i, *d2p;
  char err_str[128];

  if ((ppst->pam2[0] = (int **) malloc (d1 * sizeof (int *))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
  }

  if ((ppst->pam2[1] = (int **) malloc (d1 * sizeof (int *))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
  }

  if ((pam12 = (int *) malloc (d1 * d2 * sizeof (int))) == NULL) {
     sprintf(err_str,"Cannot allocate 2D pam matrix: %d",d1);
     s_abort (err_str,"");
   }

   if ((pam12x = (int *) malloc (d1 * d2 * sizeof (int))) == NULL) {
     sprintf(err_str,"Cannot allocate 2d pam matrix: %d",d2);
     s_abort (err_str,"");
   }

   for (i = 0, d2p = pam12; i < d1; i++, d2p += d2)
      ppst->pam2[0][i] = d2p;

   for (i = 0, d2p = pam12x; i < d1; i++, d2p += d2)
      ppst->pam2[1][i] = d2p;
}

/*
 *  initpam2(struct pstruct pst): Converts 1-D pam matrix to 2-D
 */
void    initpam2 (struct pstruct *ppst)
{
   int     i, j, k, nsq, nsqx, pam_x, sa_x;

   nsq = ppst->nsq;

   sa_x = sascii['X'];
   ppst->pam2[0][0][0] = -BIGNUM;
   ppst->pam_h = -1; ppst->pam_l = 1;

   k = 0;
   for (i = 1; i <= nsq; i++) {
     ppst->pam2[0][0][i] = ppst->pam2[0][i][0] = -BIGNUM;
     for (j = 1; j <= i; j++) {
       ppst->pam2[0][j][i] = ppst->pam2[0][i][j] = pam[k++] - ppst->pamoff;
       if (ppst->pam_l > ppst->pam2[0][i][j]) ppst->pam_l =ppst->pam2[0][i][j];
       if (ppst->pam_h < ppst->pam2[0][i][j]) ppst->pam_h =ppst->pam2[0][i][j];
     }
   }

   if (ppst->pam_x > 0) ppst->pam_x = ppst->pam2[0][sa_x][sa_x];
   else { 
     ppst->pam2[0][sa_x][sa_x]=ppst->pam_x;
     for (i=1; i<=nsq; i++)
       ppst->pam2[0][sa_x][i] = ppst->pam2[0][i][sa_x]=ppst->pam_x;
   }
   pam_x = ppst->pam_x;

   if (ppst->ext_sq_set) {	/* using extended alphabet */
     /* fill in pam2[1] matrix */
     ppst->pam2[1][0][0] = -BIGNUM;
     /* fill in additional parts of the matrix */
     for (i = 1; i <= nsq; i++) {
       ppst->pam2[0][0][i+nsq] = ppst->pam2[0][i+nsq][0] = 
       ppst->pam2[1][0][i+nsq] = ppst->pam2[1][i+nsq][0] = 
	 ppst->pam2[1][0][i] = ppst->pam2[1][i][0] = -BIGNUM;
       for (j = 1; j <= nsq; j++) {
	 ppst->pam2[0][i+nsq][j] = ppst->pam2[0][i][j+nsq] =
	   ppst->pam2[0][i+nsq][j+nsq] = ppst->pam2[1][i][j] =
	   ppst->pam2[0][i][j];

	 ppst->pam2[1][i+nsq][j] = ppst->pam2[1][i][j+nsq] = 
	   ppst->pam2[1][i+nsq][j+nsq] = pam_x;
       }
     }
   }
}

/*
 * initenv () : initializes the environment
 */
void    f_initenv (m_msg, ppst, aa0)
struct mngmsg *m_msg;
struct pstruct *ppst;
char  **aa0;
{
#ifdef PRSS
  m_msg->aln.llen = 0;
#endif
   m_msg->stages = 1;
   strncpy (m_msg->label, "s-w", sizeof(m_msg->label));
   strncpy (m_msg->f_id0,"sw",3);
   strncpy (m_msg->f_id1,"sw",3);
   ppst->histint = 2;
#ifndef GAP_OPEN
   ppst->gdelval = -12;
#else
   ppst->gdelval = -10;
#endif
   ppst->ggapval = -2;
   m_msg->nframe = -1;
   m_msg->qframe = 1;
   ppst->score_ix = 0;
   ppst->sw_flag = 1;
   m_msg->nrelv = 1;		/* number of relevant scores */
   m_msg->srelv = 1;		/* relevant scores in showbest */
   m_msg->arelv = 1;		/* number of alignment scores */
   strncpy (m_msg->alab[0],"s-w opt",20);

   alloc_pam (MAXSQ, MAXSQ, ppst);
}

static int gap_set=0;
static int del_set=0;
static int e_cut_set=0;
static int nframe_set = 0;

void    f_getopt (copt, optarg, m_msg, ppst)
char    copt;
char   *optarg;
struct mngmsg *m_msg;
struct pstruct *ppst;
{
  switch (copt) {
  case '3':
    m_msg->qframe = m_msg->nframe = 1; 
    nframe_set = 1;
    break;
  case 'E': sscanf(optarg,"%lf",&m_msg->e_cut);
    e_cut_set = 1;
    break;
  case 'F': sscanf(optarg,"%lf",&m_msg->e_low);
    e_cut_set = 1;
    break;
  case 'f':
    sscanf (optarg, "%d", &ppst->gdelval);
    if (ppst->gdelval > 0) ppst->gdelval = -ppst->gdelval;
    del_set = 1;
    break;
  case 'g':
    sscanf (optarg, "%d", &ppst->ggapval);
    if (ppst->ggapval > 0) ppst->ggapval = -ppst->ggapval;
    gap_set = 1;
    break;
  }
}

void    f_getarg (argc, argv, optind, m_msg, ppst)
int     argc;
char  **argv;
int     optind;
struct mngmsg *m_msg;
struct pstruct *ppst;
{
  /* this sets up the number of shuffles for PRSS */
#ifdef PRSS
  if (argc - optind == 4) {
      sscanf (argv[optind + 3], "%d", &m_msg->mshow);
      m_msg->mshow_flg = 1;
  }
#endif
}

void
resetp (unsigned char *aa0, int n0,
	struct mngmsg *m_msg, struct pstruct *ppst)
{
  int i;

  if (ppst->dnaseq == 1) {
    if (!nframe_set) {
      m_msg->nframe = m_msg->qframe = 2;
    }

    ppst->histint = 4;
#ifndef GAP_OPEN
    if (!del_set) ppst->gdelval = -16;
#else
    if (!del_set) ppst->gdelval = -12;
#endif
    if (!gap_set) ppst->ggapval = -4;

    if (!e_cut_set) m_msg->e_cut=2.0;

    ppst->nsq = nnt;
    ppst->nsqx = nntx;

    for (i=0; i<=ppst->nsqx; i++) {
      ppst->hsq[i] = hnt[i];
      ppst->sq[i] = nt[i];
      ppst->hsqx[i] = hntx[i];
      ppst->sqx[i] = ntx[i];
    }

    if (ppst->p_d_set)
      mk_n_pam(npam,nnt,ppst->p_d_mat,ppst->p_d_mis);
    else if (strncmp(ppst->pamfile,"BL50",4)==0) {
      strncpy (ppst->pamfile, "+5/-4", sizeof(ppst->pamfile));
    }

    /* this may not be necessary, as init_ascii fills out sascii */
    sascii = nascii;
    pam = npam;	    /* required, broken in pre-March 3 fa33t08 */

    if (ppst->ext_sq_set) init_ascii(ppst->sqx,ppst->nsqx,sascii,1);
    else init_ascii(ppst->sq,ppst->nsq,sascii,1);

    strncpy (m_msg->sqnam, "nt",sizeof(m_msg->sqnam));
    strncpy (m_msg->sqtype, "DNA",sizeof(m_msg->sqtype));
  }
}

/* query_parm()	this function asks for any additional parameters
	that have not been provided.  Could be null. */

void
query_parm (m_msp, ppst)
struct mngmsg *m_msp;
struct pstruct *ppst;
{
   char    qline[40];

#ifdef PRSS
   if (!m_msp->mshow_flg) m_msp->mshow = 200;

 l1:
   printf(" number of shuffles [%d]? ",m_msp->mshow);
   fflush(stdout);
   if (fgets (qline, sizeof(qline), stdin) == NULL) exit (0);
   else sscanf(qline,"%d",&m_msp->mshow);

   m_msp->mshow_flg = 1;

   if (m_msp->aln.llen < 1) {
     printf (" local (window) (w) or uniform (u) shuffle [u]? ");
     if (fgets (qline, sizeof(qline), stdin) == NULL) exit (0);
     else if (qline[0]=='w' || qline[0]=='W') {
       m_msp->aln.llen = 20;
       printf(" local shuffle window size [%d]? ",m_msp->aln.llen);
       if (fgets (qline, sizeof(qline), stdin) == NULL) exit (0);
       else sscanf(qline,"%d",&m_msp->aln.llen);
     }
   }

#endif
}

void
last_init (struct mngmsg *m_msg, struct pstruct *ppst)
{
  m_msg->thr_fact = 1;
#ifdef PRSS
   if (!m_msg->mshow_flg) m_msg->mshow = 200;
#endif

  initpam2(ppst);

   /* once we have a complete pam matrix, we can calculate Lambda and K 
      for "average" sequences */

   ppst->kar_p = NULL;
   init_karlin_a(ppst, ppst->aa0_f, &ppst->kar_p);
   do_karlin_a(ppst->pam2[0], ppst,ppst->aa0_f, 
	       ppst->kar_p, &ppst->Lambda, &ppst->K, &ppst->H);
}

void
f_initpam (line, ppst)
char   *line;
struct pstruct *ppst;
{}

