/* mysql_lib.c copyright (c) 2000 William R. Pearson */

/* $Name: fa33t09d0 $ - $Id: mysql_lib.c,v 1.1.2.2.2.1 2001/05/18 20:50:11 wrp Exp $ */

/* functions for opening, reading, seeking a mySQL database */

/*
  For the moment, this interface assumes that the file to be searched will
  be specified in a single, long, string with 4 parts:

  (1) a database open string. This string has four fields, separated by
      whitespace (' \t'):
        hostname:port dbname user password

  (2) a database query string that will return an unique ID and a sequence
      string

  (3) a database select string that will return a description
      given a unique ID

  (4) a database select string that well return a sequence given a
      unique ID


   This string can contain "\n". ";" are used to separate the four
   functions, which must be specified in the order shown above.
   The last (fourth) query must terminate with a ';';
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mysql.h>

#define XTERNAL
#include "uascii.h"
#include "upam.h"

int sfnum[10], nsfnum;

/* pointers to the 4 mysql strings */
static char *sql_db, *sql_query, *sql_getdesc, *sql_getseq;
static MYSQL *sql_conn;
static MYSQL_RES *sql_res=NULL;
static int sql_reopen = 0;

int
mysql_openlib(char *sname) {
  FILE *sql_file;
  char tmp_str[1024];
  char *bp, *bps, *tp;
  int i;
  char *sql_host, *sql_dbname, *sql_user, *sql_pass;
  int sql_port;

  if (sql_reopen) return 1;

  if (sname[0] == '%') strncpy(tmp_str,sname,sizeof(tmp_str));
  else {
    if ((sql_file=fopen(sname,"r"))==NULL) {
      fprintf(stderr," cannot open mySQL file: $s\n",sname);
      return 0;
    }
    if (fread(tmp_str,sizeof(char),sizeof(tmp_str),sql_file)<=0) {
      fprintf(stderr," cannot read mySQL file: $s\n",sname);
      return 0;
    }
  }

  /* copy/parse database field */
  bps = tmp_str;
  if ((bp=strchr(bps,';'))!=NULL) {
    *bp='\0';
    if ((sql_db=calloc(strlen(bps)+1,sizeof(char)))==NULL) {
      fprintf(stderr, " cannot allocate space for database name [%d], %s\n",
	      strlen(bps),bps);
      return 0;
    }
    /* have database name, parse the fields */
    else {
      strcpy(sql_db,bps);
      bps = bp+1;
      *bp=';'; /* replace ; */
      sql_host = strtok(sql_db," \t\n");
      sql_dbname = strtok(NULL," \t\n");
      sql_user = strtok(NULL," \t\n");
      sql_pass = strtok(NULL," \t\n");
      if ((tp=strchr(sql_host,':'))!=NULL) {
	*tp='\0';
	sql_port=atoi(tp+1);
      }
      else sql_port = 0;
    }
  }
  else {
    fprintf(stderr," cannot find database fields:\n%s\n",tmp_str);
    return 0;
  }

  /* copy 1st query field */
  if ((bp=strchr(bps,';'))!=NULL) {
    *bp='\0';
    if ((sql_query=calloc(strlen(bps)+1,sizeof(char)))==NULL) {
      fprintf(stderr, " cannot allocate space for query string [%d], %s\n",
	      strlen(bps),bps);
      return 0;
    }
    /* have query, copy it */
    else {
      strcpy(sql_query,bps);
      *bp=';'; /* replace ; */
      bps = bp+1;
    }
  }
  else {
    fprintf(stderr," cannot find database query field:\n%s\n",tmp_str);
    return 0;
  }

  /* copy get_desc field */
  if ((bp=strchr(bps,';'))!=NULL) {
    *bp='\0';
    if ((sql_getdesc=calloc(strlen(bps)+1,sizeof(char)))==NULL) {
      fprintf(stderr, " cannot allocate space for database name [%d], %s\n",
	      strlen(bps),bps);
      return 0;
    }
    /* have get_desc, copy it */
    else {
      strcpy(sql_getdesc,bps);
      *bp=';'; /* replace ; */
      bps = bp+1;
    }
  }
  else {
    fprintf(stderr," cannot find getdesc field:\n%s\n",tmp_str);
    return 0;
  }

  if ((bp=strchr(bps,';'))!=NULL) { *bp='\0';}

  if ((sql_getseq=calloc(strlen(bps)+1,sizeof(char)))==NULL) {
    fprintf(stderr, " cannot allocate space for database name [%d], %s\n",
	    strlen(bps),bps);
    return 0;
  }

  if (strlen(bps) > 0) {
    strcpy(sql_getseq,bps);
  }
  else {
    fprintf(stderr," cannot find getseq field:\n%s\n",tmp_str);
    return 0;
  }
  if (bp!=NULL) *bp=';';

  if ((sql_conn=mysql_init(NULL))==NULL) {
    fprintf(stderr,"*** Error - mysql_init\n");
    return 0;
  }

  if (mysql_real_connect(sql_conn,
			 sql_host,sql_user,sql_pass,
			 sql_dbname,
			 sql_port,
			 NULL,
			 0)==NULL)
    {
    fprintf(stderr,"*** Error %u - could  not open database:\n%s\n%s",
	    mysql_errno(sql_conn),tmp_str, mysql_error(sql_conn));
    return 0;
  }
  else {
    fprintf(stderr," Database %s opened on %s\n",sql_dbname,sql_host);
  }

  /* now do the query */    

  if (mysql_query(sql_conn,sql_query)) {
    fprintf(stderr,"*** Error %u - query failed:\n%s\n%s\n",
	    mysql_errno(sql_conn), sql_query, mysql_error(sql_conn));
    return 0;
  }

  if ((sql_res = mysql_use_result(sql_conn)) == NULL) {
    fprintf(stderr,"*** Error = use result failed\n%s\n",
	    mysql_error(sql_conn));
    return 0;
  }

  return 1;
}

mysql_closelib() {

  if (sql_res != NULL)
    mysql_free_result(sql_res);
  mysql_close(sql_conn);
  sql_reopen=0;
}

static char *sql_seq = NULL, *sql_seqp;
static int sql_seq_len;
static MYSQL_ROW sql_row;

mysql_getlib( unsigned char *seq,
	      int maxs,
	      char *libstr,
	      long *libpos,
	      int *lcont)
{
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *bp;
  static long seq_len;
  static unsigned char *cp_max;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;

#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif

  sfnum[0]=nsfnum = 0;

  if (*lcont==0) {
    /* get a row, with UID, sequence */
    if ((sql_row =mysql_fetch_row(sql_res))!=NULL) {
      *libpos=atol(sql_row[0]);
      sql_seq = sql_seqp = sql_row[1];
#ifndef PCOMPLIB
      /* the normal case returns only GID/sequence */
      strncpy(libstr,sql_row[0],20);
      libstr[12]='\0';
#else
      /* the PVM case also returns a long description */
      strncpy(libstr,sql_row[2],255);
      libstr[255]='\0';
#endif
    }
    else {
      mysql_free_result(sql_res);
      sql_res=NULL;
      *lcont = 0;
      *seqp = EOSEQ;
      sql_reopen = 1;
      return -1;
    }
  }

  for (cp=(unsigned char *)sql_seqp; seqp<seqm1 && *cp; ) {
    if ((*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA &&
	(*seqp++=ap[*cp++])<NA) continue;
    --seqp;
    if (*(cp-1)==0) break;
  }
  sql_seqp = (char *)cp;

  if (seqp>=seqm1) (*lcont)++;
  else {
    *lcont=0;
    if (sql_reopen) {
      mysql_free_result(sql_res);
      sql_res = NULL;
    }
  }

  *seqp = EOSEQ;
  /*   if ((int)(seqp-seq)==0) return 1; */
  return (int)(seqp-seq);
}

void
mysql_ranlib(str,cnt,libpos)
	char *str; int cnt;
	long libpos;
{
  char tmp_query[1024], tmp_val[20];
  char *bp;
  int i;

  str[0]='\0';

  /* put the UID into the query string - cannot use sprintf because of
     "%' etc */

  /*   sprintf(tmp_query,sql_getdesc,libpos); */

  if ((bp=strchr(sql_getdesc,'#'))==NULL) {
    fprintf(stderr, "no GID position in %s\n",sql_getdesc);
    return;
  }
  else {
    *bp = '\0';
    strncpy(tmp_query,sql_getdesc,sizeof(tmp_query));
    sprintf(tmp_val,"%ld",libpos);
    strncat(tmp_query,tmp_val,sizeof(tmp_query));
    strncat(tmp_query,bp+1,sizeof(tmp_query));
    *bp='#';
  }

  /*  fprintf(stderr," requesting: %s\n",tmp_query); */

  if (sql_res !=NULL) {
    mysql_free_result(sql_res);
    sql_res = NULL;
  }

  if (mysql_query(sql_conn,tmp_query)) {
    fprintf(stderr,"*** Error - query failed:\n%s\n%s",tmp_query,
	    mysql_error(sql_conn));
    goto l_seq1;
  }

  if ((sql_res = mysql_use_result(sql_conn)) == NULL) {
    fprintf(stderr,"*** Error = use result failed\n%s\n",
	    mysql_error(sql_conn));
    goto l_seq1;
  }
  
  /* have the description */
  if ((sql_row = mysql_fetch_row(sql_res))==NULL) {
/*    fprintf(stderr," cannot fetch description: %s\n",tmp_query); */
    sprintf(str,"id:%ld *** Description not found *** ",libpos);
    goto l_seq;
  }
  
  strncpy(str,sql_row[1],cnt-1);
  for (i = 2; i< mysql_num_fields(sql_res); i++) {
    strncat(str," ",cnt-1);
    strncat(str,sql_row[i],cnt-1);
  }

  str[cnt-1]='\0';
  while ((bp = strchr(str,'\001'))!=NULL ||
	 (bp=strchr(str,'\t'))!=NULL) *bp=' ';
  if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';

 l_seq:
  mysql_free_result(sql_res);
 l_seq1:
  sql_res = NULL;

  /* get the sequence, set up for mysql_getseq() */
  /* put the UID into the query string */

  if ((bp=strchr(sql_getseq,'#'))==NULL) {
    fprintf(stderr, "no GID position in %s\n",sql_getseq);
    return;
  }
  else {
    *bp = '\0';
    strncpy(tmp_query,sql_getseq,sizeof(tmp_query));
    sprintf(tmp_val,"%ld",libpos);
    strncat(tmp_query,tmp_val,sizeof(tmp_query));
    strncat(tmp_query,bp+1,sizeof(tmp_query));
    *bp='#';
  }

  if (mysql_query(sql_conn,tmp_query)) {
    fprintf(stderr,"*** Error - query failed:\n%s\n%s",tmp_query,
	    mysql_error(sql_conn));
  }

  if ((sql_res = mysql_use_result(sql_conn)) == NULL) {
    fprintf(stderr,"*** Error = use result failed\n%s\n",
	    mysql_error(sql_conn));
  }
}
