/*	May, June 1987	- modified for rapid read of database

	copyright (c) 1987,1988,1989,1992,1995,2000 William R. Pearson

	This is one of three alternative files that can be used to
	read a database.  The three files are nxgetaa.c, nmgetaa.c, and
	mmgetaa.c.

	nxgetaa.c contains the original code for reading databases, and
	is still used for Mac and PC versions of fasta33 (which do not
	use mmap).

	nmgetaa.c and mmgetaa.c are used together.  nmgetaa.c provides
	the same functions as nxgetaa.c if memory mapping is not used,
	mmgetaa.c provides the database reading functions if memory
	mapping is used. The decision to use memory mapping is made on
	a file-by-file basis.

	June 2, 1987 - added TFASTA
	March 30, 1988 - combined ffgetaa, fgetgb;
	April 8, 1988 - added PIRLIB format for unix
	Feb 4, 1989 - added universal subroutines for libraries
	December, 1995 - added range option file.name:1-1000
	September, 1999 - added option for mmap()ed files using ".xin"


	getnt.c	associated subroutines for matching sequences */

/* $Name: fa33t09d0 $ - $Id: nmgetaa.c,v 1.8.4.5.2.2 2001/04/05 14:16:08 wrp Exp $ */

/*
8-April-88
	The compile time #define PIRLIB allows this routine to be used
	to read protein and DNA sequence libraries in the NBRF/PIR
	VAX/VMS library format.  That is:

	>P1;LCBO
	This is a line of description
	GTYH ... the sequence starts on this line

	This may ease conversion from UWGCG format libraries. It
	has not been extensively tested.

	In addition, sequence libraries with a '>' in the 4th position
	are recognized as NBRF format libraries for consistency with
	UWGCG

	February 4, 1988 - this starts a major revision of the getaa
	routines.  The goal is to be able to seach the following format
	libraries:

		0 - normal FASTA format
		1 - full Genbank tape format
		2 - NBRF/PIR CODATA format
		3 - EMBL/Swiss-prot format
		4 - Intelligentics format
		5 - NBRF/PIR VMS format
		6 - GCG 2bit format

		11 - NCBI setdb/blastp (1.3.2) AA

	see file altlib.h to confirm numbers

	This is done with a new global variable and a requirement for the
	FASTLIBS file.  The FASTLIBS file will now indicate both the sequence
	type (protein = 0, DNA = 1) and the file format (the numbers shown
	above, although intelligenetics may become an alternative to Pearson).
	This will be done by always using a function pointer for getlib and
	ranlib(), and setting up a bunch of different getlib() and ranlib()
	functions.  Openlib() will be substantially simplified.
*/

/* 	Nov 12, 1987	- this version checks to see if the sequence
	is DNA or protein by asking whether > 85% is A, C, G, T

	May 5, 1988 - modify the DNA/PROTEIN checker by re-reading
	DNA sequences in order to check for 'U'.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defs.h"
#include "uascii.h"
#include "structs.h"

#ifndef SFCHAR
#define SFCHAR ':'
#endif

#ifdef VMS
#define PIRLIB
#endif

#define XTERNAL
#include "upam.h"
#undef XTERNAL

#define YES 1
#define NO 0
#define MAXLINE 512

#ifndef min
#define min(x,y) ((x) > (y) ? (y) : (x))
#endif

#ifdef __MWERKS__

/* BIGBUFF allows the Mac to use setvbuf to set up a new buffer for reading
the library, that is 32K instead of 0.5k.  This appears to increase the speed
of the program by about 33%.  */

char *bigbuf=NULL;
long fbufsize=128L;

#define BIGBUF
extern char prompt[];
#include <StandardFile.h>
extern StandardFileReply freply;
extern FSSpec q1Spec;
extern void STFileDlog(char *prompt,StandardFileReply *freply,
						SFTypeList plist,int nl);
#endif

int nsfnum;	/* number of superfamily numbers */
int sfnum[10];	/* superfamily number from types 0 and 5 */
int nsfnum_n;
int sfnum_n[10];

#ifdef SUPERFAMNUM
static char tline[MAXLINE];
#endif

#define NO_FORMAT 0
#define FASTA_FORMAT 1
#define GCG_FORMAT 2

static int seq_format=NO_FORMAT;
static char seq_title[200];

int scanseq(unsigned char *, int, char *);
void sf_sort(int *, int);

getseq(char *filen, unsigned char *seq, int maxs, char *libstr,
       int *dnaseq,long *sq0off)
{
  FILE *fptr;
  char line[512],*bp, *bp1, *bpn, *tp;
  int i, j, rn, n;
  int ic;
  int sstart, sstop, sset=0;
  int llen, l_offset;

  seq_title[0]='\0';

  sstart = sstop = -1;
#ifndef DOS
  if ((bp=strchr(filen,':'))!=NULL && *(bp+1)!='\0') {
#else
  if ((bp=strchr(filen+3,':'))!=NULL && *(bp+1)!='\0') {
#endif
    *bp='\0';
    if (*(bp+1)=='-') {
      sstart = 0;
      sscanf(bp+2,"%d",&sstop);
    }
    else {
      sscanf(bp+1,"%d-%d",&sstart,&sstop);
      sstart--;
      if (sstop <= 0 ) sstop = BIGNUM;
    }
    sset=1;
  }
  else {
    sstart = 0;
    sstop = BIGNUM;
  }

  /* check for input from stdin */
  if (strcmp(filen,"-") && strcmp(filen,"@")) {
    if ((fptr=fopen(filen,"r"))==NULL) {
      fprintf(stderr," could not open %s\n",filen);
      return 0;
    }
  }
  else {
    fptr = stdin;
  }


  rn = n=0;
  while(fgets(line,sizeof(line),fptr)!=NULL) {
#ifdef PIRLIB
    if (line[0]=='>'&& (line[3]==';'||line[3]=='>'))
      fgets(line,sizeof(line),fptr);
    else
#endif
    if (line[0]=='>') {
      l_offset = 0;
      seq_format = FASTA_FORMAT;
#ifdef STAR_X
      sascii['*'] = sascii['X'];
#endif
    sfnum[nsfnum=0]=sfnum_n[nsfnum_n=0]=0;
#ifdef SUPERFAMNUM
    strncpy(tline,line+1,sizeof(tline));
    tline[sizeof(tline)-1]='\0';

    if ((bp=strchr(tline,' ')) && (bp=strchr(bp+1,SFCHAR))) {
      if ((bp1=strchr(bp+1,SFCHAR))==NULL) {
	fprintf(stderr," second %c missing: %s\n",SFCHAR,tline);
      }
      else {
	if ((bpn=strchr(bp+1,NSFCHAR))!=NULL) *bpn = '\0';
	*bp1 = '\0';
	i = 0;
	if ((tp = strtok(bp+1," \t"))!=NULL)  {
	  sfnum[i++] = atoi(tp);
	  while ((tp = strtok((char *)NULL," \t")) != (char *)NULL) {
	    if (isdigit(*tp)) sfnum[i++] = atoi(tp);
	    if (i>=9) break;
	  }
	}
	sfnum[nsfnum=i]= 0;
	if (nsfnum>1) sf_sort(sfnum,nsfnum);
	else {
	  if (nsfnum < 1) fprintf(stderr," found | but no sfnum: %s\n",libstr);
	}
	if (bpn != NULL) {
	  tp = strtok(bpn+1," \t");
	  sfnum_n[0]=atoi(tp);
	  i = 1;
	  while ((tp=strtok(NULL," \t"))!=NULL) {
	    sfnum_n[i++] = atoi(tp);
	    if (i >= 10) {
	      fprintf(stderr,
		      " error - too many negative superfamilies: %d\n %s\n",
		      i,tline);
	      break;
	    }
	  }
	  sfnum_n[nsfnum_n=i]=0;
	  sf_sort(sfnum_n,nsfnum_n);
	}
      }
    }
    else {
      sfnum[nsfnum = 0] = 0;
      sfnum_n[nsfnum_n = 0] = 0;
      }
#else
      sfnum[nsfnum = 0] = 0;
      sfnum_n[nsfnum_n = 0] = 0;
#endif
      if ((bp=(char *)strchr(line,'\n'))!=NULL) *bp='\0';
      strncpy(seq_title,line+1,sizeof(seq_title));
      seq_title[sizeof(seq_title)-1]='\0';
      if ((bp=(char *)strchr(line,' '))!=NULL) *bp='\0';
      strncpy(libstr,line+1,12);
      libstr[12]='\0';
    }
    else if (seq_format==NO_FORMAT) {
      seq_format = GCG_FORMAT;
      sascii['*'] = sascii['X'];
      l_offset = 10;
      llen = strlen(line);
      while (strncmp(&line[llen-3],"..\n",(size_t)3) != 0) {
	if (fgets(line,sizeof(line),fptr)==NULL) return 0;
	llen = strlen(line);
      }
      bp = strtok(line," \t");
/*
      if ((bp=(char *)strchr(line,' '))!=NULL) *bp='\0';
      else if ((bp=(char *)strchr(line,'\n'))!=NULL) *bp='\0';
*/
      if (bp!=NULL) strncpy(libstr,bp,12);
      else strncpy(libstr,filen,12);
      libstr[12]='\0';
      if (fgets(line,sizeof(line),fptr)==NULL) return 0;
    }

    if (seq_format==GCG_FORMAT && strlen(line)<l_offset) continue;

    if (line[0]!='>'&& line[0]!=';') {
      for (i=l_offset; (n<maxs && rn < sstop)&&
	     ((ic=sascii[line[i]&AAMASK])<EL); i++)
	if (ic<NA && ++rn > sstart) seq[n++]= ic;
      if (ic == ES || rn > sstop) break;
    }
  }

  if (n==maxs) {
    fprintf(stderr," sequence may be truncated %d %d\n",n,maxs);
    fflush(stderr);
  }
  if ((bp=strchr(libstr,'\n'))!=NULL) *bp = '\0';
  if ((bp=strchr(libstr,'\r'))!=NULL) *bp = '\0';
  seq[n]= EOSEQ;

  if (fptr!=stdin && *dnaseq ==0 && n > 0
      && (float)scanseq(seq,n,"ACGTXN")/(float)n > 0.85) {
    *dnaseq = 1;
    /* convert from protein to DNA sequence */
    sascii = nascii;
    rewind(fptr);
    rn = n = 0;

    if (seq_format !=GCG_FORMAT) 
      while(fgets(line,sizeof(line),fptr)!=NULL) {
#ifdef PIRLIB
	if (line[0]=='>'&& (line[3]==';'||line[3]=='>'))
	  fgets(line,sizeof(line),fptr);
	else
#endif
	if (line[0]!='>'&& line[0]!=';') {
	  for (i=0; (n<maxs && rn < sstop)&&
		 ((ic=sascii[line[i]&AAMASK])<EL); i++)
	    if (ic<NA && ++rn > sstart ) seq[n++]= ic;
	  if (ic == ES || rn > sstop) break;
	}
      }
    else {
      llen = strlen(line);
      while (strncmp(&line[llen-3],"..\n",(size_t)3) != 0) {
	if (fgets(line,sizeof(line),fptr)==NULL) return 0;
	llen = strlen(line);
      }
      while (fgets(line,sizeof(line),fptr)!=NULL) {
	if (strlen(line)<l_offset) continue;
	for (i=l_offset; (n<maxs && rn < sstop) &&
	       ((ic=sascii[line[i]&AAMASK])<EL); i++)
	  if (ic<NA && ++rn > sstart ) seq[n++]= ic;
	if (ic == ES || rn > sstop ) break;
      }
    }

    if (n==maxs) {
      fprintf(stderr," sequence may be truncated %d %d\n",n,maxs);
      fflush(stderr);
    }
    seq[n]= EOSEQ;

    pam = npam;
    strcpy(sqnam,"nt");
    strcpy(sqtype,"DNA");
  }

  if (fptr!=stdin) fclose(fptr);

  if (sset==1) {
    sstart++;
    filen[strlen(filen)]=':';
    if (*sq0off==1 || sstart>=1) *sq0off = sstart;
  }

  return n;
}

getntseq(char *filen, unsigned char *seq, int maxs, char *libstr,
       int *dnaseq,long *sq0off)
{
  FILE *fptr;
  char line[512],*bp, *bp1, *tp;
  int i, j, n, rn;
  int ic;
  int sstart, sstop, sset=0;
  int llen, l_offset;

  sstart = sstop = -1;
#ifndef DOS
  if ((bp=strchr(filen,':'))!=NULL) {
#else
  if ((bp=strchr(filen+3,':'))!=NULL) {
#endif
    *bp='\0';
    if (*(bp+1)=='-') {
      sstart = 0;
      sscanf(bp+2,"%d",&sstop);
    }
    else {
      sscanf(bp+1,"%d-%d",&sstart,&sstop);
      sstart--;
      if (sstop <= 0) sstop = BIGNUM;
    }
    sset=1;
  }
  else {
    sstart = 0;
    sstop = BIGNUM;
  }
    

  if (strcmp(filen,"-") && strcmp(filen,"@")) {
    if ((fptr=fopen(filen,"r"))==NULL) {
      fprintf(stderr," could not open %s\n",filen);
      return 0;
    }
  }
  else {
    fptr = stdin;
  }

  if (sset==1) {
    filen[strlen(filen)]=':';
    if (*sq0off==1 || sstart>1) *sq0off = sstart;
  }

  rn = n = 0;
  while(fgets(line,sizeof(line),fptr)!=NULL) {
#ifdef PIRLIB
    if (line[0]=='>'&& (line[3]==';'||line[3]=='>'))
      fgets(line,sizeof(line),fptr);
    else
#endif
    if (line[0]=='>') {
      l_offset = 0;
      seq_format = FASTA_FORMAT;
#ifdef SUPERFAMNUM
    strncpy(tline,line+1,sizeof(tline));
    tline[sizeof(tline)-1]='\0';
    sfnum[nsfnum=0]=0;
    if ((bp=strchr(tline,' ')) &&
	 (bp=strchr(bp+1,SFCHAR))) {

      if ((bp1=strchr(bp+1,SFCHAR))==NULL) {
	fprintf(stderr," second %c missing: %s\n",SFCHAR,tline);
      }
      else {
	*bp1 = '\0';
	i = 0;
	if ((tp = strtok(bp+1," \t"))!=NULL) {
	  sfnum[i++] = atoi(tp);
	  while ((tp = strtok((char *)NULL," \t")) != (char *)NULL) {
	    sfnum[i++] = atoi(tp);
	    if (i>=9) break;
	  }
	}
	*bp1 = SFCHAR;
	sfnum[nsfnum=i]= 0;
	if (nsfnum>1) sf_sort(sfnum,nsfnum);
	else {
	  if (nsfnum<1) fprintf(stderr," found | but no sfnum: %s\n",libstr);
	}
      }
    }
    else {
      sfnum[0] = nsfnum = 0;
      }
#else
      sfnum[0] = nsfnum = 0;
#endif
	if ((bp=(char *)strchr(line,'\n'))!=NULL) *bp='\0';
	strncpy(seq_title,line+1,sizeof(seq_title));
	seq_title[sizeof(seq_title)-1] = '\0';
	if ((bp=(char *)strchr(line,' '))!=NULL) *bp='\0';
	strncpy(libstr,line+1,12);
	libstr[12]='\0';
    }
    else if (seq_format==NO_FORMAT) {
      seq_format = GCG_FORMAT;
      l_offset = 10;
      llen = strlen(line);
      while (strncmp(&line[llen-3],"..\n",(size_t)3) != 0) {
	if (fgets(line,sizeof(line),fptr)==NULL) return 0;
	llen = strlen(line);
      }
      if ((bp=(char *)strchr(line,' '))!=NULL) *bp='\0';
      else if ((bp=(char *)strchr(line,'\n'))!=NULL) *bp='\0';
      strncpy(libstr,line,12);
      libstr[12]='\0';
      if (fgets(line,sizeof(line),fptr)==NULL) return 0;
    }

    if (seq_format==GCG_FORMAT && strlen(line)<l_offset) continue;

    if (line[0]!='>'&& line[0]!=';') {
      for (i=l_offset; (n<maxs && rn < sstop)&&
	     ((ic=nascii[line[i]&AAMASK])<EL); i++)
	if (ic<NA && ++rn > sstart) seq[n++]= ic;
      if (ic == ES || rn > sstop) break;
    }
  }

  if (n==maxs) {
    fprintf(stderr," sequence may be truncated %d %d\n",n,maxs);
    fflush(stderr);
  }
  if ((bp=strchr(libstr,'\n'))!=NULL) *bp = '\0';
  if ((bp=strchr(libstr,'\r'))!=NULL) *bp = '\0';
  seq[n]= EOSEQ;

  if (fptr!=stdin) fclose(fptr);

  /*
  if (sset) {
    if (sstart <= 0) sstart = 1;
    if (sstop <= 0) sstop = n;
    sstart--;
    sstop--;
    for (i=0, j=sstart; j<=sstop; i++,j++)
      seq[i] = seq[j];
    n = sstop - sstart +1;
    seq[n]=EOSEQ;
  }
  */

  *dnaseq = 1;
  return n;
}

int
gettitle(char *filen, char *title, int len) {
  FILE *fptr;
  char line[512];
  char *bp;
  int ll,sset;
#ifdef MSDOS
  char *strpbrk();
#endif

  sset = 0;

  if (strncmp(filen,"-",1)==0 || strncmp(filen,"@",1)==0) {
    strncpy(title,seq_title,len);
    title[len-1]='\0';
    return (int)strlen(title);
  }

  if ((bp=strchr(filen,':'))!=NULL) { *bp='\0'; sset=1;}
	  

  if ((fptr=fopen(filen,"r"))==NULL) {
    fprintf(stderr," file %s was not found\n",filen);
    fflush(stderr);
    return 0;
  }

  if (sset==1) filen[strlen(filen)]=':';

  while(fgets(line,sizeof(line),fptr)!=NULL) {
    if (line[0]=='>'|| line[0]==';') goto found;
  }
  fclose(fptr);
  title[0]='\0';
  return 0;

 found:

#ifdef PIRLIB
  if (line[0]=='>'&&(line[3]==';'||line[3]=='>')) {
    if ((bp = strchr(line,'\n'))!=NULL) *bp='\0';
    ll=strlen(line); line[ll++]=' '; line[ll]='\0';
    fgets(&line[ll],sizeof(line)-ll,fptr);
  }
#endif

#ifdef MSDOS
  bp = strpbrk(line,"\n\r");
#else
  bp = strchr(line,'\n');
#endif
  if (bp!=NULL) *bp = 0;
  strncpy(title,line,len);
  title[len-1]='\0';
  fclose(fptr);
  return strlen(title);
}

FILE *libf=NULL;

#ifdef NOLIB
int leof = 0;
#endif

#ifndef USE_FSEEKO
#define FSEEK fseek
#define FTELL ftell
typedef long FSEEK_T;
static long lpos;
#else
#define FSEEK fseeko
#define FTELL ftello
typedef off_t FSEEK_T;
 static off_t lpos;
#endif


char lline[MAXLINE+1];

int lfflag=0;	/* flag for CRLF in EMBL CDROM files */
#define LFCHAR '\015'  /* for MWC 5.5 */

#ifndef NOLIB
#include "altlib.h"

int (*getlib)();
void (*ranlib)();

#define GETLIB agetlib
#define RANLIB aranlib
#else
#define LASTLIB 10
#define GETLIB getlib
#define RANLIB ranlib
#endif

/* the following is from fgetgb.c */

#ifdef __MWERKS__
SFTypeList llist={'TEXT',0L,0L,0L};
#define getenv mgetenv
#define LLN 1
#endif

#include <fcntl.h>
#ifndef O_RAW
#ifdef O_BINARY
#define O_RAW O_BINARY
#else
#define O_RAW 0
#endif		/* O_BINARY */
#endif		/* O_RAW */
int libfd= -1;

extern int outtty;	/* flag for no interaction */
#ifdef MSDOS
#define RBSTR "rb"	/* read file in binary mode */
#else
#define RBSTR "r"
#endif

void closelib(int iln, int libtype);
static int last_iln= -1;

#ifdef USE_MMAP
#include "mm_file.h"
struct m_file_str *m_file_arr[MAX_LF];
struct m_file_str *load_mmap(FILE *, char *, int, int);
void reload_mmap(struct m_file_str *);
void reopen_ncbl2(struct m_file_str *);
struct m_file_str *ncbl2_openlib(char *, int);
#endif

/* a file name for openlib may now include a library type suffix */

int openlib(int iln, char *lname, int *lb_types, int ldnaseq)  {
  char rline[10],sname[MAX_FN], iname[MAX_FN], *bp, *getenv ();
  int wcnt, ll, opnflg;
  int libtype;
#ifdef USE_MMAP
  FILE *libi=NULL;
  struct m_file_str *m_fptr;
#endif

#ifdef USE_MMAP
  /* if the file mmap()ed and has been opened - use it */
  if ((m_fptr=m_file_arr[iln]) != NULL) {
    if (m_fptr->ranlib == ranliba[NCBIBL20]) reopen_ncbl2(m_fptr);
    else reload_mmap(m_fptr);
    getlib = m_fptr->getlib;
    ranlib = m_fptr->ranlib;
    return 1;
  }
#endif

  if (lname[0]=='#') return -9;
  wcnt = 0;

  strncpy(sname,lname,MAX_FN);
  sname[MAX_FN-1]='\0';
    /* check for library type */
  if ((bp=strchr(sname,' '))!=NULL) {
    *bp='\0';
    sscanf(bp+1,"%d",&libtype);
    if (libtype<0 || libtype >= LASTLIB) {
      fprintf(stderr," invalid library type: %d (>%d)- resetting\n%s\n",
	      libtype,LASTLIB,lname);
      libtype=0;
    }
  }
  else libtype=0;
  lb_types[iln]=libtype;

    /* m_file_arr[iln] will be NULL for a non-mmap()ed file
       or for one that has not yet been mmap()ed */

#ifdef USE_MMAP
  if (last_iln >= 0 && m_file_arr[iln]==NULL && last_iln != iln)
    closelib(last_iln, libtype);
#else
  if (last_iln >= 0 && last_iln!=iln) closelib(last_iln, libtype);
#endif

  last_iln = iln;

#ifdef __MWERKS__
  HSetVol(NULL,q1Spec.vRefNum,q1Spec.parID);
  if (bigbuf==NULL) {
    if ((bp=getenv("BUFSIZE"))!=NULL) sscanf(bp,"%ld",&fbufsize);
    else fbufsize=128l;
    if ((bigbuf=malloc((long)(fbufsize*1024l)))==NULL)
      fprintf(stderr," cannot allocate %ld K buffer\n",fbufsize);
  }
#endif

#ifndef NOLIB
  getlib=getliba[libtype];
  ranlib=ranliba[libtype];

#ifdef USE_MMAP
  /* check for possible mmap()ed files */
  if (getlibam[libtype]!=NULL) {
    /* look for .xin file */
    newname(iname,sname,"xin",sizeof(iname));
    if ((libi=fopen(iname,"r"))!=NULL) { /* have a *.xin file, use mmap */
      if ((m_file_arr[iln]=
	   load_mmap(libi,sname,libtype,ldnaseq)) ==NULL) {
	fclose(libi);
      }
      else {
	getlib=m_file_arr[iln]->getlib=getlibam[libtype];
	ranlib=m_file_arr[iln]->ranlib=ranlibam[libtype];
	return 1;
      }
    }
  }
#endif

 l1:
  if (libtype<=LASTTXT) opnflg=((libf=fopen(sname,RBSTR))!=NULL);
#ifdef NCBIBL13
  else if (libtype==NCBIBL13) opnflg=(ncbl_openlib(sname,ldnaseq)!= -1);
#endif
#ifdef NCBIBL20
  else if (libtype==NCBIBL20) {
#ifdef USE_MMAP
    opnflg=((m_file_arr[iln]=ncbl2_openlib(sname,ldnaseq))!=NULL);
#else
    opnflg=(ncbl2_openlib(sname,ldnaseq)!= 0);
#endif
  }
#endif
#ifdef MYSQL_DB
  /* a mySQL filename contains mySQL commands, not sequences */
  else if (libtype==MYSQL_LIB) {
    opnflg=mysql_openlib(sname);
  }
#endif

  if (!opnflg) {
#else
  l1:	if ((libf=fopen(sname,"r"))==NULL) {
#endif

#ifdef __MWERKS__
    rline[0]='\0';
    sprintf(prompt," cannot open %s\r Select library filename",sname);
    STFileDlog(prompt,&freply,llist,LLN);
    if (freply.sfGood==TRUE) {
      PtoCstr((StringPtr)freply.sfFile.name);
      strcpy(sname,(char *)freply.sfFile.name);
      strcpy(lname,sname);
      q1Spec.vRefNum = freply.sfFile.vRefNum;
      q1Spec.parID = freply.sfFile.parID;
      HSetVol(NULL,q1Spec.vRefNum,q1Spec.parID);
      goto l1;
    }
    else return -1;
  }
#else
  if (outtty) {
    fprintf(stderr," cannot open %s library\n",sname);
    fprintf(stderr," enter new file name or <RET> to quit ");
    fflush(stderr);
    if (fgets(sname,MAX_FN,stdin)==NULL) return -1;
    if ((bp=strchr(sname,'\n'))!=0) *bp='\0';
    if (strlen(sname)==0) return (-1);
    if (++wcnt > 10) return -1;
    strcpy(lname,sname);
    goto l1;
  }
  else return -1;
  }

#endif	/* __MWERKS__ */

#ifndef NOLIB
#ifdef BIGBUF
  if (bigbuf!=NULL)
    setvbuf(libf,bigbuf,_IOFBF,(size_t)(fbufsize*1024l));
#endif
  if (libtype<=LASTTXT) {
    lpos = FTELL(libf);
    if (fgets(lline,MAXLINE,libf)==NULL) return -1;
#ifdef __MWERKS__
    if (libtype==EMBLSWISS || libtype==VMSPIR ||libtype==FULLGB) {
      fgets(lline,MAXLINE,libf);
      lfflag = (lline[0]==LFCHAR);
      rewind(libf);
      lpos = 0;
      fgets(lline,MAXLINE,libf);
      if (lfflag) {getc(libf);
      /*		fprintf(stderr," lfflag is set\n");  */
      }
    }
#endif
  }		
#else		/* NOLIB */
  lpos = FTELL(libf);
  if (fgets(lline,MAXLINE,libf)==NULL) return -1;
  leof = 0;
#endif	/* NOLIB */
  return 1;
}

void
closelib(int iln, int libtype) {

#ifdef MMAP
  if (getlibam[libtype]!=NULL) {
    close_mmap();
    return;
  }
#endif

  if (libf!=NULL) {
    fclose(libf);
    libf = NULL;
  }

#ifndef NOLIB
#ifdef NCBIBL13
  if (libtype == NCBIBL13) ncbl_closelib();
#endif
#ifdef NCBIBL20
  if (libtype == NCBIBL20) ncbl2_closelib();
#endif
#ifdef MYSQL_DB
  if (libtype == MYSQL_LIB) mysql_closelib();
#endif
#endif
}

GETLIB(unsigned char *seq,
       int maxs,
       char *libstr,
       FSEEK_T *libpos,
       int *lcont)
{
  int ll;
  int ic, i;
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *linep, *bp, *bp1, *bpa, *tp;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='>' && lline[0]!=';') {
      lpos = FTELL(libf);
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
    }
#ifdef SUPERFAMNUM
    strncpy(tline,lline+1,sizeof(tline));
    tline[sizeof(tline)-1]='\0';
    sfnum[0]=nsfnum=0;
    if ((bp=strchr(tline,' ')) && (bp=strchr(bp+1,SFCHAR))) {
      if ((bpa = strchr(bp+1,'\001'))!=NULL) *bpa = '\0';
      if ((bp1=strchr(bp+1,SFCHAR))==NULL) {
/*	fprintf(stderr," second %c missing: %s\n",SFCHAR,libstr); */
      }
      else {
	*bp1 = '\0';
	i = 0;
	if ((tp = strtok(bp+1," \t"))!=NULL) {
	  sfnum[i++] = atoi(tp);
	  while ((tp = strtok((char *)NULL," \t")) != (char *)NULL) {
	    if (isdigit(*tp)) sfnum[i++] = atoi(tp);
	    if (i>=9) break;
	  }
	}
	sfnum[nsfnum=i]= 0;
	if (nsfnum>1) sf_sort(sfnum,nsfnum);
	else {
	  if (nsfnum<1) fprintf(stderr," found | but no sfnum: %s\n",libstr);
	}
      }
    }
    else {
      sfnum[0] = nsfnum = 0;
      }
#else
    sfnum[0]=nsfnum = 0;
#endif

#ifdef PCOMPLIB
    strncpy(libstr,lline+1,255);
    libstr[255]='\0';
#else
    strncpy(libstr,lline+1,20);
    libstr[12]='\0';
    if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
#endif
    if ((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';

    *libpos = lpos;

    /* make certain we have the end of the line */
    while (strchr((char *)lline,'\n')==NULL) {
      if (strlen(lline)<MAXLINE/2) 
	fgets(&lline[strlen(lline)],MAXLINE/2,libf);
      else 
      fgets(&lline[MAXLINE/2],MAXLINE/2,libf);
    }
    lline[MAXLINE-1]='\0';
  }

  lline[0]='\0';
  while (seqp<seqm1 && fgets((char *)seqp,(size_t)(seqm-seqp),libf)!=NULL) {
    if (*seqp=='>') goto new;
    if (*seqp==';') {
      if (strchr((char *)seqp,'\n')==NULL) goto cont;
      continue;
    }
    for (cp=seqp; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
    if (*seqp==ES) goto done;
    lpos = FTELL(libf);
  }
  goto done;
new:
  strncpy(lline,(char *)seqp,MAXLINE);
  lline[MAXLINE]='\0';
  /* be certain to get complete line, if possible */
  if (strchr(lline,'\n')==NULL)
    fgets(&lline[strlen(lline)],MAXLINE-strlen(lline),libf);
  lline[MAXLINE]='\0';
  if (strchr(lline,'\n')==NULL && strchr((char *)seqp,'\n')!=NULL)
    lline[strlen(lline)-1]='\n';
goto done;

cont:
  fgets(lline,MAXLINE,libf);
  seqm1 = seqp;
done:
  if (seqp>=seqm1) (*lcont)++;
  else {
#ifdef NOLIB
    leof = 1;
#endif
    *lcont=0;
  }

  *seqp = EOSEQ;
  /*  if ((int)(seqp-seq)==0) return 1; */
  return (int)(seqp-seq);
}

void
RANLIB(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp;
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);

  if (lline[0]=='>' || lline[0]==';') {
    strncpy(str,lline+1,cnt);
    str[cnt-1]='\0';
    /*
    if ((bp = strchr(str,SFCHAR))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\001'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';
    */
    bp = str;
    while ((bp = strchr(str,'\001'))!=NULL ||
	   (bp = strchr(str,'\t'))!=NULL) *bp=' ';
    if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
    if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';
  }
  else {
    str[0]='\0';
  }
#ifdef NOLIB
  leof=0;
#endif
}

#ifndef NOLIB
unsigned char *cpsave;

lgetlib(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     FSEEK_T *libpos;
     int *lcont;
{
  int i, n, ll;
  int ic;
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *linep, *bp;
  char desc[128], acc[128], ver[128];

  seqp = seq;
  seqm = &seq[maxs-11];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='L' || lline[1]!='O' || 
	   strncmp(lline,"LOCUS",5)) { /* find LOCUS */
      lpos = FTELL(libf);
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      if (lfflag) getc(libf);
    }
    *libpos=lpos;
    strncpy(libstr,&lline[12],12);
    libstr[12]='\0';

#ifdef PCOMPLIB
    if ((bp=strchr(libstr,' '))!=NULL) *(bp+1) = '\0';
    fgets(desc,sizeof(desc),libf);
    while (desc[0]!='D' || desc[1]!='E' || strncmp(desc,"DEFINITION",10))
      fgets(desc,sizeof(desc),libf);
    if ((bp = strchr(&desc[12],'\n'))!=NULL) *bp='\0';
    /* get the accession line */
    fgets(acc,sizeof(acc),libf);
    while (acc[0]!='A' || acc[1]!='C' || strncmp(acc,"ACCESSION",9))
      fgets(acc,sizeof(acc),libf);
    if ((bp = strchr(&acc[12],'\n'))!=NULL) *bp='\0';
    if ((bp = strchr(&acc[12],' '))!=NULL) *bp='\0';

    /*
      fgets(ver,sizeof(ver),libf);
      while (ver[0]!='V' || ver[1]!='E' || strncmp(acc,"VERSION",7))
      fgets(ver,sizeof(ver),libf);
      if ((bp = strchr(&ver[12],'\n'))!=NULL) *bp='\0';
      bp_gid = strchr(&ver[12],':');
      if (bp_gid != NULL) {
      if ((bp=strchr(bp_gid,' '))!=NULL) *bp='\0';
    */    

    /* add accession to locus */
    if ((bp=strchr(&desc[12],' '))!=NULL) {
      *bp = '\0';
      if (strcmp(&acc[12],&desc[12])==0) {
	*bp = ' ';
	strncat(libstr,&desc[11],255-strlen(libstr));
      }
      else {
	*bp = ' ';
	strncat(libstr,&acc[11],255-strlen(libstr));
	strncat(libstr,&desc[11],255-strlen(libstr));
      }
    }
    else {
      strncat(libstr,&acc[11],255-strlen(libstr));
      strncat(libstr,&desc[11],255-strlen(libstr));
    }    
    libstr[255]='\0';
    fgets(lline,MAXLINE,libf);
#endif

    while (lline[0]!='O' || lline[1]!='R' ||
	   strncmp(lline,"ORIGIN",6)) { /* find ORIGIN */
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      if (lfflag) getc(libf);
    }
  }
  else {
    for (cp= cpsave; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
  }

  lline[0]='\0';
  while (seqp<seqm1 && fgets(lline,MAXLINE,libf)!=NULL) {
    if (lfflag) getc(libf);
    if (lline[0]=='/') goto new;
    for (cp= (unsigned char *)&lline[10]; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
  }
  goto done;
new:
  lpos = FTELL(libf);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);

done:
  if (seqp>=seqm1) {
    cpsave = cp;
    (*lcont)++;
  }
  else *lcont=0;

  *seqp = EOSEQ;
  /*  if ((int)(seqp-seq)==0) return 1; */
  return (int)(seqp-seq);
}

/* this code seeks to provide both the various accession numbers
   necessary to identify the sequence, and also some description.

   Unfortunately, the various contributors to Genbank use three
   slightly different formats for including the accession number.

(1)LOCUS       HSJ214M20  107422 bp    DNA             HTG       16-JUN-2000
   DEFINITION  Homo sapiens chromosome 6 clone RP1-214M20 map p12.1-12.3, ***
               SEQUENCING IN PROGRESS ***, in unordered pieces.
   ACCESSION   AL121969

(2)LOCUS       AL359201   117444 bp    DNA             HTG       15-JUN-2000
   DEFINITION  Homo sapiens chromosome 1 clone RP4-671C13 map p13.2-21.1, ***
               SEQUENCING IN PROGRESS ***, in unordered pieces.
   ACCESSION   AL359201

(3)LOCUS       BB067000      280 bp    mRNA            EST       19-JUN-2000
   DEFINITION  BB067000 RIKEN full-length enriched, 15 days embryo male testis Mus
               musculus cDNA clone 8030456L01 3', mRNA sequence.
   ACCESSION   BB067000

This makes it more difficult to both provide the accession number in a
standard location and to conserve definition space
*/

void
lranlib(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp, acc[MAXLINE], desc[MAXLINE];
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);

  /* genbank fields start at 12 */
  strncpy(str,&lline[12],12);
  if ((bp=strchr(str,' '))!=NULL) *bp = '\0';
  str[12]='\0';
  fgets(desc,MAXLINE,libf);
  if (lfflag) getc(libf);
  while (desc[0]!='D' || desc[1]!='E' || strncmp(desc,"DEFINITION",10))
    fgets(desc,MAXLINE,libf);

  if ((bp = strchr(&desc[12],'\n'))!=NULL) *bp='\0';

  /* get the accession line */
  strncpy(acc,desc,sizeof(acc));
  while (acc[0]!='A' || acc[1]!='C' || strncmp(acc,"ACCESSION",9))
    fgets(acc,MAXLINE,libf);
  if ((bp = strchr(&acc[12],'\n'))!=NULL) *bp='\0';
  if ((bp = strchr(&acc[12],' '))!=NULL) *bp='\0';

  /* add accession to locus  - copy from &[11] to add space */
  if ((bp=strchr(&desc[12],' '))!=NULL) {
    *bp = '\0';
    if (strcmp(&acc[12],&desc[12])==0) {
      *bp = ' ';
      strncat(str,&desc[11],cnt-strlen(str));
    }
    else {
      *bp = ' ';
      strncat(str,&acc[11],cnt-strlen(str));
      strncat(str,&desc[11],cnt-strlen(str));
    }
  }
  else {
    strncat(str,&acc[11],cnt-strlen(str));
    strncat(str,&desc[11],cnt-strlen(str));
  }    
  str[cnt-1]='\0';

  FSEEK(libf,seek,0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);
}

pgetlib(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     FSEEK_T *libpos;
     int *lcont;
{
  int i, n, ll;
  int ic;
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *linep;

  seqp = seq;
  seqm = &seq[maxs-11];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='E' || lline[1]!='N' || strncmp(lline,"ENTRY",5))
      { /* find ENTRY */
	lpos = FTELL(libf);
	if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      }
    strncpy(libstr,&lline[16],8);
    libstr[8]='\0';
    *libpos = lpos;
    while (lline[2]!='Q' || lline[0]!='S' || strncmp(lline,"SEQUENCE",8))
      { /* find SEQUENCE */
	if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      }
    fgets(lline,MAXLINE,libf); /* get the extra line */
  }
  else {
    for (cp= cpsave; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
    if (*seqp==ES) goto done;
  }

  lline[0]='\0';
  while (seqp<seqm1 && fgets(lline,MAXLINE,libf)!=NULL) {
    if (lline[0]=='/') goto new;
    for (cp= (unsigned char *)&lline[8]; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    };
    if (*seqp==ES) goto done;
  }
  goto done;
new:
  lpos = FTELL(libf);
  fgets(lline,MAXLINE,libf);

done:
  if (seqp>=seqm1) {
    cpsave = cp;
    (*lcont)++;
  }
  else *lcont=0;

  *seqp = EOSEQ;
  /*  if ((int)(seqp-seq)==0) return 1; */
  return (int)(seqp-seq);
}

void
pranlib(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp;
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);

  strncpy(str,&lline[16],8);
  str[8]='\0';
  fgets(lline,MAXLINE,libf);
  while (lline[0]!='T' || lline[1]!='I' || strncmp(lline,"TITLE",5))
    fgets(lline,MAXLINE,libf);
  strncpy(&str[8],&lline[16],cnt-9);
  str[cnt-1]='\0';
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';

  FSEEK(libf,seek,0);
  fgets(lline,MAXLINE,libf);
}

long seqsiz;

egetlib(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     FSEEK_T *libpos;
     int *lcont;
{
  int ll;
  int ic;
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *linep;
  char id[11];  /* Holds Identifier */

  seqp = seq;
  seqm = &seq[maxs-11];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='I' || lline[1]!='D') { /* find ID */
      lpos = FTELL(libf);
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      if (lfflag) getc(libf);
    }
    sscanf(&lline[5],"%s",id);
    sprintf(libstr,"%-12.12s",id);
    libstr[12]='\0';
    *libpos = lpos;
    while (lline[0]!='S' || lline[1]!='Q') { /* find ORIGIN */
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      if (lfflag) getc(libf);
    }
    sscanf(&lline[14],"%ld",&seqsiz);
  }
  else {
    for (cp= cpsave; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
    if (*seqp==ES) goto done;
  }

  lline[0]='\0';
  while (seqp<seqm1 && fgets(lline,MAXLINE,libf)!=NULL) {
    if (lfflag) getc(libf);
    if (lline[0]=='/') goto new;
    lline[70]='\0';
    for (cp= (unsigned char *)&lline[5]; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
    if (*seqp==ES) goto done;
  }
  goto done;
new:	lpos = FTELL(libf);
fgets(lline,MAXLINE,libf);
if (lfflag) getc(libf);
goto done;

done:	if (seqp>=seqm1) {
  cpsave = cp;
  (*lcont)++;
  seqsiz -= (long)(seqp-seq);
}
else *lcont=0;

*seqp = EOSEQ;
/* if ((int)(seqp-seq)==0) return 1; */
/*	if (*lcont==0 && (long)(seqp-seq)!=seqsiz)
	printf("%s read %d of %d\n",libstr,(int)(seqp-seq),seqsiz);
	*/
return (int)(seqp-seq);
}

void
eranlib(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp;
  char id[11];  /* Holds Identifier */
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);

  sscanf(&lline[5],"%s",id);
  sprintf(str,"%-10.10s ",id);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);
  while (lline[0]!='D' || lline[1]!='E') fgets(lline,MAXLINE,libf);
  strncpy(&str[11],&lline[5],cnt-11);
  str[cnt-1]='\0';
  if ((bp = strchr(str,'\r'))!=NULL) *bp='\0';
  if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';

  FSEEK(libf,seek,0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);
}

igetlib(seq,maxs,libstr,libpos,lcont)
	unsigned char *seq;
	int maxs;
	char *libstr;
	FSEEK_T *libpos;
	int *lcont;
{
	int i, n, ll;
	int ic;
	register unsigned char *cp, *seqp;
	register int *ap;
	unsigned char *seqm, *seqm1;
	char *linep, *bp;

	seqp = seq;
	seqm = &seq[maxs-9];
	seqm1 = seqm-1;
#if !defined(TFAST)
	ap = sascii;
#else
	ap = nascii;
#endif
	if (*lcont==0) {
		while (lline[0]!=';') {
			lpos = FTELL(libf);
			if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
			}
		*libpos = lpos;
		while (lline[0]==';') fgets(lline,MAXLINE,libf);
		strncpy(libstr,lline+1,12);
		libstr[12]='\0';
		if((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';
		}

	lline[0]='\0';
	while (seqp<seqm1 && fgets((char *)seqp,(size_t)(seqm-seqp),libf)!=NULL) {
		if (*seqp=='>') goto new;
		if (*seqp==';') {
			if (strchr((char *)seqp,'\n')==NULL) goto cont;
			continue;
			}
		for (cp=seqp; seqp<seqm1; ) {
			if ((*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
		            (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA &&
			    (*seqp++=ap[*cp++])<NA) continue;
			    if (*(--seqp)>NA) break;
			    }
		if (*seqp==ES) goto done;
		lpos = FTELL(libf);
		}
	goto done;
new:	strncpy(lline,(char *)seqp,MAXLINE);
	lline[MAXLINE-1]='\0';
	if (strchr((char *)seqp,'\n')==NULL)
	    fgets(lline,MAXLINE-strlen(lline),libf);
	goto done;

cont:
	fgets(lline,MAXLINE,libf);
	seqm1 = seqp;

done:	if (seqp>=seqm1) {
		(*lcont)++;
		}
	else {
	*lcont=0;
		}


	*seqp = EOSEQ;
	/*	if ((int)(seqp-seq)==0) return 1; */
	return (int)(seqp-seq);
	}

void
iranlib(str,cnt,seek)
	char *str; int cnt; FSEEK_T seek;
{
	char *bp;
	int ll;
	char tline[MAX_FN];

	FSEEK(libf, seek, 0);
	fgets(lline,MAXLINE,libf);

	if (lline[0]=='>' || lline[0]==';') {
		strncpy(tline,lline+1,sizeof(tline));
		tline[sizeof(tline)-1]='\0';
		if ((bp = strchr(tline,'\n'))!=NULL) *bp='\0';
		}
	else {
		tline[0]='\0';
		}

	while (lline[0]==';') fgets(lline,MAXLINE,libf);
	if ((bp=strchr(lline,'\n'))!=NULL) *bp=0;
	if ((bp=strchr(lline,' '))!=NULL) *bp=0;
	strncpy(str,lline,cnt);
	strncat(str,"  ",cnt-strlen(str));
	strncat(str,tline,cnt-strlen(str));
	str[cnt-1]='\0';
	
	FSEEK(libf,seek,0);
	fgets(lline,MAXLINE,libf);
	}

vgetlib(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     FSEEK_T *libpos;
     int *lcont;
{
  int i, n, ll;
  int ic, ich;
  register unsigned char *cp, *seqp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  char *linep, *bp, *tp;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='>' && lline[0]!=';') {
      lpos = FTELL(libf);
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
      if (lfflag) getc(libf);
    }

#ifdef SUPERFAMNUM
    if ((bp=strchr(&lline[1],' ')) &&
	 (bp=strchr(bp+1,SFCHAR))) {
      i=0;
      if ((tp = strtok(bp+1," \t\n"))!=NULL) sfnum[i++] = atoi(tp);
      while ((tp = strtok(NULL," \t")) != NULL) {
	sfnum[i++] = atoi(tp);
	if (i>=10) break;
      }
      sfnum[nsfnum=i]= 0;
      if (nsfnum>1) sf_sort(sfnum,nsfnum);
      else {
	if (nsfnum < 1) fprintf(stderr," found | but no sfnum: %s\n",libstr);
      }
    }
    else sfnum[0]=nsfnum=0;
#else
    sfnum[0]=nsfnum=0;
#endif

    if ((bp=strchr(lline,'\n'))!=NULL) *bp='\0';
    strncpy(libstr,&lline[4],12);
    libstr[12]='\0';
    if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
    if ((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';
    
    fgets(lline,MAXLINE,libf);
    if (lfflag) getc(libf);
#ifdef PCOMPLIB
    strcat(libstr," ");
    strncat(libstr,lline,255-strlen(libstr));
    if ((bp=strchr(libstr,'\n'))!=NULL) *bp='\0';
    libstr[255]='\0';
#endif
    *libpos = lpos;
  }

  lline[0]='\0';
  while (seqp<seqm1 && fgets((char *)seqp,(size_t)(seqm-seqp),libf)!=NULL) {
    if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
    if (*seqp=='>') goto new;
    if (*seqp==';') {
      if (strchr((char *)seqp,'\n')==NULL) goto cont;
      continue;
    }
    for (cp=seqp; seqp<seqm1; ) {
      if ((*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA &&
	  (*seqp++=ap[*cp++])<NA) continue;
      if (*(--seqp)>NA) break;
    }
    if (*seqp==ES) goto done;
    lpos = FTELL(libf);
  }
  goto done;
new:
  strncpy(lline,(char *)seqp,MAXLINE);
  lline[MAXLINE-1]='\0';
  if (strchr((char *)seqp,'\n')==NULL) {
    fgets(&lline[strlen(lline)],MAXLINE-strlen(lline),libf);
    if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
  }
  goto done;

cont:
  fgets(lline,MAXLINE,libf);
  if (lfflag && (ich=getc(libf))!=LFCHAR) ungetc(ich,libf);
  seqm1 = seqp;

done:
  if (seqp>=seqm1) {
    (*lcont)++;
  }
  else {
    *lcont=0;
  }

  *seqp = EOSEQ;
  /*   if ((int)(seqp-seq)==0) return 1;*/
  return (int)(seqp-seq);
}

void
vranlib(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp, *llp;
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);

  if (lline[0]=='>'&&(lline[3]==';'||lline[3]=='>')) {
    strncpy(str,&lline[4],cnt);
    str[cnt-1]='\0';

    if ((bp = strchr(str,':'))!=NULL) *bp='\0';
    if ((bp=strchr(str,'\r'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';

    fgets(lline,MAXLINE,libf);
    if (lfflag) getc(libf);

    /* skip over redundant stuff */
    for (llp=lline,bp=str; *llp==*bp; llp++,bp++);
    if ((int)(llp-lline)<5) llp = lline;

    if ((bp=strchr(llp,'\r'))!=NULL) *bp=' ';
    if ((bp=strchr(llp,'\n'))!=NULL) *bp='\0';
    strncat(str," ",(size_t)cnt);
    strncat(str,llp,(size_t)cnt-strlen(str));
  }
  else {
    str[0]='\0';
  }

  FSEEK(libf,seek,0);
  fgets(lline,MAXLINE,libf);
  if (lfflag) getc(libf);
}

static char gcg_type[10];
static long gcg_len;
static int gcg_bton[4]={2,4,1,3};

gcg_getlib(seq,maxs,libstr,libpos,lcont)
     unsigned char *seq;
     int maxs;
     char *libstr;
     FSEEK_T *libpos;
     int *lcont;
{
  char dummy[20];
  char gcg_date[6];
  int i, n, ll;
  int ic, ich;
  register unsigned char *cp, *seqp, stmp;
  register int *ap;
  unsigned char *seqm, *seqm1;
  long r_block, b_block;
  char *linep, *bp;

  seqp = seq;
  seqm = &seq[maxs-9];
  seqm1 = seqm-1;
#if !defined(TFAST)
  ap = sascii;
#else
  ap = nascii;
#endif
  if (*lcont==0) {
    while (lline[0]!='>' && lline[0]!=';') {
      lpos = FTELL(libf);
      if (fgets(lline,MAXLINE,libf)==NULL) return (-1);
    }
    sscanf(&lline[4],"%s %s %s %s %ld",
	   libstr,gcg_date,gcg_type,dummy,&gcg_len);

    fgets(lline,MAXLINE,libf);
    while (strchr((char *)lline,'\n')==NULL) {
      if (strlen(lline)<MAXLINE/2) 
	fgets(&lline[strlen(lline)],MAXLINE/2,libf);
      else 
      fgets(&lline[strlen(lline)-MAXLINE/2],MAXLINE/2,libf);
    }
    lline[MAXLINE-1]='\0';
#ifndef PCOMPLIB
    libstr[12]='\0';
#else
    strncat(libstr," ",1);
    strncat(libstr,lline,255-strlen(libstr));
    if ((bp = strchr(libstr,'\n'))!=NULL) *bp='\0';
    libstr[255]='\0';
#endif
    *libpos = lpos;
  }

  lline[0]='\0';

  r_block = b_block = min((size_t)(seqm-seqp),gcg_len);
  if (gcg_type[0]=='2') {
    r_block = (r_block+3)/4;
  }

  fread((char *)seqp,(size_t)r_block,(size_t)1,libf);
  if (gcg_type[0]=='A') 
    for (cp=seqp; seqp<seq+r_block; ) *seqp++ = ap[*cp++];
  else if (gcg_type[0]=='2') {
    seqp = seq + r_block;
    cp = seq + 4*r_block;
    while (seqp > seq) {
      stmp = *--seqp;
      *--cp = gcg_bton[stmp&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
      *--cp = gcg_bton[(stmp >>= 2)&3];
    }
  }
  if (4 * r_block >= gcg_len) {
    fgets(lline,MAXLINE,libf);
    *lcont = 0;
  }
  else {
    if (gcg_type[0]=='2') b_block = 4*r_block;
    gcg_len -= b_block;
    (*lcont)++;
  }

  seq[b_block] = EOSEQ;
  /*   if (b_block==0) return 1; else */
  return b_block;
}

void
gcg_ranlib(str,cnt,seek)
     char *str; int cnt; FSEEK_T seek;
{
  char *bp, *bp1, *bp2, *llp;
  int ll;

  FSEEK(libf, seek, 0);
  fgets(lline,MAXLINE,libf);

  if (lline[0]=='>'&&(lline[3]==';'||lline[3]=='>')) {
    strncpy(str,&lline[4],cnt);
    str[cnt-1]='\0';
    if ((bp = strchr(str,' '))!=NULL) *bp='\0';
    else if ((bp=strchr(str,'\r'))!=NULL) *bp='\0';
    else if ((bp = strchr(str,'\n'))!=NULL) *bp='\0';
    else str[cnt-1]='\0';

    fgets(lline,MAXLINE,libf);

    /* check beginning of line it is a duplicate */
    for (llp=lline,bp=str; *llp == *bp; llp++,bp++);
    if ((int)(llp-lline)<5) llp = lline;

    /* here we would like to skip over some species stuff */
	/*
    if ((bp1 = strchr(llp,';'))!=NULL && (int)(bp1-llp)<50) {
      if ((bp2 = strchr(bp1+1,';'))!=NULL && (int)(bp2-bp1)<50) {
	*(bp2+1)='\0'; bp1 = bp2+2;
      }
      else {bp1=llp;}
    }
    else if ((bp1=strchr(llp,'.'))!=NULL && *(bp1+1)==' ') {
      *(bp1+1) = '\0'; bp1 += 2;}
    else bp1 = llp;
    */
    
    bp1 = llp;
    if ((bp=strchr(bp1,'\r'))!=NULL) *bp='\0';
    if ((bp=strchr(bp1,'\n'))!=NULL) *bp='\0';
    strncat(str," ",(size_t)cnt);
    strncat(str,bp1,(size_t)cnt-strlen(str));
    if (bp1!=llp) strncat(str,llp,(size_t)cnt-strlen(str));
  }
  else {
    str[0]='\0';
  }

  FSEEK(libf,seek,0);
  fgets(lline,MAXLINE,libf);
}

#endif	/* NOLIB */

int scanseq(unsigned char *seq, int n, char *str)
{
	int tot,i;
	char aaray[MAXSQ];		/* this must be set > nsq */
	
	for (i=0; i<MAXSQ; i++)  aaray[i]=0;
	for (i=0; (size_t)i < strlen(str); i++) aaray[sascii[str[i]]]=1;
	for (i=tot=0; i<n; i++) tot += aaray[seq[i]];
	return tot;
	}

#ifndef MPI_SRC
void
revcomp(seq,n)
	unsigned char *seq; int n;
{
  unsigned char tmp;
  int i, ni;

  for (i=0; i< n; i++) {
    if (nt[seq[i]]=='A') seq[i] = nascii['T'];
    else if (nt[seq[i]]=='C') seq[i] = nascii['G'];
    else if (nt[seq[i]]=='G') seq[i] = nascii['C'];
    else if (nt[seq[i]]=='T') seq[i] = nascii['A'];
    else if (nt[seq[i]]=='R') seq[i] = nascii['Y'];
    else if (nt[seq[i]]=='Y') seq[i] = nascii['R'];
    else if (nt[seq[i]]=='M') seq[i] = nascii['K'];
    else if (nt[seq[i]]=='K') seq[i] = nascii['M'];
    else if (nt[seq[i]]=='D') seq[i] = nascii['H'];
    else if (nt[seq[i]]=='H') seq[i] = nascii['D'];
    else if (nt[seq[i]]=='V') seq[i] = nascii['B'];
    else if (nt[seq[i]]=='B') seq[i] = nascii['V'];
  }

  for (i=0, ni = n-1; i< n/2; i++,ni--) {
    tmp = seq[i];
    seq[i] = seq[ni];
    seq[ni] = tmp;
  }
}
#endif

void
sf_sort(s,n)
     int *s, n;
{
  int gap, i, j;
  int itmp;

  if (n == 1) return;
	
  for (i=0; i<n-1; i++)
    if (s[i]>s[i+1]) goto l2;
  return;

l2:
  for (gap=n/2; gap>0; gap/=2)
    for (i=gap; i<n; i++)
      for (j=i-gap; j>=0; j -= gap) {
	if (s[j] <= s[j+gap]) break;
	itmp = s[j];
	s[j]=s[j+gap];
	s[j+gap]=itmp;
      }
}
