/* copyright (c) 1996, 1997, 1998, 1999 William R. Pearson and the
   U. of Virginia */

/* $Name: fa33t09d0 $ - $Id: p_complib.c,v 1.12.2.1.2.1 2001/03/29 22:29:44 wrp Exp $ */

/*
 * Concurrent read version
 * pcomplib.c : Parallel library search
 * 
 *	#define FIRSTNODE 0/1 (in msg.h) can be used to reserve one node
 *	for collecting results
 *
 * Parallel specific options (from doinit.c):
 *	-J # jump to query #
 *	-I   self-comparison, do (N choose 2) comparisons
 *	-T # number of workers
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#ifdef UNIX
#include <values.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#else
#define MAXFLOAT 1e+37
#endif

#include <math.h>

#ifdef PVM_SRC
#include "pvm3.h"
char *mp_verstr="v3.36, June 12, 2000 PVM";
#endif

#ifdef MPI_SRC
#include "mpi.h"
char *mp_verstr="v3.36, June 12, 2000 MPI";
#endif

#include "msg.h"
#include "defs.h"
#include "param.h"
#include "p_mw.h"
#include "structs.h"

char pgmdir[120];
char workerpgm[120];
char managepgm[120];

#define XTERNAL
#include "upam.h"
#undef XTERNAL
#include "uascii.h"

/********************************/
/* global variable declarations */
/********************************/
char gstring2[MAX_STR];                  /* string for label */
char gstring3[MAX_STR];                  /* string for label */
char hstring1[MAX_STR];
int sfnum;

/********************************/
/* extern variable declarations */
/********************************/
extern char *prog_func;		/* function label */
extern char *verstr;

/********************************/
/*extern function declarations  */
/********************************/
/********************************/

/* open sequence file (pxgetaa.c) */
extern int h_openlib(char *lname, char *libenv, int ldnaseq);
/* check for DNA sequence (nxgetaa.c) */
extern int scanseq(unsigned char *seq, int n, char *str);
/* 1d to 2d pam (initxx.c) */
extern void initpam2 (struct pstruct *ppst);
/* initialize environment (doinit.c) */
extern void initenv (int argc, char **argv, struct mngmsg *m_msg,
		     struct pstruct *ppst, unsigned char **aa0, int outtty);

/* print hist, summaries, timing information */
void prhist(FILE *, struct mngmsg, struct pstruct, char *);
void printsum(FILE *);
extern void ptime (FILE *, time_t);

/* reset parameters if DNA sequence (initxx.c) */
#ifndef FASTX
extern void resetp (unsigned char *, int, struct mngmsg *, struct pstruct *);
#else
extern void reseta (unsigned char *, int, struct mngmsg *, struct pstruct *);
#endif
/* read a sequence (nxgetaa.c) */
#ifndef FASTX
extern int (*h_getlib) (unsigned char *seq, int maxs,
		      char *libstr, long *libpos, int *lcont);
#define HGETLIB (*h_getlib)
#else
extern int (*h_getntlib) (unsigned char *seq, int maxs,
			  char *libstr, long *libpos, int *lcont);
#define HGETLIB (*h_getntlib)
#endif
void h_closelib();		/* close the query library */
extern void reseq (long);	/* reset file pointer if DNA sequence */


/* these functions are in scaleswn.c */
extern void process_hist(struct stat_str *sptr, int nstat, struct pstruct pst);
extern double zs_to_E(float zs, int n1, int isdna);
extern float (*find_zp)(int score, int length, double comp); /* does not use var. corr. */
extern float find_z(int score, int length, double comp);
void addhistz(float zscore);	/* scaleswn.c */

void selectbestz(struct beststr **, int, int);
void sortbestz(struct beststr **, int);
void sortbeste(struct beststr **, int);
void sortbest(struct beststr **, int, int);

void showbest (FILE *fp, struct beststr **bptr, int nbest,
	       int qlib, struct mngmsg *m_msg, struct pstruct pst,
	       char *gstring2);

void showalign (FILE *fp, unsigned char *aa0, 
		struct beststr **bptr, int nbest,int qlib, struct mngmsg m_msg,
		struct pstruct pst, char *gstring2);


#ifdef PVM_SRC
char worknode[120];
int pinums[MAXNOD],hosttid;
int narch;
struct pvmhostinfo *hostp;
#endif

FILE *outfd;			/* Output file */
int outtty=1;

extern time_t s_time ();                 /* fetches time for timing */

/* this information is global for fsigint() */
time_t tstart, tscan, tprev, tdone;	/* Timing */
time_t tdstart, tddone, time();
int max_nodes, nnodes;			/* number of nodes */
int node_map[MAXWRKR];
int tot_speed;
int  nlib = 0, nclib=0, qlib = 0;	/* number of sequences scanned */
unsigned long ntt  = 0l, qtt = 0l;	/* number of residues scanned */
int ntt_c=0, qtt_c=0;
int nls_tot0, nls_cnt0;

extern int max_workers;
int  wlsn [MAXWRKR + 1];	/* number of library sequences in worker */
int  clsn [MAXWRKR + 1];	/* number of 1st library sequence in worker */
int  slsn [MAXWRKR + 1];	/* correction for continuation lines */

int  wstage2[MAXWRKR +1];	/* count of second stage scores */
struct stage2_str  *wstage2p[MAXWRKR+1]; /* list of second stage sequences */
int  wstage2i[MAXWRKR+1];	/* index into second stage sequences */
struct beststr *bbptr,
     **wstage2b[MAXWRKR+1];	/* reverse pointers to bestr */

struct sql *ldes [MAXWRKR + 1];/* descriptive lines for all lib sequences */
char *bline_buf, *bline_buf_mx, *bline_bufp;
int bline_inc, max_bline_b;

main (argc, argv)
int argc;
char *argv [];
{
  unsigned char *aa00, *aa01, *aa0p0, *aa0p1;
  char nodefile[240];
  struct pstruct pst;
  int i_score, r_score;

  /* from manage code */
  struct mngmsg m_msg0;	/* Message from host to manager */
  struct qmng_str qm_msg0, qm_msg1;	/* stuff updated for each query */
  struct qmng_str *qm_msp0, *qm_msp1;	/* pointer to stuff updated */
  struct wrkmsg s_msg;	/* worker start msg */
  int last_msg_b[10];	/* last set of numbers */
  long curtype = ONETYPE;	/* current message type */
  struct stat stbuf;
  long lbsiz;			/* Library size */
  char libn [120];		/* full pathname of library */
  struct beststr *best,		/* array of best scores */
                 **bptr;	/* array of pointers */
  struct comstr bestr[BFR+1];	/* temporary structure array */
  struct comstr2 bestr2[BFR2+1];	/* temporary structure array */
  struct a_struct *aln_d_base=NULL;	/* alignment info for -m 9 */
  int qres_bufsize;		/* buffer size for results */
  struct sqlc seqds [NDES];/* sequence descriptions */
  struct stat_str *stats;
  int best_flag = 1;		/* bptr[] must be re-initialized */
  int fast_flag = 0;		/* send new sequences before old displayed */
  int nstats;
  int nbest, nres;		/* number of best scores */
  long parcelsiz;		/* size of a node chunk */
  float zbestcut = -BIGNUM;	/* z-value cutoff */
  int lcnt, ldcnt;		/* counters */
  int nls_cnt, nls_tot, nopt;
  int i, j, k, is, iw, ires, naa0 = 0;

  FILE *fdata=NULL;		/* file for full results */
  struct sql *desptr;
  char libstr[11];		/* string for labeling full results */
  
  int node, snode;		/* Number of nodes */
  int bufid, numt, tid;

  int leng,type;		/* leng is length of the descriptive line */
  int maxn = MAXTST;		/* max size of the query sequence */
  long lmark;			/* seek into library of current sequence */
  int lcont;			/* continued sequence */
  int var;			/* dummy variable for message */
  int n_proc, t_proc, n_tmp;
  char errstr[120];
  int dohist=0;			/* flag for z-value processing */
  int tm_best, t_rbest, t_best, t_n1;
  float zscore;			/* tmp value */
  double k_H, k_comp;
  char *bp;
  char stat_string[MAX_STR];
#ifdef MPI_SRC
  MPI_Status mpi_status;
#endif

  void fsigint();
  
  signal(SIGHUP,SIG_IGN);
  if (signal(SIGINT,SIG_IGN) != SIG_IGN) signal(SIGINT,fsigint);
  if (signal(SIGQUIT,SIG_IGN) != SIG_IGN) signal(SIGQUIT,fsigint);
/*  if (signal(SIGSEGV,SIG_IGN) != SIG_IGN) signal(SIGSEGV,fsigint); */

  /* Initialization */

#ifdef UNIX
  outtty=isatty(1);
#endif

  /* BFR must be %6 = 0 for TFASTA */
  if ((BFR%6) != 0) {
    fprintf(stderr," BFR size %d not %%6=0 - recompile\n",BFR);
    exit(1);
  }

#ifdef MPI_SRC
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&tid);
  if (tid > 0) {
    workcomp(tid); 
    MPI_Finalize();
    exit(0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&nnodes);
  if (nnodes <= 1) {
    fprintf(stderr," nnodes = %d; no workers available\n",nnodes);
    exit(1);
  }
  else fprintf(stderr," have %d nodes\n",nnodes);

  tot_speed = nnodes*100;
#endif

  h_init (&pst);
  
#ifdef PVM_SRC
  strcpy (workerpgm, WORKERPGM);
#endif
  
  initenv (argc, argv, &m_msg0, &pst, &aa00, outtty);
  fprintf (stderr, "Pcomp library processor\n");
  fprintf (stderr, "Using %s [%s]\n", prog_func, verstr);
  
  tstart = tscan = s_time();
  tdstart = time(NULL);
  

#ifdef PVM_SRC
  if ((hosttid=pvm_mytid())<0) {
    pvm_perror("initialization");
    fprintf(stderr,"can't initialize %s\n", argv[0]);
    pvm_exit();
    exit(1);
  }
  
  pvm_config(&nnodes,&narch,&hostp);
  fprintf(stderr,"nnodes: %d, narch: %d\n",nnodes, narch);
  max_nodes = nnodes;

/*  if (nnodes < 2 ) nnodes = 4; */
  if (max_workers > 0  && nnodes > max_workers) {
    nnodes = max_workers+FIRSTNODE;
    fprintf(stderr," workers reset from %d to %d\n",
	    max_nodes,nnodes-FIRSTNODE);
  }
  
  strcpy(nodefile,pgmdir);
  strcat(nodefile,workerpgm);
  
  for (i=0; i< nnodes; i++) node_map[i] = i;

  if (nnodes < max_nodes) {
    hostp++;	/* bump over host name for spawn */
    rand_nodes(node_map,nnodes,max_nodes-1);
    for (i=0; i<nnodes; i++) {
      numt+=pvm_spawn(nodefile,NULL,PvmTaskHost,hostp[node_map[i]].hi_name,
		      1,&pinums[i]);
    }
  }
  else {
    for (i=j=0; i<nnodes && j < MAXWRKR; i++) {
      n_proc = hostp[i].hi_speed%100;
      if (n_proc == 0) n_proc = 1;
      if (n_proc > max_workers) n_proc = max_workers;

      n_tmp =pvm_spawn(nodefile,NULL,PvmTaskHost,hostp[i].hi_name,
		       n_proc,&pinums[j]);
      if (n_tmp < n_proc)
	fprintf(stderr," spawn problem: %d\n", pinums[j]);
      if (n_tmp > 0) {
	for (k=j; k < j+n_tmp; k++) node_map[k]=i;
	j += n_tmp;
      }
    }
    nnodes = numt = j;
  }

  if (numt < nnodes) {
    if (numt <= 0) {
      pvm_perror("");
      pvm_exit();
      exit(1);
    }
    nnodes = numt;
  }

  for (tot_speed=0,i=FIRSTNODE; i<nnodes; i++) {
    if (pinums[i]<0) {
      fprintf(stderr," tids %d %8o\n",i,pinums[i]);
      pvm_perror("");
      pvm_exit();
      exit(1);
    }
    else {
      fprintf(stderr," tids %d %8o %s %5d\n",i,pinums[i],
	      hostp[node_map[tidtonode(pinums[i])]].hi_name,
	      hostp[node_map[tidtonode(pinums[i])]].hi_speed);
      tot_speed +=(hostp[node_map[tidtonode(pinums[i])]].hi_speed/100);
    }
  }

  strncpy(worknode,nodefile,sizeof(worknode));
  fprintf (stderr, "%3d worker programs loaded from %s\n",
	   nnodes-FIRSTNODE,worknode);
#endif  

  /* need to allocate two aa0 arrays so that the old is saved for alignments */

  /* Allocate space for the query sequence */
  if ((aa00 = (unsigned char *) malloc (MAXTST * sizeof (char))) == NULL)
    s_abort ("Unable to allocate query sequence", "");
  
  if ((aa01 = (unsigned char *) malloc (MAXTST * sizeof (char))) == NULL)
    s_abort ("Unable to allocate query sequence", "");
  
  /* Query library */
  if (m_msg0.tname[0] == '\0') {
      if (m_msg0.quiet == 1) s_abort("query sequence undefined","");
      
      fprintf(stderr, "Pvcomplib [%s]\n",mp_verstr);
    l1:	fprintf (stderr, "Enter query library name: ");
      fflush  (stderr);
      if (fgets (m_msg0.tname, 80, stdin) == NULL)
	s_abort ("Unable to read query library name","");
      if ((bp=strchr(m_msg0.tname,'\n'))!=NULL) *bp='\0';
      if (m_msg0.tname[0] == '\0') goto l1;
    }
  
  /* Open query library */
  if (h_openlib (m_msg0.tname, "",m_msg0.ldnaseq) <= 0)
    s_abort ("Could not open library: ", m_msg0.tname);
  
  /* Fetch first sequence */
  lcont = 0;
  while (qlib < m_msg0.ql_off) {
    pst.n0 = qm_msg0.n0 = m_msg0.n0 = 
      HGETLIB (aa00, MAXTST, qm_msg0.libstr, &lmark, &lcont);
    qlib++;
    qtt += m_msg0.n0;
    qm_msg0.slist = 0;

    if (m_msg0.n0 <= 0)
      s_abort ("Unable to fetch sequence from library: ", m_msg0.tname);
  }

#ifndef FASTX	/* don't scan and change sascii only if !FASTX */
  if (pst.dnaseq == 0 && 
      (float) scanseq (aa00, m_msg0.n0, "ACGTXN")/(float) m_msg0.n0 > 0.85) {
      /* DNA Sequence */
      pst.dnaseq = m_msg0.ldnaseq = 1;
      sascii = nascii;
      reseq(0l);
      qm_msg0.n0 = m_msg0.n0 = pst.n0 =
	HGETLIB (aa00, MAXTST, qm_msg0.libstr, &lmark, &lcont);
      qm_msg0.slist = 0;
    }
#endif

#ifndef FASTX
  resetp (aa00, qm_msg0.n0, &m_msg0, &pst);
#else
  reseta (aa00, qm_msg0.n0, &m_msg0, &pst);
#endif

  leng = strlen (qm_msg0.libstr);
  sprintf (&qm_msg0.libstr[leng - 1], " %d %s", qm_msg0.n0, sqnam);
  qm_msg0.seqnm = qlib-1;
  
  /* Library */

  if (strlen (m_msg0.lname) == 0) {
    if (m_msg0.quiet == 1) s_abort("library name undefined","");
  l2: 	fprintf (stderr, "Library file name: ");
    fflush  (stderr);
    if (fgets (m_msg0.lname, 80, stdin) == NULL)
      s_abort ("Unable to read library name","");
    if ((bp=strchr(m_msg0.lname,'\n'))!=NULL) *bp='\0';
    if (m_msg0.lname[0] == '\0') goto l2;
  }
  
  /* Get additional parameters here */
  query_parm (&m_msg0, &pst);
  
/* a sanity check */
   if (m_msg0.revcomp && !(pst.dnaseq==1)) {
     fprintf(stderr," cannot reverse complement %s\n",sqtype);
     m_msg0.revcomp = 0;
   }

  last_init(&m_msg0, &pst);
#if !defined(TFASTA) && !defined(TFASTX)
  m_msg0.nitt1 = m_msg0.qframe-1;
#else
  m_msg0.nitt1 = m_msg0.nframe-1;
#endif
  
  if (m_msg0.maxn <= 0) m_msg0.maxn = MAXLIB;
  if (m_msg0.maxn < 2 * m_msg0.dupn) m_msg0.maxn = 5*m_msg0.dupn;
  pst.maxlen = m_msg0.maxn;


  /* ******************** */
  /* initial manager code */
  /* ******************** */
  
  outfd = stdout;
  if (m_msg0.outfile[0]!='\0') {
    if ((outfd = fopen(m_msg0.outfile,"w"))==NULL) {
      fprintf(stderr, "cannot open %s for output\n", m_msg0.outfile);
      outfd = stdout;
    }
  }

  /* Label the output */
  printf("Query library %.50s  vs  %.60s library\n",
	  m_msg0.tname, m_msg0.lname);
  if (outfd != stdout)
    fprintf (outfd, "Query library %.50s  vs  %.60s library\n",
	     m_msg0.tname, m_msg0.lname);
  fflush (stdout);
  
  /* Allocate space for saved scores */
  if ((best = 
       (struct beststr *)malloc((MAXBEST+1)*sizeof(struct beststr)))==NULL)
    s_abort ("Cannot allocate best struct","");
  if ((bptr = 
       (struct beststr **)malloc((MAXBEST+1)*sizeof(struct beststr *)))==NULL)
    s_abort ("Cannot allocate bptr","");
  
  /* Initialize bptr */
  for (nbest = 0; nbest < MAXBEST+1; nbest++)
    bptr[nbest] = &best[nbest];

  best++; bptr++;
  best[-1].score[0]=best[-1].score[1]=best[-1].score[2]=1000000000;
  best[-1].zscore=MAXFLOAT;
  best_flag = 0;
  
  if ((stats =
       (struct stat_str *)calloc((size_t)MAXSTATS,sizeof(struct stat_str)))
      ==NULL)
    s_abort ("Cannot allocate stats struct","");
  nstats = 0;

  strcpy (s_msg.libenv, "");
  /* Build the worker start message */
  if ((bp = (char *) strchr (m_msg0.lname, ' ')) != NULL) *bp = '\0';
  if (stat (m_msg0.lname, &stbuf) == -1)
    s_abort (m_msg0.lname, ": library not found");
  
  s_msg.lb_size = stbuf.st_size;
  if (bp != NULL) *bp = ' ';
  
/* fprintf(stderr," lb_size: %d; tot_speed: %d\n",s_msg.lb_size,tot_speed); */

  /* Send each worker node the start messages */
  
#ifdef PVM_SRC 
#ifdef ROUTE_DIRECT
  pvm_setopt(PvmRoute,PvmRouteDirect);
#endif
  pvm_initsend(PvmDataRaw);
  pvm_pkint(&nnodes,1,1);
  pvm_pkint(pinums,nnodes,1);
  pvm_pkbyte((char *)&m_msg0,(int)sizeof(m_msg0),1);
  for (node = FIRSTNODE; node<nnodes; node++) 
    if (pvm_send(pinums[node],STARTTYPE0)<0) {
      pvm_perror("pvm_send1");
      pvm_exit();
      exit(1);
    }
#endif

#ifdef MPI_SRC
  /* each node gets 1/nnodes library */
  s_msg.p_size = s_msg.lb_size/(nnodes-1);
#endif
  /* divide up the database and send it off */
  s_msg.lb_stop = s_msg.lb_off = 0;
  for (node = FIRSTNODE; node<nnodes; node++)  {
#ifdef MPI_SRC
    MPI_Send(&m_msg0,(int)sizeof(m_msg0),MPI_BYTE,node,STARTTYPE0,
	     MPI_COMM_WORLD);
    s_msg.lb_stop = s_msg.lb_off + s_msg.p_size;
    MPI_Send(&s_msg,(int)sizeof(s_msg),MPI_BYTE,node,STARTTYPE1,
	     MPI_COMM_WORLD);
#endif

#ifdef PVM_SRC
    pvm_initsend(PvmDataRaw);
    s_msg.p_size = (int)((float)s_msg.lb_size*
	   (float)(hostp[node_map[tidtonode(pinums[node])]].hi_speed/100)/
	   (float)tot_speed+0.5);
/*  fprintf(stderr," p_size: %d; lb_off: %d\n",  s_msg.p_size,s_msg.lb_off);
    fflush(stderr);
*/    
    s_msg.lb_stop = s_msg.lb_off + s_msg.p_size;
    pvm_pkbyte((char *)&s_msg,(int)sizeof(s_msg),1);
    if (pvm_send(pinums[node],STARTTYPE2)<0) {
      pvm_perror("pvm_send1a");
      pvm_exit();
      exit(1);
    }
#endif
    s_msg.lb_off = s_msg.lb_stop+1;
  }

  /* now send pst, sascii */
#ifdef PVM_SRC
  pvm_initsend(PvmDataRaw);
  pvm_pkbyte((char *)&pst,(int)sizeof(pst),1);
  pvm_pkbyte((char *)sascii,(int)sizeof(aascii),1);
  for (node = FIRSTNODE; node< nnodes; node++)
    pvm_send(pinums[node],STARTTYPE1);

  pvm_initsend(PvmDataRaw);
  pvm_pkint(pam12,m_msg0.pamd1*m_msg0.pamd2,1);
  for (node = FIRSTNODE; node< nnodes; node++)
    pvm_send(pinums[node],STARTTYPE2);
  
  pvm_initsend(PvmDataRaw);
  pvm_pkint(pam12x,m_msg0.pamd1*m_msg0.pamd2,1);
  for (node = FIRSTNODE; node< nnodes; node++)
    pvm_send(pinums[node],STARTTYPE3);
#endif
#ifdef MPI_SRC
  for (node=FIRSTNODE; node < nnodes; node++) {
    MPI_Send(&pst,(int)sizeof(pst),MPI_BYTE,node,STARTTYPE2,
	     MPI_COMM_WORLD);
    MPI_Send(sascii,(int)sizeof(sascii)/sizeof(int),MPI_INT,node,STARTTYPE2,
	     MPI_COMM_WORLD);
    MPI_Send(pam12,m_msg0.pamd1*m_msg0.pamd2,MPI_INT,node,STARTTYPE2,
	     MPI_COMM_WORLD);
    MPI_Send(pam12x,m_msg0.pamd1*m_msg0.pamd2,MPI_INT,node,STARTTYPE2,
	     MPI_COMM_WORLD);
  }
#endif
  
  /* Receive HGETLIB data from all the workers */
  nls_cnt0 = nls_tot0 = 0;
  for (node = FIRSTNODE; node < nnodes; node++) {
#ifdef PVM_SRC
    bufid = pvm_recv(-1,STARTTYPE0);
    pvm_bufinfo(bufid,NULL,NULL,&tid);
    snode = tidtonode(tid)-FIRSTNODE;
    pvm_upkint(&lcnt,1,1);
    pvm_freebuf(bufid);
#endif
#ifdef MPI_SRC
    MPI_Recv(&lcnt,1,MPI_INT,MPI_ANY_SOURCE,STARTTYPE0,
	     MPI_COMM_WORLD,&mpi_status);
    snode= mpi_status.MPI_SOURCE;
#endif

    wlsn [snode-FIRSTNODE] = lcnt;
    fprintf(stderr," %d descriptions from %d\n",lcnt,snode);
    if ((desptr=ldes[snode-FIRSTNODE]=
	 (struct sql *)calloc(lcnt+1,sizeof(struct sql)))==NULL)
      s_abort("cannot allocate descriptor space","");

  }
  
  /* have the counts, now get the data */
  ntt = 0l;
  ntt_c = 0;
  for (node = FIRSTNODE; node < nnodes; node++) {
    if (wlsn[node-FIRSTNODE] == 0) {
      fprintf(stderr," skipping descriptions from %d\n",node);
      continue;
    }

    bline_inc = m_msg0.aln.llen;
    max_bline_b = (bline_inc+1) * (wlsn[node-FIRSTNODE]+1);
    if ((bline_buf = (char *)calloc(max_bline_b,sizeof(char)))==NULL)
      s_abort("cannot allocate bline descriptor space","");

    bline_bufp=bline_buf;
    bline_buf_mx = bline_buf+max_bline_b;

#ifdef PVM_SRC
    pvm_initsend(PvmDataRaw);
    pvm_pkint(&m_msg0.n0,1,1);
    pvm_send(pinums[node],STARTTYPE1);
#endif
#ifdef MPI_SRC
    MPI_Send(&m_msg0.n0,1,MPI_INT,node,STARTTYPE1,MPI_COMM_WORLD);
#endif
    for (lcnt=0; lcnt < wlsn[node-FIRSTNODE]; ) {
#ifdef PVM_SRC
      bufid = pvm_recv(pinums[node],ONETYPE);
      pvm_upkbyte((char *)seqds,sizeof(struct sqlc)*NDES,1);
      pvm_freebuf(bufid);
#endif
#ifdef MPI_SRC
      MPI_Recv(seqds,sizeof(struct sqlc)*NDES,MPI_BYTE,
	       node,ONETYPE,MPI_COMM_WORLD,&mpi_status);
#endif
      for (i=0; i<NDES; i++,lcnt++) {
	if (lcnt >= wlsn[node-FIRSTNODE]) break;
	desptr = &ldes[node-FIRSTNODE][lcnt];
	memcpy(desptr,&seqds[i],sizeof(struct sql));
	desptr->bline = bline_bufp;
	strncpy(desptr->bline,seqds[i].bline,bline_inc);
	bline_bufp += bline_inc+1;
      }
    }

    fprintf(stderr," have %d results from %d\n", lcnt, node);

    slsn[node-FIRSTNODE]=wlsn[node-FIRSTNODE];
    for (lcnt=0; lcnt < wlsn[node-FIRSTNODE]; lcnt++) {
      ntt += ldes[node-FIRSTNODE][lcnt].n1;
      if (ntt > MAXLONG) {ntt -= MAXLONG; ntt_c++;}
      if (ldes[node-FIRSTNODE][lcnt].cont) slsn[node-FIRSTNODE]--;
    }
  }
  
  /* print out all descriptions */
  /*
  for (node = FIRSTNODE; node < nnodes; node++)
    for (lcnt = 0; lcnt < wlsn[node-FIRSTNODE]; lcnt ++)
      printf("%2d:%3d\t%s\n",node,lcnt,ldes[node-FIRSTNODE][lcnt].bline);
  */

  /* Calculate cumulative totals and send to workers for a self search */

  clsn [0] = nlib = nclib= 0;
  for (node = FIRSTNODE; node < nnodes-1; node++) {
    /* clsn[] is for the next node */
    clsn[node-FIRSTNODE+1] = nclib += wlsn[node-FIRSTNODE];
    nlib += slsn[node-FIRSTNODE];
  }
  nlib += slsn[nnodes-1-FIRSTNODE];

  if (m_msg0.self)
    for (node = FIRSTNODE; node < nnodes; node++) {
#ifdef PVM_SRC
      pvm_initsend(PvmDataRaw);
      pvm_pkint(&clsn[node-FIRSTNODE],1,1);
      pvm_send(pinums[node],STARTTYPE1);
#endif
#ifdef MPI_SRC
    MPI_Send(&clsn[node-FIRSTNODE],1,MPI_INT,node,STARTTYPE1,MPI_COMM_WORLD);
#endif
      fprintf(stderr,"sending lend: %d to worker %d\n",clsn[node-FIRSTNODE],node);
    }

  last_msg_b[0] = m_msg0.nbr_seq = nlib;

  qres_bufsize = BFR;
  /* if BFR is too big for this library, reduce it */
  while ( nlib*(m_msg0.nitt1+1)/(2*nnodes) < qres_bufsize) {
    qres_bufsize /= 2;
    if ((qres_bufsize%(m_msg0.nitt1+1))!= 0) {
      qres_bufsize *= (m_msg0.nitt1+1);
      break;
    }
    if (qres_bufsize < 50) break;
  }
  last_msg_b[1] = qres_bufsize;

  fprintf(stderr," using BFR=%d/%d\n",qres_bufsize,BFR);

#ifdef PVM_SRC
  pvm_initsend(PvmDataRaw);
  pvm_pkint(last_msg_b,2,1);
  for (node=FIRSTNODE; node < nnodes; node++)
    pvm_send(pinums[node],STARTTYPE0);
#endif
#ifdef MPI_SRC
  for (node=FIRSTNODE; node < nnodes; node++)
    MPI_Send(last_msg_b,2,MPI_INT,node,STARTTYPE0,MPI_COMM_WORLD);
#endif  


  tscan = tprev = s_time();

/**************************************
  The logic of this section has been simplified to allow multistage
  comparison functions to be used and alignments to be generated.

  	send 1st query to workers
	get next query sequence from host (m_msp1)
    L1:	get results from next-1 search (m_msp0)
	sort the results of the next-1 search
	(possibly) do additional stages of search
	(possibly produce alignments for search
	send next query to workers (m_msp1)
	display result of next-1 search (m_msp0)
	get next query sequence from host (m_msp1)
	goto L1;


As a result of the interleaving, there must be two qm_msg structures,
one for the next-1 sequence (which is required for labeling the
output), and one for the next sequence (which is sent to the workers
while the results are being displayed.  qm_msp0 and qm_msp1 alternate
between these two structures.
***************************************/

/*
  qm_msp0 points to the older qm_msg 
  qm_msp1 points to the newer qm_msg
  the assignment below goes with curtype==ONETYPE
*/
  qm_msp0 = &qm_msg0;
  qm_msp1 = &qm_msg1;
  /* skip this; each value should be filled explicitly */
  /* memcpy(qm_msp1,qm_msp0,sizeof(qm_msg0)); */
  
  aa0p0 = aa00;		/* aa0p0 is the "old" sequence */
  aa0p1 = aa01;		/* aa0p1 is the "new" sequence */

/* ******************** */
/* manage code          */
/* ******************** */
/* At this point we have received all the sequence description data
   from the workers and also know the number of sequences in each
   worker's chunk. Now, send query sequence to each worker */

  if (m_msg0.dfile[0] && (fdata=fopen(m_msg0.dfile,"w"))!=NULL)
    fprintf(fdata,"%3d%-50s\n",qlib,qm_msp0->libstr);

#ifdef PVM_SRC
  pvm_initsend(PvmDataRaw);
  pvm_pkbyte((char *)qm_msp0,sizeof(qm_msg0),1);
  if (qm_msp0->n0 > 0) pvm_pkbyte((char *)aa0p0,qm_msp0->n0+1,1);
  for (node = FIRSTNODE; node < nnodes; node++)
    pvm_send(pinums[node],MSEQTYPE);
#endif
#ifdef MPI_SRC
  for (node = FIRSTNODE; node < nnodes; node++) {
    MPI_Send(qm_msp0,sizeof(qm_msg0),MPI_BYTE,node,MSEQTYPE,MPI_COMM_WORLD);
    if (qm_msp0->n0 > 0) {
      MPI_Send(aa0p0,qm_msp0->n0+1,MPI_BYTE,node,MSEQTYPE1,MPI_COMM_WORLD);
    }
  }
#endif

  /* Get next query sequence */ 
  lcont = 0; 	/* required for linux alpha? */

  qm_msp1->n0 = HGETLIB(aa0p1,MAXTST,qm_msp1->libstr, &lmark, &lcont);

  qm_msp1->slist = 0;
  qm_msp1->seqnm = qlib;
  leng = strlen (qm_msp1->libstr);
  sprintf (&(qm_msp1->libstr[leng]), " %d %s", qm_msp1->n0, sqnam);

  naa0 = 0;  /* reset node counter */

  /* sit in loop and collect results */
  nbest = nopt = 0;
  zbestcut = -BIGNUM;

  /* fast_flag == 1 => send new sequences immediately */
  fast_flag = ((m_msg0.stages==1) && (m_msg0.markx != 9) && (m_msg0.ashow == 0));

  while (1) {
#ifdef PVM_SRC
    bufid = pvm_recv(-1,curtype);
    pvm_bufinfo(bufid,NULL,NULL,&tid);
    pvm_upkbyte((char *)&bestr[0],sizeof(struct comstr)*(qres_bufsize+1),1);
    snode = tidtonode(tid)-FIRSTNODE;
    pvm_freebuf(bufid);
#endif

#ifdef MPI_SRC
    MPI_Recv(bestr,sizeof(struct comstr)*(qres_bufsize+1),
	     MPI_BYTE,MPI_ANY_SOURCE,curtype,MPI_COMM_WORLD,&mpi_status);
    snode = mpi_status.MPI_SOURCE;
#endif

    nres = bestr[qres_bufsize].seqnm & ~FINISHED;
    if (bestr[qres_bufsize].seqnm&FINISHED) {	/* a worker is finished */
      naa0++;

      /* send a new query sequence if no more processing required */
      if (fast_flag) {
#ifdef PVM_SRC
	pvm_initsend(PvmDataRaw);
	pvm_pkbyte((char *)qm_msp1,sizeof(qm_msg1),1);
	if (qm_msp1->n0 != -1) pvm_pkbyte((char *)aa0p1,qm_msp1->n0+1,1);
	pvm_send(tid,MSEQTYPE);
#endif
#ifdef MPI_SRC
	MPI_Send(qm_msp1,sizeof(qm_msg1),MPI_BYTE,snode,MSEQTYPE,MPI_COMM_WORLD);
	if (qm_msp1->n0 != -1)
	  MPI_Send(aa0p1,qm_msp1->n0+1,MPI_BYTE,snode,MSEQTYPE1,MPI_COMM_WORLD);
#endif
      }
    }

#ifdef DEBUG
    if (pst.debug_lib)
      fprintf(stderr," unpacking %d from %d; nbest %d\n",nres,snode,nbest);
#endif

    /* this section is now more complex because can get groups of
       sequence results; e.g. forward and reverse frame */

    t_best = t_rbest = -1;
    for (ires = 0; ires < nres; ires++) {
      desptr = &ldes[snode-FIRSTNODE][bestr[ires].seqnm];

      /* save raw results */
      if (fdata) {
	strncpy(libstr,desptr->bline,10);
	if ((bp=strchr(libstr,' '))!=NULL) *bp='\0';
	fprintf(fdata,"%-10s\t%4d\t%4d\t%d\t%4d\t%4d\t%4d\t%8ld\n",
		libstr,desptr->sfnum[0],desptr->n1,bestr[ires].frame,
		bestr[ires].score[0],bestr[ires].score[1],bestr[ires].score[2],
		desptr->lseek);
      }

      i_score = bestr[ires].score[pst.score_ix];
      k_comp = bestr[ires].comp;
      k_H = bestr[ires].H;

      if (i_score > t_best) {
	tm_best = t_best = i_score;
	t_n1 = desptr->n1;
      }

      if (pst.zsflag >= 10 && bestr[ires].rscore > t_rbest)
	t_rbest = bestr[ires].rscore;

      if (nstats < MAXSTATS) {
	if (bestr[ires].frame == m_msg0.nitt1) {
	  stats[nstats].n1 = t_n1;
	  stats[nstats].comp = k_comp;
	  stats[nstats].H = k_H;

	  if (pst.zsflag > 10) {
	    tm_best = t_rbest;
	    t_rbest = -1;
	  }
	  stats[nstats++].score = tm_best;
	  t_best = -1;
	}
      }
      else if (!dohist && pst.zsflag >=0 ) {
	pst.n0 = qm_msp0->n0;
	process_hist(stats,nstats,pst);
	dohist = 1;
	for (i=0; i<nbest; i++)
	  bptr[i]->zscore = (*find_zp)(bptr[i]->score[pst.score_ix],
				       bptr[i]->n1, bptr[i]->comp);
      }

      if (dohist) {
	zscore=(*find_zp)(i_score,desptr->n1,k_comp);
	if (bestr[ires].frame == m_msg0.nitt1) {
	  addhistz((*find_zp)(tm_best,t_n1,k_comp));
	  t_best = t_rbest = -1;
	}
      }
      else zscore = (float) i_score;

      if (zscore > zbestcut) {
	if (nbest>=MAXBEST) {
	  selectbestz(bptr, nbest-MAXBEST/4-1, nbest); 
	  nbest -= MAXBEST/4;
	  zbestcut = bptr[nbest-1]->zscore;
	  best_flag = 0;
	}
	/* if zbestcut == -BIGNUM, bptr[] has not been reinitialized */
	else if (best_flag) bptr[nbest]=&best[nbest];

	bptr[nbest]->seqnm  = bestr[ires].seqnm ;
	bptr[nbest]->score[0] = bestr[ires].score[0];
	bptr[nbest]->score[1] = bestr[ires].score[1];
	bptr[nbest]->score[2] = bestr[ires].score[2];
	bptr[nbest]->comp = bestr[ires].comp;
	bptr[nbest]->H = bestr[ires].H;

	bptr[nbest]->zscore = zscore;
	bptr[nbest]->wrkr   = snode;
	bptr[nbest]->desptr = desptr;
	bptr[nbest]->lseek = desptr->lseek;
	bptr[nbest]->n1 = desptr->n1;
	bptr[nbest]->frame = bestr[ires].frame;

	/*
	bptr[nbest]->swscore = bestr[ires].swscore;
	if (bestr[ires].swscore > -1) {
	  nopt++;
	  bptr[nbest]->a_len = bestr[ires].a_len;
	  bptr[nbest]->percent = bestr[ires].percent;
	  bptr[nbest]->gpercent = bestr[ires].gpercent;
	  bptr[nbest]->min0 = bestr[ires].min0;
	  bptr[nbest]->min1 = bestr[ires].min1;
	  bptr[nbest]->max0 = bestr[ires].max0;
	  bptr[nbest]->max1 = bestr[ires].max1;
	  bptr[nbest]->ngap_q = bestr[ires].ngap_q;
	  bptr[nbest]->ngap_l = bestr[ires].ngap_l;
	}
	else {
	  bptr[nbest]->percent = -1.0;
	  bptr[nbest]->min0 = bptr[nbest]->min1 = bptr[nbest]->max0 = 
	    bptr[nbest]->max1 = 0;
	}
	*/

	nbest++;
      }
    }	/* for loop */
    if (naa0 < nnodes-FIRSTNODE) continue;

    gstring2[0]='\0';

#ifdef PVM_SRC
    bufid = pvm_recv(pinums[FIRSTNODE],PARAMTYPE);
    pvm_upkbyte(gstring2,sizeof(gstring2),1);
    pvm_upkbyte(gstring3,sizeof(gstring3),1);
    pvm_freebuf(bufid);
#endif
#ifdef MPI_SRC
    MPI_Recv(gstring2,sizeof(gstring2),MPI_BYTE,FIRSTNODE,PARAMTYPE,
	     MPI_COMM_WORLD,&mpi_status);
#endif
/* **************************************************************** */
/* if no more worker computation is required, start the next search */
/* **************************************************************** */


/* *************************** */
/* analyze the last results    */
/* *************************** */
    
    if (!dohist) {
      if (nbest < 20 || pst.zsflag <= 0) {
	pst.zsflag = -1;
      }
      else {
	pst.n0 = qm_msp0->n0;
	process_hist(stats,nstats,pst);
	for (i=0; i<nbest; i++)
	  bptr[i]->zscore = (*find_zp)(bptr[i]->score[pst.score_ix],
				       bptr[i]->n1,bptr[i]->comp);
      }
    }

    if (!m_msg0.noshow) {
      if (pst.zsflag>=0) {
	sortbestz (bptr, nbest);
	for (i=0; i<nbest; i++)
	  bptr[i]->escore = zs_to_E(bptr[i]->zscore,bptr[i]->n1,pst.dnaseq);
	sortbeste(bptr, nbest);
      }
      else sortbest(bptr,nbest,pst.score_ix);

/*      else sortorder(bptr,nbest,wlsn,nnodes); */

/* if more than one stage or markx==9, calculate opt scores or do alignment */
/* send results to workers as available */

/* wstage2[] FIRSTNODE .. nnodes has the next sequence to be do_opt()/do_walign()ed */
/* wstage2p[] is a list of sequence numbers/frames, to be sent to workers */
/* wstage2b[] is a list of bptr's that shares the index with wstage2p[] */

      if (m_msg0.stages > 1 || m_msg0.markx==9) {

	/* calculate a new mshow if necessary */
	if (pst.zsflag >= 0) {
	  for (i=0; i<nbest && bptr[i]->escore< m_msg0.e_cut; i++) {}
	  m_msg0.mshow = i;
	}

	/* allocate space for a_struct info */
	if (m_msg0.markx==9) {
	  if ((aln_d_base=(struct a_struct *)
	       calloc((size_t)m_msg0.mshow,sizeof(struct a_struct)))==NULL) {
	    fprintf(stderr," cannot allocate a_struct %d\n", m_msg0.mshow);
	    exit(1);
	  }
	}

	/* initialize the counter for each worker to 0 */
	for (iw = FIRSTNODE; iw < nnodes; iw++) wstage2[iw] = 0;

	/* for each result, bump the counter for the worker
	   that has the sequence */

	for (is = 0; is < m_msg0.mshow; is++ ) {
	  wstage2[bptr[is]->wrkr]++;
	  bptr[is]->aln_d = &aln_d_base[is];
	}

	/* now allocate enough space to send each worker a
	   list of its sequences stage2_str {seqnm, frame} */
	for (iw = FIRSTNODE; iw < nnodes; iw++) {
	  if (wstage2[iw]>0) {
	    if ((wstage2p[iw]=
		 (struct stage2_str *)
		 calloc(wstage2[iw],sizeof(struct stage2_str)))==NULL) {
	      sprintf(errstr," cannot allocate sequence listp %d %d",
		      iw,wstage2[iw]);
	      s_abort(errstr,"");
	    }

	    /* allocate space to remember the bptr's for each result */
	    if ((wstage2b[iw]=(struct beststr **)
		 calloc(wstage2[iw],sizeof(struct beststr *)))==NULL) {
	      sprintf(errstr," cannot allocate sequence listb %d %d",
		      iw,wstage2[iw]);
	      s_abort(errstr,"");
	    }
	    wstage2i[iw]=0;
	  }
	  else {
	    (void *)wstage2p[iw] = (void *)wstage2b[iw] = NULL;
	  }
	}

	/* for each result, set wstage2p[worker][result_index_in_worker] */
	for (is = 0; is < m_msg0.mshow; is++) {
	  iw=bptr[is]->wrkr;
	  wstage2p[iw][wstage2i[iw]].seqnm = bptr[is]->seqnm;
	  wstage2p[iw][wstage2i[iw]].frame = bptr[is]->frame;
	  wstage2b[iw][wstage2i[iw]] = bptr[is];
	  wstage2i[iw]++;
	}
	/* at this point, wstage2i[iw] should equal wstage2[iw] */

	naa0 = 0;
	for (node = FIRSTNODE; node < nnodes; node++) {

	  /* if a worker has no results, move on */
	  if (wstage2[node]<=0) { naa0++; continue;}

	  qm_msp0->slist = wstage2[node];	/* set number of results to return */
	  qm_msp0->s_func = 1;			/* set s_funct for do_opt/do_walign */
#ifdef PVM_SRC
	  pvm_initsend(PvmDataRaw);
	  pvm_pkbyte((char *)qm_msp0,sizeof(qm_msg0),1);
	  pvm_send(pinums[node],MSEQTYPE);	/* send qm_msp0 */

	  pvm_initsend(PvmDataRaw);		/* send the list of seqnm/frame */
	  pvm_pkbyte((char *)wstage2p[node],wstage2[node]*sizeof(struct stage2_str),1);
	  pvm_send(pinums[node],LISTTYPE);
#endif
#ifdef MPI_SRC
	MPI_Send(qm_msp0,sizeof(qm_msg0),MPI_BYTE,node,MSEQTYPE,
		 MPI_COMM_WORLD);
	MPI_Send((char *)wstage2p[node],wstage2[node]*
		 sizeof(struct stage2_str),MPI_BYTE,node,LISTTYPE,
		 MPI_COMM_WORLD);
#endif
	}
	
	/* reset the index of results to obtain */
	for (iw = 0; iw < nnodes; iw++) wstage2i[iw]=0;
	
	while (naa0 < nnodes-FIRSTNODE) {
#ifdef PVM_SRC
	  bufid = pvm_recv(-1,LISTRTYPE);	/* wait for results */
	  pvm_bufinfo(bufid,NULL,NULL,&tid);
	  /* get a chunk of comstr2 results */
	  pvm_upkbyte((char *)&bestr2[0],sizeof(struct comstr2)*(BFR2+1),1);
	  snode = (iw=tidtonode(tid))-FIRSTNODE;
	  pvm_freebuf(bufid);
#endif
#ifdef MPI_SRC
	  MPI_Recv((char *)&bestr2[0],sizeof(struct comstr2)*(BFR2+1),
		   MPI_BYTE,MPI_ANY_SOURCE,LISTRTYPE,MPI_COMM_WORLD,
		   &mpi_status);
	  snode = mpi_status.MPI_SOURCE;
	  iw = snode;
#endif

	  /* get number of results in this message */
	  nres = bestr2[BFR2].seqnm & ~FINISHED;
	  /* check to see if finished */
	  if (bestr2[BFR2].seqnm&FINISHED) {naa0++;}

	  /* count through results from a specific worker */
	  for (i=0,is=wstage2i[iw]; i < nres; i++,is++) {
	    /* get the (saved) bptr for this result */
	    bbptr=wstage2b[iw][is];
	    /* consistency check seqnm's must agree */
	    if (wstage2p[iw][is].seqnm ==  bbptr->seqnm) {
	      if (m_msg0.stages > 1) {
		bbptr->score[0] = bestr2[i].score[0];
		bbptr->score[1] = bestr2[i].score[1];
		bbptr->score[2] = bestr2[i].score[2];
	      }
	      if (m_msg0.markx == 9) {
		bbptr->swscore = bestr2[i].swscore;
		memcpy(bbptr->aln_d,&bestr2[i].aln_d,sizeof(struct a_struct));
		/*
		bbptr->a_len = bestr2[i].a_len;
		bbptr->min0 = bestr2[i].min0;
		bbptr->max0 = bestr2[i].max0;
		bbptr->min1 = bestr2[i].min1;
		bbptr->max1 = bestr2[i].max1;
		bbptr->ngap_q = bestr2[i].ngap_q;
		bbptr->ngap_l = bestr2[i].ngap_l;
		*/
		bbptr->percent = bestr2[i].percent;
		bbptr->gpercent = bestr2[i].gpercent;
	      }
	    }
	    else fprintf(stderr,"phase error in phase II return %d %d",
			 iw,i);
	  }
	  wstage2i[iw] += nres;
	}
	for (iw=FIRSTNODE; iw < nnodes; iw++) {
	  if ((void *)wstage2p[iw]!=NULL) free((void *)wstage2p[iw]);
	  if ((void *)wstage2b[iw]!=NULL) free((void *)wstage2b[iw]);
	}
      }

      tdone = s_time();
      tddone = time(NULL);

      fprintf (outfd, "%3d>>%s\n", qlib,qm_msp0->libstr);

      m_msg0.n0 = qm_msp0->n0;
      strncpy(m_msg0.libstr,qm_msp0->libstr,sizeof(m_msg0.libstr));
      prhist (outfd,m_msg0,pst,gstring2);

      if (bptr[0]->escore < m_msg0.e_cut) {
	showbest (outfd, bptr, nbest, qlib, &m_msg0,pst,gstring2);

	if (m_msg0.markx==9) {
	  fprintf(outfd,"\n>>>%s#%d%s%s, %d %s vs %s library\n",
		  m_msg0.tname,qlib,qm_msp0->libstr,
		  (m_msg0.revcomp ? "-":"\0"), qm_msp0->n0, m_msg0.sqnam,
		  m_msg0.lname);
	}
	else if (m_msg0.markx==10) {
	  if ((bp=strchr(qm_msp0->libstr,' '))!=NULL) *bp = '\0';
	  fprintf(outfd,"\n>>>%s#%d%s%s, %d %s vs %s library\n",
		  m_msg0.tname,qlib,qm_msp0->libstr,
		  (m_msg0.revcomp ? "-":"\0"), qm_msp0->n0, m_msg0.sqnam,
		  m_msg0.lname);
	  if (bp!=NULL) *bp=' ';
	  fprintf(outfd,"; mp_name: %s\n",argv[0]);
	  fprintf(outfd,"; mp_ver: %s\n",mp_verstr);
	  fprintf(outfd,"; mp_argv:");
	  for (i=0; i<argc; i++)
	    fprintf(outfd," %s",argv[i]);
	  fputc('\n',outfd);
	  fputs(gstring3,outfd);
	  fputs(hstring1,outfd);
	}

	if (m_msg0.ashow != 0)
	  m_msg0.n0 = qm_msp0->n0;
	  strncpy(m_msg0.libstr,qm_msp0->libstr,sizeof(m_msg0.libstr));
	  showalign (outfd, aa0p0, bptr, nbest, qlib, m_msg0, pst, gstring2);
	/*	else fprintf(stderr,"%s %d opt\n",qm_msp0->libstr,nopt); */
      }
      else {
	if (m_msg0.markx<9) fprintf(outfd,"!! No sequences with E() < %f\n",m_msg0.e_cut);
	else {
	  fprintf(outfd,"\n>>>%s#%d%s%s, %d %s vs %s library\n",
		  m_msg0.tname,qlib,qm_msp0->libstr,(m_msg0.revcomp ? "-":"\0"), qm_msg0.n0, m_msg0.sqnam,
		  m_msg0.lname);
	  fprintf(outfd,">>>!!! No sequences with E() < %f\n",m_msg0.e_cut);
	}
      }
    }

    summ_stats(stat_string);
    if (m_msg0.markx < 9 ) {
    /*
      fprintf(outfd,"/ ** %s ** /\n",gstring2);
      fprintf(outfd,"/ ** %s ** /\n",stat_string);
    */
      fprintf(outfd,"/** search time: ");
      ptime(outfd,tdone-tprev);
      fprintf(outfd," **/\n");
      tprev = tdone;
    }
    else if (m_msg0.markx == 9) {
      fprintf(outfd,">>>***\n");
      fprintf(outfd,"/** %s **/\n",gstring2);
      fprintf(outfd,"/** %s **/\n",stat_string);
      fprintf(outfd,">>><<<\n");
    }
    else if (m_msg0.markx == 10) {
      fprintf(outfd,">>><<<\n");
    }
    fflush(outfd);
    
/* *********************** */
/* end of analysis/display */
/* *********************** */


/* *********************** */
/* start the next search   */                                           
/* *********************** */

    if (fdata) {
      fprintf(fdata,"/** %s **/\n",gstring2);
      fprintf(fdata,"%3d%-50s\n",qlib-1,qm_msp1->libstr);
      fflush(fdata);
    }
    
    dohist = 0;
    nstats = 0;

    if (!fast_flag) {
#ifdef PVM_SRC
      pvm_initsend(PvmDataRaw);
      pvm_pkbyte((char *)qm_msp1,sizeof(qm_msg1),1);
      if (qm_msp1->n0 != -1) pvm_pkbyte((char *)aa0p1,qm_msp1->n0+1,1);
      for (node = FIRSTNODE; node < nnodes; node++)
	pvm_send(pinums[node],MSEQTYPE);
#endif
#ifdef MPI_SRC
      for (node=FIRSTNODE; node < nnodes; node++) {
	MPI_Send(qm_msp1,sizeof(qm_msg1),MPI_BYTE,node,MSEQTYPE,
		 MPI_COMM_WORLD);
	if (qm_msp1->n0 != -1) {
	  MPI_Send(aa0p1,qm_msp1->n0+1,MPI_BYTE,node,MSEQTYPE1,MPI_COMM_WORLD);
	}
      }
#endif
    }

    if (qm_msp1->n0 != -1) {
      qtt += qm_msp1->n0;
      qlib++;
    }
    if (qm_msp1->n0 == -1) goto done;
    
    naa0 = 0;
    best_flag = 1;
    nbest = nopt = 0;
    zbestcut = -BIGNUM;
    if (curtype == ONETYPE) {
      curtype = TWOTYPE;
      qm_msp0 = &qm_msg1;
      qm_msp1 = &qm_msg0;
      aa0p0 = aa01;
      aa0p1 = aa00;
    }
    else  {
      curtype = ONETYPE;
      qm_msp0 = &qm_msg0;
      qm_msp1 = &qm_msg1;
      aa0p0 = aa00;
      aa0p1 = aa01;
    }

    lcont = 0;	/* required for linux alpha */
    qm_msp1->n0 = HGETLIB(aa0p1,MAXTST,qm_msp1->libstr, &lmark, &lcont);
    qm_msp1->slist = 0;
    leng = strlen (qm_msp1->libstr);
    sprintf (&(qm_msp1->libstr[leng]), " %d %s", qm_msp1->n0, sqnam);
    qm_msp1->seqnm = qlib;
  }	    /* while loop */
  
  /* ******************** */
  /* end of manage code   */
  /* ******************** */

 done:
  tdone = s_time();
  if (m_msg0.markx == 10 || m_msg0.markx == 9) fputs(">>>///\n",outfd);
  printsum(outfd);
  if (outfd!=stdout) printsum(stdout);
  printsum(stderr);
#ifdef PVM_SRC
  pvm_exit();
#endif
#ifdef MPI_SRC
  MPI_Finalize();
#endif
  h_closelib();
  exit(0);
}   /* End of main program */

scanseq(unsigned char *seq, int n, char *str)
{
  int tot,i;
  char aaray[MAXSQ];		/* this must be set > nsq */
	
  for (i=0; i<MAXSQ; i++)  aaray[i]=0;
  for (i=0; i<strlen(str); i++) aaray[sascii[str[i]]]=1;
  for (i=tot=0; i<n; i++) tot += aaray[seq[i]];
  return tot;
}

void
printsum(FILE *fd)
{
  char tstr1[26], tstr2[26];
  double db_tt;

  strncpy(tstr1,ctime(&tdstart),sizeof(tstr1));
  strncpy(tstr2,ctime(&tddone),sizeof(tstr1));
  tstr1[24]=tstr2[24]='\0';

  /* Print timing to output file as well */

  if (qtt_c==0) {
    fprintf(fd, "%ld residues in %d query   sequences\n", qtt, qlib);
  }
  else {
    db_tt = (double)qtt_c*(double)MAXLONG + (double)qtt;
    fprintf(fd, "%.0f residues in %d query   sequences\n", db_tt, qlib);
  }

  if (ntt_c==0) {
    fprintf(fd, "%ld residues in %d library sequences\n", ntt, nlib);
  }
  else {
    db_tt = (double)ntt_c*(double)MAXLONG + (double)ntt;
    fprintf(fd, "%.0f residues in %d library sequences\n", db_tt, nlib);
  }

  /*  if (m_msg0.lsize_cut > 0)
    fprintf(fd, "%d residues excluded from %d sequences < %d\n",
	    nls_tot0,nls_cnt0,m_msg0.lsize_cut);
	    */

  fprintf(fd," %d processors (%d workers) were used\n",
	  nnodes+-FIRSTNODE+1,nnodes-FIRSTNODE);
  fprintf(fd," Pvcomplib [%s]\n start: %s done: %s\n",mp_verstr,tstr1,tstr2);
  fprintf(fd," Loading time: ");
  ptime(fd, tscan - tstart);
  fprintf (fd," Scan time: ");
  ptime (fd, tdone - tscan);
  fprintf (fd,"\n");
  fprintf (fd, "\nFunction used was %s [%s]\n", prog_func, verstr);
}

void fsigint()
{
  int i;

  tdone = s_time();
  tddone = time(NULL);

  if (outfd!=stdout) fprintf(outfd,"/*** interrupted ***/\n");
  fprintf(stderr,"/*** interrupted ***/\n");

  printsum(stdout);
  if (outfd!=stdout) printsum(outfd);

#ifdef PVM_SRC
  for (i=FIRSTNODE; i<nnodes; i++) pvm_kill(pinums[i]);
  pvm_exit();
#endif
#ifdef MPI_SRC
  MPI_Abort(MPI_COMM_WORLD,1);
  MPI_Finalize();
#endif
  exit(1);
}
