
/* copyright (c) 1996, 1997, 1998, 1999 William R. Pearson and the
   U. of Virginia */

/* $Name: fa33t09d0 $ - $Id: uthr_subs.c,v 1.3 2000/01/09 18:04:22 wrp Exp $ */

/* this file isolates the pthreads calls from the main program */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <signal.h>

#include "param.h"

#define XTERNAL
#include "thr.h"
#undef XTERNAL
#include "uthr_subs.h"

extern void *work_thread (void *work_info);

/* start the threads working */

void init_thr(int nthreads, struct thr_str *work_info)
{
  int status, i;

  if (nthreads > MAX_WORKERS)  {
    fprintf ( stderr," cannot start %d threads, max: %d\n",
	      nthreads, MAX_WORKERS);
    exit(1);
  }

  /* mutex and condition variable initialisation */

  status = mutex_init(&reader_mutex, USYNC_THREAD, 0);
  check(status,"Reader_mutex init bad status\n");
   
  status = mutex_init(&worker_mutex, USYNC_THREAD, 0);
  check(status,"Worker_mutex init bad status\n");

  status = cond_init(&reader_cond_var, USYNC_THREAD, 0);
  check(status,"Reader_cond_var init bad status\n");

  status = cond_init(&worker_cond_var, USYNC_THREAD, 0);
  check(status,"Worker_cond_var init bad status\n");

  status = mutex_init(&start_mutex, USYNC_THREAD, 0);
  check(status,"Start_mutex init bad status\n");

  status = cond_init(&start_cond_var, USYNC_THREAD, 0);
  check(status,"Start_cond_var init bad status\n");

  /* start the worker threads */

  for (i=0; i < nthreads; i++) {
    status=thr_create(NULL, 1000000, &work_thread,
		      (void *)&work_info[i],
		      THR_BOUND, &threads[i]);
    check(status,"create failed\n");
  }
}

void start_thr()
{
  int status;

  /* tell threads to proceed */

  status = mutex_lock(&start_mutex);
  check(status,"Start_mutex lock bad status in main\n");

  start_thread = 0;  /* lower predicate */

  status = cond_broadcast(&start_cond_var);
  status = mutex_unlock(&start_mutex);
  check(status,"Start_mutex unlock bad status in main\n");
}

void get_rbuf(struct buf_head **cur_buf, int max_work_buf)
{
  int status;

  status = mutex_lock(&reader_mutex);  /* lock reader_buf structure */

  check(status,"Reader_mutex lock in master bad status\n");

  /* no reader bufs:  wait for signal to proceed */
  while (num_reader_bufs == 0) {
    cond_wait(&reader_cond_var,&reader_mutex);
  }

  *cur_buf = reader_buf[reader_buf_readp];  /* get the buffer address */
  reader_buf_readp = (reader_buf_readp+1)%(max_work_buf);  /* increment index */
  num_reader_bufs--;

  status = mutex_unlock(&reader_mutex);  /* unlock structure */
  check(status,"Reader_mutex unlock in master bad status\n");
}

void put_rbuf(struct buf_head *cur_buf, int max_work_buf)
{
  int status;

  /* give the buffer to a thread, and wait for more */
  status = mutex_lock(&worker_mutex);  /* lock worker_buf_structure */
  check(status,"Worker_mutex lock in master bad status\n");

  /*  Put buffer onto available for workers list */
  worker_buf[worker_buf_readp] = cur_buf;
  worker_buf_readp = (worker_buf_readp+1)%(max_work_buf);
  num_worker_bufs++;   /* increment number of buffers available to workers */

  /*  Signal one worker to wake and start work */
  status = cond_signal(&worker_cond_var);

  status = mutex_unlock(&worker_mutex);
  check(status,"Worker_mutex unlock in master bad status\n"); 
}

void put_rbuf_done(int nthreads, struct buf_head *cur_buf, int max_work_buf)
{
  int status, i;
  void *exit_value;

  /* give the buffer to a thread, and wait for more */
  status = mutex_lock(&worker_mutex);  /* lock worker_buf_structure */
  check(status,"Worker_mutex lock in master bad status\n");

  /*  Put buffer onto available for workers list */
  worker_buf[worker_buf_readp] = cur_buf;
  worker_buf_readp = (worker_buf_readp+1)%(max_work_buf);
  num_worker_bufs++;   /* increment number of buffers available to workers */

  /*  Signal one worker to wake and start work */

  reader_done = 1;
  status = cond_broadcast(&worker_cond_var);

  status = mutex_unlock(&worker_mutex);
  check(status,"Worker_mutex unlock in master bad status\n"); 

  /* wait for all buffers available (means all do_workers are done) */
 
  for (i=0; i < nthreads; i++) {
    status = thr_join( threads[i], NULL, &exit_value);
    check(status,"join bad status\n");

/*
    status = thr_kill( threads[i]);
    check(status,"detach bad status\n");
*/
  } 
}

void wait_thr()
{
  int status;

  /* Wait on master to give start signal */
  status = mutex_lock(&start_mutex);
  check(status,"Start_mutex lock bad status in worker\n");

  while (start_thread) {
         status = cond_wait(&start_cond_var, &start_mutex);
         check(status,"Start_cond_wait bad status in worker\n");
  }

  status = mutex_unlock(&start_mutex);
  check(status,"Start_mutex unlock bad status in worker\n");
}

int get_wbuf(struct buf_head **cur_buf, int max_work_buf)
{
  int status;

  /* get a buffer to work on */
  status = mutex_lock(&worker_mutex);
  check(status,"First worker_mutex lock in worker bad status\n");

  /*  No worker_bufs available:  wait for reader to produce some */
  while (num_worker_bufs == 0) {
    /*  Exit if reader has finished */
    if (reader_done) {
      mutex_unlock(&worker_mutex);
      return 0;
    }
    cond_wait(&worker_cond_var,&worker_mutex);
  } /* end while */

  /*  Get the buffer from list */
  *cur_buf = worker_buf[worker_buf_workp];
  worker_buf_workp = (worker_buf_workp+1)%(max_work_buf);
  num_worker_bufs--;

  status = mutex_unlock(&worker_mutex);
  check(status,"First worker_mutex unlock in worker bad status\n");
  return 1;
}

void put_wbuf(struct buf_head *cur_buf, int max_work_buf)
{
  int status;

  /* put buffer back on list for reader */
  status = mutex_lock(&reader_mutex);
  check(status,"Reader_mutex lock in worker bad status\n");
    
  reader_buf[reader_buf_workp] = cur_buf;
  reader_buf_workp = (reader_buf_workp+1)%(max_work_buf);
  num_reader_bufs++;

     /* No reader_bufs available:  wake reader */
  if (num_reader_bufs == 1) {
    cond_signal(&reader_cond_var);
  }

  status = mutex_unlock(&reader_mutex);
  check(status,"Reader_mutex unlock in worker bad status\n");
}
