
/* copyright (c) 1996, 1997, 1998, 1999 William R. Pearson and the
   U. of Virginia */

/* $Name: fa33t09d0 $ - $Id: workacc.c,v 1.5 2000/07/21 22:41:23 wrp Exp $ */

/* Concurrent read version */

#include <stdio.h>
#include <stdlib.h>

#include "param.h"

#define XTERNAL
#include "uascii.h"
#include "upam.h"
#undef XTERNAL

char err_str[128];

/* Initialization - set up defaults - assume protein sequence */
void w_init ()
{
    sascii = aascii;
}

/* allocate memory for pam12 matrix */
int alloc_pam12 (int d1, int d2)
{

  if ((pam12 = (int *) calloc ((size_t)d1 * d2,sizeof (int))) == NULL) {
    sprintf(err_str,"Cannot allocate 2d pam matrix: %d %d",d1,d2);
    return -1;
  }
  if ((pam12x = (int *) calloc ((size_t)d1 * d2,sizeof (int))) == NULL) {
    sprintf(err_str,"Cannot allocate 2d pam matrix: %d %d",d1,d2);
    return -1;
  }
  else return 1;
}


/* allocate memory for pam matrix */
int    alloc_pam2 (int d1, int d2, struct pstruct *ppst)
{
  int     i, *d2p;

  if ((ppst->pam2[0] = (int **) calloc ((size_t)d1,sizeof (int *))) == NULL) {
    sprintf (err_str,"Cannot allocate pst->pam2[0] matrix %d",d1);
    return -1;
  }

  if ((ppst->pam2[1] = (int **) calloc ((size_t)d1,sizeof (int *))) == NULL) {
    sprintf (err_str,"Cannot allocate pst->pam2[0] matrix %d",d1);
    return -1;
  }

  for (i = 0, d2p = pam12;  i < d1; i++, d2p += d2) ppst->pam2[0][i] = d2p;
  for (i = 0, d2p = pam12x; i < d1; i++, d2p += d2) ppst->pam2[1][i] = d2p;
  return 1;
}

void
aancpy(char *to, char *from, int count, struct pstruct pst)
{
  char *tp, *sq;
  int nsq;

  tp=to;

  if (pst.ext_sq_set) {
    nsq = pst.nsqx;
    sq = pst.sqx;
  }
  else {
    nsq = pst.nsq;
    sq = pst.sq;
  }

  while (count-- && *from) {
    if (*from <= nsq) *tp++ = sq[*(from++)];
    else *tp++ = *from++;
  }
  *tp='\0';
}


/* copies from from to to shuffling */

void
shuffle(unsigned char *from, unsigned char *to, int n)
{
  int i,j; unsigned char tmp;

  if (from != to) memcpy((void *)to,(void *)from,(size_t)n);

  for (i=n; i>1; i--) {
    j = nrand(i);
    tmp = to[j];
    to[j] = to[i-1];
    to[i-1] = tmp;
  }
  to[n] = 0;
}

/* copies from from to from shuffling */
void
wshuffle(unsigned char *from, unsigned char *to, int n, int wsiz, int *ieven)
{
  int i,j, k, mm; 
  unsigned char tmp, *top;

  memcpy((void *)to,(void *)from,n);
	
  mm = n%wsiz;

  if (*ieven) {
    for (k=0; k<(n-wsiz); k += wsiz) {
      top = &to[k];
      for (i=wsiz; i>1; i--) {
	j = nrand(i);
	tmp = top[j];
	top[j] = top[i-1];
	top[i-1] = tmp;
      }
    }
    top = &to[n-mm];
    for (i=mm; i>1; i--) {
      j = nrand(i);
      tmp = top[j];
      top[j] = top[i-1];
      top[i-1] = tmp;
    }
    *ieven = 0;
  }
  else {
    for (k=n; k>=wsiz; k -= wsiz) {
      top = &to[k-wsiz];
      for (i=wsiz; i>1; i--) {
	j = nrand(i);
	tmp = top[j];
	top[j] = top[i-1];
	top[i-1] = tmp;
      }
    }
    top = &to[0];
    for (i=mm; i>1; i--) {
      j = nrand(i);
      tmp = top[j];
      top[j] = top[i-1];
      top[i-1] = tmp;
    }
    *ieven = 1;
  }
  to[n] = 0;
}

void initseq(char **seqc0, char **seqc1, int seqsiz)	/* initialize arrays */
{
  *seqc0=(char *)calloc((size_t)seqsiz,sizeof(char));
  *seqc1=(char *)calloc((size_t)seqsiz,sizeof(char));
  if (*seqc0==NULL || *seqc1==NULL)
    {fprintf(stderr,"cannot allocate consensus arrays %d\n",seqsiz);
     exit(1);}
}

void freeseq(char **seqc0, char **seqc1)
{
  free(*seqc0); free(*seqc1);
}

#ifndef MPI_SRC
void
revcomp(seq,n)
	unsigned char *seq; int n;
{
  unsigned char tmp;
  int i, ni;

  for (i=0; i< n; i++)
    if (nt[seq[i]]=='A') seq[i] = nascii['T'];
    else if (nt[seq[i]]=='C') seq[i] = nascii['G'];
    else if (nt[seq[i]]=='G') seq[i] = nascii['C'];
    else if (nt[seq[i]]=='T') seq[i] = nascii['A'];
    else if (nt[seq[i]]=='R') seq[i] = nascii['Y'];
    else if (nt[seq[i]]=='Y') seq[i] = nascii['R'];
    else if (nt[seq[i]]=='M') seq[i] = nascii['K'];
    else if (nt[seq[i]]=='K') seq[i] = nascii['M'];
    else if (nt[seq[i]]=='D') seq[i] = nascii['H'];
    else if (nt[seq[i]]=='H') seq[i] = nascii['D'];
    else if (nt[seq[i]]=='V') seq[i] = nascii['B'];
    else if (nt[seq[i]]=='B') seq[i] = nascii['V'];

  for (i=0, ni = n-1; i< n/2; i++,ni--) {
    tmp = seq[i];
    seq[i] = seq[ni];
    seq[ni] = tmp;
  }
}
#endif
